package edu.tuwien.kbs;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Calendar;

import org.apache.commons.lang3.text.WordUtils;

import edu.tuwien.kbs.keywords.KeywordGenerator;
import edu.tuwien.kbs.keywords.KeywordResults;
import edu.tuwien.kbs.ldm.SemSearchLDM;
import edu.tuwien.kbs.search.*;
import edu.tuwien.kbs.util.Util;

public class SemSearchCli {

    public static String PropertiesFileName = "config.properties";

    public static void main(String args[]) { // throws FileNotFoundException

	
	// Start time
	Calendar cal1 = Calendar.getInstance();
	System.out.println("Start: " + cal1.getTime().toString());
	long start = System.nanoTime();

	// Main
	ISemSearch s = null;
	String useCase = Util.getValueFromArgs(args, "c");
	boolean useKeyword = Util.getBoolFromArgs(args, "k");
	boolean log = Util.getBoolFromArgs(args, "l");
	//boolean keepViews = Util.getBoolFromArgs(args, "v");
	String query = Util.getValueFromArgs(args, "q");
	String from = Util.getValueFromArgs(args, "p");
	String limit = Util.getValueFromArgs(args, "r1");
	int timeLimit = 600;
	if(limit!=null) timeLimit = Integer.valueOf(limit);
	
	if (useCase == null) throw new RuntimeException("Please choose either NR, VIA, LDM.");
		
	if (useCase.equals("NR")) {
	    s = new SemSearchNR(query, from, SemSearch.Router_Modality_Foot, timeLimit, null, null, null); // , log, true, keepViews	    
	} else if (useCase.contains("VIA")) {
	    s = new SemSearchVIA(query, from, SemSearch.Router_Modality_Foot, timeLimit, null, log, true, null, null);	 
	} else if (useCase.contains("LDM")) {	     
	    s = new SemSearchLDM(query, from, SemSearch.Router_Modality_Foot, timeLimit, null, null, null);	 
	    
	} else if (useCase.contains("MEET")) {
	    s = null;
	} else if (useCase.contains("GEN_KW") ) {
	    KeywordGenerator gen = null;
	    try {
		gen = new KeywordGenerator();
		gen.initKeywords();
	    } catch (Exception e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	    }
	    
	    
	} else if (useCase.contains("TEST_KW")) {
	    int len = Integer.parseInt(Util.getValueFromArgs(args, "l"));
	    testKeywords(len);
	}
	
	if(s != null) {
        	try {
        	    s.setStartCal(start);
        	    s.evalQuery(useKeyword);  // Toogle Query /Keywords
        	} catch (Exception e) {
        	    e.printStackTrace();
        }
	
	}
	
	long milliseconds = (System.nanoTime() - start) / 1000000;

	// End time
	Calendar cal2 = Calendar.getInstance();
	System.out.println("End: " + cal2.getTime().toString()); // + "  Total time:; " + milliseconds);

	
    }
  
    private static void testKeywords(int len) {

	KeywordGenerator gen = null;
	String inputString = null;

	try {
	    gen = new KeywordGenerator();
	} catch (Exception e) {
	    e.printStackTrace();
	}

	// Scanner keys = new Scanner(System.in);
	System.out.println("Enter Keyword:");
	
	BufferedReader br = new BufferedReader(new InputStreamReader(System.in));

	// Read the username from the command-line; need to use try/catch with
	try {
	    inputString = br.readLine();
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	}
	
	String[] inputs = inputString.split(" ");
	KeywordResults kwds = gen.getKeywords(inputs, len, true);
	
	for (int i = 0; i < kwds.size(); i++) {
	    String searchStr = kwds.getSearchString(i);
	    
	    System.out.println("Keywords for " + searchStr + ":");
	    ArrayList<String>  kws = kwds.getKeywords(i); // String[]
	    for (int j = 0; j < kws.size(); j++) {
		System.out.println( WordUtils.capitalize(kws.get(j)));// if(kws[j]!=null)   kws[j]
	    }
	    System.out.println();
	}

	System.out.println("Keywords combinations: ");
	// gen.combineKeywords(kwds);
	String[] combs = kwds.getCombinations();
	for (int i = 0; i < combs.length; i++) {
	    System.out.println( combs[i]);
	}

	
	
    }
    
    
}
