package edu.tuwien.kbs.db;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import com.vividsolutions.jts.geom.Geometry;

import edu.tuwien.kbs.geo.GeoUtil;
import edu.tuwien.kbs.routing.*;

/**
 * <p>
 * Title: Results of a query evaluation
 * </p>
 * <p>
 * Description:
 * </p>
 * 
 * @author Patrik Schneider
 */

public class QueryResultGIS implements IQueryResult {

    protected ArrayList<PointOfInterest> mResultsPoi;// ArrayList<String[]>
    protected ArrayList<AreaOfInterest> mResultsAoi;// ArrayList<String[]>
    protected int mColumns;
    protected String mName;
    protected int mGeoID;    
    protected HashMap<String, String> mExistingPois;
    
    // Have to match SQL in QueryRewritter
    protected static int Postgres_Result_PosId = 0;
    protected static int Postgres_Result_PosName = 1;
    protected static int Postgres_Result_PosDistance = 2;
    protected static int Postgres_Result_PosX = 3;
    protected static int Postgres_Result_PosY = 4;
    protected static int Postgres_Result_PosGeometry = 5;
    protected static int Postgres_Result_PosSource = 6;
    protected static int Postgres_Result_PosPostCode = 7;
    protected static int Postgres_Result_PosPostCity = 8;
    protected static int Postgres_Result_PosPostStreet = 9;
    protected static int Postgres_Result_PosPostHNr = 10;
    protected static int Postgres_Result_PosUrl = 11;
    protected static int Postgres_Result_PosEmail = 12;
    protected static int Postgres_Result_PosPhone = 13;
    protected static int Postgres_Result_PosDesc = 14;
    protected static int Postgres_Result_PosOpen = 15;
       
    

    public QueryResultGIS(String name, int geoID)  {  // ,  int posID, int posName, int posCoorX, int posCoorY

	mResultsPoi = new ArrayList<PointOfInterest>(); // new ArrayList<String[]>();		
	mResultsAoi = new ArrayList<AreaOfInterest>();
	mName = name;
	mExistingPois = new HashMap<String, String>();

    }

    public int getPosCoorX() {
	return Postgres_Result_PosX;
    }
    
    public int getPosCoorY() {
	return Postgres_Result_PosY;
    }
    
    public int getColumns() {
	return mColumns;
    }
    
    public int getSize() {
	return mResultsPoi.size() + mResultsAoi.size() ;
    }
    
//    public ArrayList<String[]>  getResults() {
//	return mResults;
//    }

    public void fill(ResultSet rset, int columnCount, boolean isPOI) throws SQLException {
	
	mColumns = columnCount;

	
	// Fetch each row from the result set and move it to ArrayList
        while (rset.next()) {
        
            String id = rset.getString("id");
            String name = rset.getString("name");
            String posx = rset.getString("posx");
            String posy = rset.getString("posy");
            
            if(isPOI) {
        	PointOfInterest poi = new PointOfInterest(id, name, posx, posy, mGeoID);
        	mResultsPoi.add(poi);
        	
		// Direction information
		if (columnCount > Postgres_Result_PosSource)
		    poi.Source = rset.getString(Postgres_Result_PosSource + 1);
		if (columnCount > Postgres_Result_PosPostCode)
		    poi.PostCode = rset.getString(Postgres_Result_PosPostCode + 1);
		if (columnCount > Postgres_Result_PosPostCity)
		    poi.City = rset.getString(Postgres_Result_PosPostCity + 1);
		if (columnCount > Postgres_Result_PosPostStreet)
		    poi.Street = rset.getString(Postgres_Result_PosPostStreet + 1);
		if (columnCount > Postgres_Result_PosPostHNr)
		    poi.HouseNr = rset.getString(Postgres_Result_PosPostHNr + 1);
		if (columnCount > Postgres_Result_PosUrl)
		    poi.Url = rset.getString(Postgres_Result_PosUrl + 1);
		if (columnCount > Postgres_Result_PosEmail)
		    poi.Email = rset.getString(Postgres_Result_PosEmail + 1);
		if (columnCount > Postgres_Result_PosPhone)
		    poi.Phone = rset.getString(Postgres_Result_PosPhone + 1);
		if (columnCount > Postgres_Result_PosDesc)
		    poi.Description = rset.getString(Postgres_Result_PosDesc + 1);
		if (columnCount > Postgres_Result_PosOpen)
		    poi.OpeningTimes = rset.getString(Postgres_Result_PosOpen + 1);
       	
            }
            else {
    	    	AreaOfInterest areaoi = new AreaOfInterest(id, name, posx, posy, mGeoID);  
    	        areaoi.Area = GeoUtil.wktToPoly(rset.getString(Postgres_Result_PosGeometry + 1));
    	        mResultsAoi.add(areaoi);
            }
        
        }
}
    
    public ArrayList<AreaOfInterest>  selectAreaOfInts(boolean doubleCheckContainment, Geometry[] isochBuffer) { 
    
	ArrayList<AreaOfInterest> areasoi = new ArrayList<AreaOfInterest>();
	
	for (int j = 0; j < mResultsAoi.size(); j++) {

//	    String[] vals = mResults.get(j);	    

	    AreaOfInterest areaoi = (AreaOfInterest) mResultsAoi.get(j);
	    String id = areaoi.Id; //  vals[Postgres_Result_PosId];
	    
//	    AreaOfInterest areaoi = new AreaOfInterest(vals[Postgres_Result_PosId], vals[Postgres_Result_PosName], 
//		    vals[Postgres_Result_PosX], vals[Postgres_Result_PosY], mGeoID);
	    
	    // Needed for spatial predicates
//	    areaoi.Area = GeoUtil.wkbToPoly(vals[Postgres_Result_PosGeometry]);
	    
	    // Eliminate duplicated (through ID)
	    if(mExistingPois.containsKey(id)) continue;
	    
	    // Important to check again, if POI is really in IsochroneBuffer, because there are false positives form the Spatial Query	    
	    if(doubleCheckContainment && !checkContainment(isochBuffer, areaoi.getPoint())) continue;

	    areasoi.add(areaoi);
	    mExistingPois.put(id, "");

	}
	
	return areasoi;

    }
    
    public ArrayList<PointOfInterest>  selectPOIs(boolean doubleCheckContainment, Geometry[] isochBuffer) { //  String spatialRelation, QueryResult queryResult2
	
	ArrayList<PointOfInterest> pois = new ArrayList<PointOfInterest>();
	
	for (int j = 0; j < mResultsPoi.size(); j++) {
	    
	    // String[] vals = mResults.get(j);
	    PointOfInterest poi = (PointOfInterest) mResultsPoi.get(j);
	    String id = poi.Id; // vals[Postgres_Result_PosId];

    
	    // Eliminate duplicated (through ID)
	    if(mExistingPois.containsKey(id)) continue;
	    
	    // Important to check again, if POI is really in IsochroneBuffer, because there are false positives form the Spatial Query	    
	    if(doubleCheckContainment && !checkContainment(isochBuffer, poi.getPoint())) continue;

	    pois.add(poi);
	    mExistingPois.put(id, "");
	    
	}

	return pois;
	
    }    
    
    protected boolean checkContainment(Geometry[] isochBuffer, Geometry poi) {
	boolean contained = false;
	for (int i = 0; i < isochBuffer.length; i++) {
	    if( isochBuffer[i].contains(poi))  {
		contained = true;		
		break;
	    }
	}			
	return contained;
    }
    
    
    /*
     * (non-Javadoc)
     * 
     * @see edu.tuwien.ps.db.IQueryResult#printOut(boolean)
     */
    @Override
    public void printOut(boolean silent) {

	System.out.println("Result for: " + mName);

	if (silent) {
	    // Print count
	    System.out.println("Results: " + mResultsPoi.size());
	} else {
	    for (int i = 0; i < mResultsPoi.size(); i++) {

		FeatureOfInterest foi = (FeatureOfInterest) mResultsPoi.get(i);
		System.out.println(foi.toString());
	    }
	}

    }

}
