package edu.tuwien.kbs.dl;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map.Entry;
import java.util.Set;

import org.javatuples.Pair;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.io.*;
import org.semanticweb.owlapi.model.*;

import edu.tuwien.kbs.search.SemSearch;

public class DLKBLoader {

    protected DLKB mKB;
    protected boolean mIsConsistent = true;
    protected IRI mLabelIRI;
    protected OWLOntology mOntoTBox;
    // protected OWLOntology mOntoABox;
    OWLOntologyManager mManager;
    protected static String LabelIriString = "http://www.w3.org/2000/01/rdf-schema#label";
    
    public DLKBLoader() {

	mLabelIRI = IRI.create(LabelIriString);
	mManager = OWLManager.createOWLOntologyManager();

	//File rootDir = new File(System.getProperty( "user.dir" ) + "/data/" );
	//AutoIRIMapper autoIRIMapper = new AutoIRIMapper(rootDir, false);
	//autoIRIMapper.update();
	//Set<IRI> iris =  autoIRIMapper.getOntologyIRIs();	
	//mManager.addIRIMapper(autoIRIMapper);

	    //IRI iIRI = IRI.create("file://Users/patrik/ITS/myits-svn/dev-java/SemSearch/data/GeoConceptsKBS_TBox.owl");
	    //OWLImportsDeclarationImpl imp = new OWLImportsDeclarationImpl(iIRI);
	    //mManager.makeLoadImportRequest(imp);
    }

    public DLKBLoader(DLKB kb) {
	this();	
	mKB = kb;
    }

/*    private void printOntologyAndImports(OWLOntology ontology) {

	printOntology(ontology);
	
	// List the imported ontologies
	for(OWLOntology importedOntology : ontology.getImports()) {
	    System.out.println("Imports:");
	    printOntology(importedOntology);
	}
    } 
 
    private void printOntology(OWLOntology ontology) {
	IRI ontologyIRI = ontology.getOntologyID().getOntologyIRI();
	IRI documentIRI = mManager.getOntologyDocumentIRI(ontology);
	System.out.println(ontologyIRI.toQuotedString());
	System.out.println(" from " + documentIRI.toQuotedString());
    }	*/

    public void loadKB(String uriTBox) { // loadTBox

	if(mOntoTBox!=null) return;
	
	try {

	    File fileT = new File(uriTBox);
	    mOntoTBox = mManager.loadOntologyFromOntologyDocument(fileT);
	    
	    // printOntologyAndImports(mOntoTBox);

	} catch (OWLOntologyCreationIOException e) {
	    e.printStackTrace();
	} catch (UnparsableOntologyException e) {
	    e.printStackTrace();
	} catch (UnloadableImportException e) {
	    e.printStackTrace();
	} catch (OWLOntologyCreationException e) {
	    e.printStackTrace();
	}	    
	    
    }    
 
    public Set<OWLNamedIndividual> getIndividuals() {
	
	return mOntoTBox.getIndividualsInSignature(); // mOntoABox
    }    
    
    public ArrayList<Pair<String,String>> getConceptAssertions() {
	
	ArrayList<Pair<String,String>> results = new ArrayList<Pair<String,String>>();

	
	for (OWLClassAssertionAxiom assertion : mOntoTBox.getAxioms(AxiomType.CLASS_ASSERTION,true)) {
	    //OWLClass  cls = assertion.getClassExpression().asOWLClass();;
	    OWLNamedIndividual  indiv = assertion.getIndividual().asOWLNamedIndividual();
	    
	    String conceptUri = getShortUriForConcept(assertion.getClassExpression()); // SemSearch.DefaultNamespacePrefixTBox + ":" + cls.getIRI().getFragment();
	    String indUri = SemSearch.DefaultNamespacePrefixABox + ":" + indiv.getIRI().getFragment();
	    
	    System.out.println("C: " + conceptUri + " I: " + indUri);
	    results.add(new Pair<String,String>(indUri, conceptUri));

	}
	
	return  results;
	
    }  
    
    public Set<OWLClass> getTBoxConcepts() {
	
	return mOntoTBox.getClassesInSignature();
    }
    
    public Set<OWLObjectProperty> getTBoxObjectRoles() {
	
	return mOntoTBox.getObjectPropertiesInSignature();
    }
    
    public ArrayList<String> getDomainsOfObjectRole(OWLObjectProperty role) {
	
	ArrayList<String> returns = new ArrayList<String>();
	
	Set<OWLClassExpression> domains = role.getDomains(mOntoTBox);
		
	for (OWLClassExpression concept : domains) {
	    if(! concept.isAnonymous()) {
		String conceptUri = getShortUriForConcept(concept); // SemSearch.DefaultNamespacePrefixTBox + ":" + rangeConcept.getIRI().getFragment();	
		returns.add(conceptUri);
	    }
	}
	
	return returns;
    }
    
    public ArrayList<Pair<String,Integer>> getRangesOfObjectRole(OWLObjectProperty role) {
	
	ArrayList<Pair<String,Integer>> returns = new ArrayList<Pair<String,Integer>>();
	
	Set<OWLClassExpression> domains = role.getRanges(mOntoTBox);
		
	for (OWLClassExpression concept : domains) {
	    if(! concept.isAnonymous()) {
		OWLClass conceptTemp = concept.asOWLClass();		
		String conceptUri = getShortUriForConcept(conceptTemp); // SemSearch.DefaultNamespacePrefixTBox + ":" + rangeConcept.getIRI().getFragment();	
		Pair<String,Integer> firstConcept = new Pair<String, Integer>(conceptUri, 0);
		returns.add(firstConcept);
		
		// Also add subconcepts
		HashMap<String,Integer> subConcepts = new HashMap<String,Integer>();
		getConnectedSubConceptUris(subConcepts, conceptTemp, 0);
		for (Entry<String, Integer> value : subConcepts.entrySet()) {
			Pair<String,Integer> subConcept  = new Pair<String, Integer>(value.getKey(), value.getValue());
			returns.add(subConcept);
		}

		
	    }
	}
	
	return returns;
    }
    
   
    public void getSuperConceptUris(HashMap<String,String> results, OWLClass startConcept) { // OWLClass

	Set<OWLClassExpression> superConcepts = startConcept.getSuperClasses(mOntoTBox);
	for (OWLClassExpression superCo : superConcepts) {
	    if(! superCo.isAnonymous()) {
		String conceptUri = getShortUriForConcept(superCo); // SemSearch.DefaultNamespacePrefixTBox + ":" + rangeConcept.getIRI().getFragment();		
		results.put(conceptUri, null);
		getSuperConceptUris(results, superCo.asOWLClass());
		
	    }
	}
    }    
    
    
   public String getShortUriForConcept(OWLClassExpression concept) {

       // Note very elegant to use the Default Namespace, but difficult to get prefixes directly from OWLOntology
	OWLClass conceptTemp = concept.asOWLClass();
	return SemSearch.DefaultNamespacePrefixTBox + ":" + conceptTemp.getIRI().getFragment();		
   
    }

   public String getShortUriForConcept(OWLClass concept) {	
	return SemSearch.DefaultNamespacePrefixTBox + ":" + concept.getIRI().getFragment();		  
   }

   public String getShortUriForObjectRole(OWLObjectProperty objRole) {	
	return SemSearch.DefaultNamespacePrefixTBox + ":" + objRole.getIRI().getFragment();		 
   }

   public String getShortUriForDataRole(OWLDataProperty dataRole) {	
	return SemSearch.DefaultNamespacePrefixTBox + ":" + dataRole.getIRI().getFragment();		 
  }   

   public String getShortUriForIndividual(OWLNamedIndividual namedIndiv) {	
	return SemSearch.DefaultNamespacePrefixABox + ":" + namedIndiv.getIRI().getFragment();		 
   }   
   
   
    
    public void getConnectedConceptToSuperConceptUris(HashMap<String,Integer> results, OWLClass startConcept) { // OWLClass

	Set<OWLClassExpression> superConcepts = startConcept.getSuperClasses(mOntoTBox);
	for (OWLClassExpression superCo : superConcepts) {
	    if(! superCo.isAnonymous()) {
		getConnectedConceptUris(results, superCo.asOWLClass());	    
		getConnectedConceptToSuperConceptUris(results, superCo.asOWLClass());
	    }
	}
    }
    
    
    
    public void getConnectedConceptUris(HashMap<String, Integer> results, OWLClass startConcept) { // OWLClass

	// ArrayList<String> results = new ArrayList<String>();

	for (OWLObjectProperty prt : mOntoTBox.getObjectPropertiesInSignature()) {

	    Set<OWLClassExpression> doms = prt.getDomains(mOntoTBox);
	    Set<OWLClassExpression> ranges = prt.getRanges(mOntoTBox);
	    
	    if (doms.contains(startConcept)) {
		for (OWLClassExpression rcls : ranges) {
		    // Just consider "real" concepts no composites
		    if (!rcls.isAnonymous()) {
			OWLClass rangeConcept = rcls.asOWLClass();
			// Ignore recursive Object roles (domain and range are equal). This would lead to too many combinations
			if (! rcls.equals(startConcept)) {
			    String conceptUri = getShortUriForConcept(rangeConcept); // SemSearch.DefaultNamespacePrefixTBox + ":" + rangeConcept.getIRI().getFragment();
			    // Check to avoid cycles
			    if (!results.containsKey(conceptUri)) {
				// Assign concept to be reached
				results.put(conceptUri, 0);
				// Search ObjectRole Graph for further concepts
				getConnectedConceptUris(results, rangeConcept);
				// Subclasses are also reachable
				getConnectedSubConceptUris(results, rangeConcept, 0);
			    }
			}
		    }
		}
	    }
	}

	// return results;
    }
    
    public void getConnectedSubConceptUris(HashMap<String,Integer> results, OWLClass startConcept, int level) {
	
	Set<OWLClassExpression> subConcepts = startConcept.getSubClasses(mOntoTBox);
	level++;
	
	for (OWLClassExpression subCls : subConcepts) {
	    OWLClass subConcept = subCls.asOWLClass();
	    String subConceptUri = getShortUriForConcept(subCls); // SemSearch.DefaultNamespacePrefixTBox + ":" + subConcept.getIRI().getFragment();

	    if(! results.containsKey(subConceptUri)) {
		// Assign concept to be reached
		results.put(subConceptUri,level);
		
		// Search ObjectRole Graph for further concepts
		// getConnectedConceptUris(results, subConcept); PS: Maybe better, but for now use just use Subconcepts
		getConnectedSubConceptUris(results, subConcept,level );
	    }
	    
	}
	
    }
    
    
    public ArrayList<Pair<String,String>> getTBoxConceptsRoleLabels() {

	// IRI labelIRI = IRI.create(LabelIriString);
	
	ArrayList<Pair<String,String>> results = new ArrayList<Pair<String,String>>();


	    // Read concepts
	    for (OWLClass cls : mOntoTBox.getClassesInSignature()) {
		
		// Get URI
		String conceptUri = getShortUriForConcept(cls); // SemSearch.DefaultNamespacePrefixTBox + ":" + cls.getIRI().getFragment(); // Note very elegant to use the Default Namespace, but difficult to get prefixes directly from OWLOntology
		// Get label (for now in german)
		ArrayList<String> indValues = GetLabelValue(cls, mOntoTBox);
		
		// Store in results
		for (String indValue : indValues) {
		    results.add(new Pair<String,String>(conceptUri, indValue));
		    System.out.println("C: " + conceptUri + " " + indValue);
		    
		}
				
		
	    }
	    
	    // Read object roles
	    for (OWLObjectProperty cls : mOntoTBox.getObjectPropertiesInSignature()) {
		
		// Get URI
		String indUri = getShortUriForObjectRole(cls); // SemSearch.DefaultNamespacePrefixTBox + ":" + cls.getIRI().getFragment(); 
		// Get label (for now in german)
		ArrayList<String> indValues = GetLabelValue(cls, mOntoTBox);
		
		// Store in results
		for (String indValue : indValues) {
		    results.add(new Pair<String,String>(indUri, indValue));
		    System.out.println("OR: " + indUri + " " + indValue);
		}
	    }
	    
	    // Read data roles
	    for (OWLDataProperty cls : mOntoTBox.getDataPropertiesInSignature()) {
		
		// Get URI
		String indUri = getShortUriForDataRole(cls); // SemSearch.DefaultNamespacePrefixTBox + ":" + cls.getIRI().getFragment(); // Note very elegant to use the Default Namespace, but difficult to get prefixes directly from OWLOntology
		// Get label (for now in german)
		ArrayList<String> indValues = GetLabelValue(cls, mOntoTBox);
		
		// Store in results
		for (String indValue : indValues) {
		    results.add(new Pair<String,String>(indUri, indValue));
		    System.out.println("DR: " + indUri + " " + indValue);
		}
	    }
	    
	return results;	
	
    }    
    
    public ArrayList<Pair<String,String>> getABoxIndividualLabels() {

	
	ArrayList<Pair<String,String>> results = new ArrayList<Pair<String,String>>();
	

	    // Read indiviudals
	    for (OWLNamedIndividual cls : mOntoTBox.getIndividualsInSignature()) { // mOntoABox
				
		// Get URI
		String indUri = getShortUriForIndividual(cls); // SemSearch.DefaultNamespacePrefixABox + ":" + cls.getIRI().getFragment(); // Note very elegant to use the Default Namespace, but difficult to get prefixes directly from OWLOntology
		// System.out.println(indUri);
		
		// Get label (for now in german)
		ArrayList<String> indValues = GetLabelValue(cls, mOntoTBox); // mOntoABox
		for (String indValue : indValues) {
		    // Store in results
		    Pair<String,String> p = new Pair<String,String>(indUri, indValue);
		    results.add(p);
		    
		}
		

	    }

	return results;

    }
    
    protected ArrayList<String> GetLabelValue(OWLEntity entity, OWLOntology ontology) {
	
	ArrayList<String> vals = new ArrayList<String>();
	
	Set<OWLAnnotation> notations = entity.getAnnotations(ontology);
	for (OWLAnnotation no : notations) {
	    if(no.getProperty().getIRI().equals(mLabelIRI)) {
		if (no.getValue() instanceof OWLLiteral) { 
		    OWLLiteral indLit = (OWLLiteral) no.getValue();
		    vals.add(indLit.getLiteral());
		}
	    }
	}
	
	return vals;
    }

}
