package edu.tuwien.kbs.geo;


import org.apache.lucene.spatial.DistanceUtils;
import org.javatuples.Pair;


import com.vividsolutions.jts.algorithm.ConvexHull;
import com.vividsolutions.jts.geom.*;
import com.vividsolutions.jts.io.ParseException;
import com.vividsolutions.jts.io.*;
import com.vividsolutions.jts.operation.buffer.BufferOp;



public class GeoUtil {

    public static double EARTH_RADIUS = 6372;
    public static String GML_Namespace = "http://www.opengis.net/gml/3.2";

    protected static PrecisionModel precisionModel = new PrecisionModel(
	    PrecisionModel.FLOATING);
    protected static GeometryFactory geometryFactory = new GeometryFactory(
	    precisionModel);
    protected static SimpleGMLWriter simpleGMLWriter = new SimpleGMLWriter();

    protected static int BUFFER_Quadrant_Segments = 3;
    protected static double BUFFER_Distance_Iso = 0.00029;


    public static String writeToWKB(Geometry[] isochBuffer, boolean simply) {
	    
	GeometryCollection col = new GeometryCollection(isochBuffer,
		geometryFactory);

	WKBWriter writer = new WKBWriter(2);
	byte[] retBytes = new byte[0];
	
	if(simply) {	
	    ConvexHull cv = new ConvexHull(col);
	    retBytes = writer.write(cv.getConvexHull());
	}
	else {
	    retBytes = writer.write(col);
	}

	return WKBWriter.toHex(retBytes);

    }
    
    public static Point wktToPoint(String wellKnownText) {

	WKTReader reader = new WKTReader(geometryFactory);	
	Point result = null;
	try {
	    Geometry geo = reader.read(wellKnownText);
	    if (geo instanceof Point) result = (Point) geo;
	    
	} catch (ParseException e) {

	    // e.printStackTrace();
	}
	
	return result;
	
    }
    
    public static LineString wktToLineString(String wellKnownText) {

	WKTReader reader = new WKTReader(geometryFactory);	
	LineString result = null;
	try {
	    Geometry geo = reader.read(wellKnownText);
	    if (geo instanceof LineString) result = (LineString) geo;
	    
	} catch (ParseException e) {

	    // e.printStackTrace();
	}
	
	return result;
	
    }
    
    public static Polygon wktToPoly(String wellKnownText) {

	WKTReader reader = new WKTReader(geometryFactory);	
	Polygon pol = null;
	try {
	    Geometry geo = reader.read(wellKnownText);
	    if (geo instanceof Polygon) pol = (Polygon) geo;
	    
	} catch (ParseException e) {

	    // e.printStackTrace();
	}
	
	return pol;
	
    }
    
    public static LineString coordinatesToLineString(String coordinates) {
	
	// StringTokenizer tokz = StringTokenizer(coordVal, "; ");
	String[] xml2 = coordinates.split("[ ]");
	Coordinate[] coords = new Coordinate[xml2.length];

	for (int j = 0; j < coords.length; j++) {
	    String[] ps = xml2[j].split("[,]");
	    coords[j] = new Coordinate(Double.valueOf(ps[0]), Double.valueOf(ps[1]));
	}

	return geometryFactory.createLineString(coords);
    
    }
    
    
    public static String pointToCoord(Point p, boolean switchXY, char delimiter) {
	
	StringBuffer coorString = new StringBuffer();
	
	if(! switchXY) {
	    coorString.append(p.getX()); // Double.toString(p.getX())
	    coorString.append(delimiter);
	    coorString.append(p.getY());
	}
	else {
	    coorString.append(p.getY()); // Double.toString(p.getY())
	    coorString.append(delimiter);
	    coorString.append(p.getX());
	}
	
	return coorString.toString();
    }
    
/*
    public static Point createPoint(String text, int geoID, boolean switchXY) { 
	
	String[] coor = text.split(" ");
	
	return createPoint(coor[0], coor[1], geoID, switchXY);
    }
    */
    
    public static Point createPoint(String x, String y, int geoID, boolean switchXY) { 

	double dx = Double.valueOf(x);
	double dy = Double.valueOf(y);
	
	Coordinate co = null;
	if(switchXY)
	    co = new Coordinate(dy, dx);
	else
	    co = new Coordinate(dx, dy);
	
	return geometryFactory.createPoint(co);

    }

    public static Pair<Geometry, Point> createBox2DForIso(Point point_a, Point point_b, double buffer_Distance_Box,  int geoID,
	    boolean log) {

	
	// For GML, switch x and y, hence y is first!!!
	// Point temp_a =  geometryFactory.createPoint(new Coordinate(point_a.getY(), point_a.getX()));
	// Point temp_b =  geometryFactory.createPoint(new Coordinate(point_b.getY(), point_b.getX()));

	Point[] points = new Point[] { point_a, point_b };
	Geometry g1 = new GeometryFactory().createMultiPoint(points);
	
	// First create envelope of the points
	Geometry g2 = g1.getEnvelope();
	
	Point centroid = g2.getCentroid();
	
	// And then blow up
	Geometry g3 = g2.buffer(buffer_Distance_Box, BUFFER_Quadrant_Segments, BufferOp.CAP_SQUARE); // BufferOp.CAP_ROUND

	if (log) 
	    simpleGMLWriter.writeGeometry(new Geometry[] {g3}, "log/bbox2d_ViA.gml");

	// More efficient to use WKB instead of WKT
	// WKBWriter writer = new WKBWriter(2);
	// byte[] retBytes = writer.write(g3);
	
	return new Pair<Geometry, Point>(g3, centroid); // WKBWriter.toHex(retBytes)
	
	
    }
    
    public static String createBox2DForIso(Point point_a, int geoID, double distance,
	    boolean log) {

	
	// For GML, switch x and y, hence y is first!!!
	double y = point_a.getY();
	double x = point_a.getX();

	Point pMin = GeoUtil.calcDistance(x, y, distance, DistanceUtils.DEG_45_AS_RADS,
		geoID); // 315
	Point pMax = GeoUtil.calcDistance(x, y, (-1) * distance,
		DistanceUtils.DEG_45_AS_RADS, geoID); // 135

	Point[] points = new Point[] { pMin, pMax };
	Geometry g1 = new GeometryFactory().createMultiPoint(points);
	Geometry g2 = g1.getEnvelope();

	if (log) simpleGMLWriter.writeGeometry(new Geometry[] { g2 }, "log/box2d_NR.gml");
	
	// return g2.toText();
	
	// More efficient to use WKB instead of WKT
	WKBWriter writer = new WKBWriter(2);
	byte[] retBytes = writer.write(g2);
	return WKBWriter.toHex(retBytes);

	

    }


    
    public static Point calcDistance(double lat, double lon, double distance,
	    double bearing, int geoID) {


	//double brng = Math.toRadians(bearing);
	double lat1 = Math.toRadians(lat);
	double lon1 = Math.toRadians(lon);

	double[] result = new double[2];
	result = DistanceUtils.pointOnBearing(lat1, lon1, distance, bearing,
		result, EARTH_RADIUS);

	// double d = distance / EARTH_RADIUS; // * 0.9 Substract 10% because
	// distance through BOX3D is not very accurate

	// double lat2 = Math.asin( Math.sin(lat1) * Math.cos(d) +
	// Math.cos(lat1) * Math.sin(d) * Math.cos(brng) );
	// double lon2 = lon1 + Math.atan2((Math.sin(brng) * Math.sin(d) *
	// Math.cos(lat1)), (Math.cos(d) - Math.sin(lat1) * Math.sin(lat2)));
	// lon2 = (lon2 + 3 * Math.PI) % (2 * Math.PI) - Math.PI;

	Coordinate c = new Coordinate(Math.toDegrees(result[0]),
		Math.toDegrees(result[1]));
	CoordinateSequence coors = geometryFactory
		.getCoordinateSequenceFactory().create(new Coordinate[] { c });

	Point p = new Point(coors, geometryFactory);

	return p;

    }

}
