package edu.tuwien.kbs.keywords;

import java.util.ArrayList;
import org.javatuples.Pair;
import nu.xom.*;

public class KeywordResults {
    
    public static String mTUNamespace = "http://www.kr.tuwien.ac.at/myits/kw/1.0";

    
    // A String representing the input search string and a String array representing matched KWs
    ArrayList<Pair<String, ArrayList<String> >>  mSearchStrings; // String[]
    ArrayList<String> mCombinations;
    public boolean mIsSpatialPredicate; // Always a single SearchString
    
    public KeywordResults() {
	mSearchStrings = new ArrayList<Pair<String, ArrayList<String> >>(); // String[]
	mCombinations = new ArrayList<String>();	
    }
    
    public KeywordResults(boolean isSpatialPredicate) {
	this();
	mIsSpatialPredicate = isSpatialPredicate;
	
    }
    
    @Override 
    public String toString() {
	
	String ret = "";
	
	for (int i = 0; i < mSearchStrings.size(); i++) {
	    Pair<String, ArrayList<String> > search = mSearchStrings.get(i);
	    ArrayList<String> vals = search.getValue1();
	    if(vals.size() > 0) {
		if(ret.length() > 0) ret = ret + " ";
		ret = ret + vals.get(0);
	    }
	}
	
	return ret;
	
    }
    
    public String serializeToXml() {

	
	Element root = new Element("tuw:Results",mTUNamespace);
	
	// Put information as validity and spell corretion of entered search strings
	Element searchstrings = new Element("tuw:searchstrings",mTUNamespace);
	root.appendChild(searchstrings);
	
	for (int i = 0; i < mSearchStrings.size(); i++) {
	    Pair<String, ArrayList<String> > search = mSearchStrings.get(i); // String[]

	    Element sel = new Element("tuw:searchstring", mTUNamespace);
	    sel.appendChild(search.getValue0());
	    boolean isValid = (search.getValue1().size() > 0); // length
	    Attribute valid = new Attribute("valid", Boolean.toString(isValid)); // tuw:
	    sel.addAttribute(valid);

	    searchstrings.appendChild(sel);
	}

	// Put information of generated keywords from the search stings
	Element keywords = new Element("tuw:keywords",mTUNamespace);
	root.appendChild(keywords);	
	
	for (int i = 0; i < mCombinations.size(); i++) {	    
	    String comb = mCombinations.get(i);
	    
	    Element kw = new Element("tuw:keyword",mTUNamespace);
	    kw.appendChild(comb);

	    keywords.appendChild(kw);	    
	}
	
	Document doc = new Document(root);
	
	return doc.toXML();
    }   
    
    public void add(String searchString, ArrayList<String> kws) { // String[] kws
	mSearchStrings.add(new Pair<String, ArrayList<String> >(searchString, kws)); // String[]

    }
    
    public void addKWCombination(String comb) {
	mCombinations.add(comb);
    }
    
    public void addKWCombinations(ArrayList<String> combs) {
	mCombinations.addAll(combs);
    }
    
    public int size() {
	return mSearchStrings.size();
    }

    public boolean isSpatialPredicate() {	
	return mIsSpatialPredicate;	
    }
    
    public String getSearchString(int pos) {	
	return mSearchStrings.get(pos).getValue0();	
    }
    
    public ArrayList<String> getKeywords(int pos) {	// String[]
	return mSearchStrings.get(pos).getValue1();	
    }

    public String[] getCombinations() {
	return mCombinations.toArray(new String[mCombinations.size()]);
    }   


    public void setCombinations(ArrayList<String> newCombinations) {
	 mCombinations = newCombinations;
    }   
}
