package edu.tuwien.kbs.ldm;

import java.io.StringWriter;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.TimeUnit;

import org.apache.lucene.spatial.DistanceUtils;
import org.javatuples.Pair;
import org.javatuples.Septet;
import org.javatuples.Tuple;

import com.clarkparsia.owlgres.cli.queryRewrite;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;


import edu.tuwien.kbs.db.IStoreConnection;
import edu.tuwien.kbs.db.QueryResultGIS;
import edu.tuwien.kbs.db.StoreConnectionPostgreSQL;
import edu.tuwien.kbs.geo.GeoUtil;
import edu.tuwien.kbs.geo.SimpleGMLWriter;

import edu.tuwien.kbs.routing.AreaOfInterest;
import edu.tuwien.kbs.routing.PointOfInterest;
import edu.tuwien.kbs.search.*;
import edu.tuwien.kbs.util.Util;
import edu.tuwien.kbs.keywords.SpatialPredicatesMyITS;
import edu.tuwien.kbs.ldm.parser.*;

import com.clarkparsia.owlgres.cli.queryRewrite;

public class SemSearchLDM extends SemSearch {

    
    public static int Result_Limit = 10; // The 11 closest should be a good limit as it reduces the amount of routes calculated (before 20, 15)
    public static int DefaultTimeLimit = 900; // 15 minutes in seconds
    protected boolean doubleCheckContainment = true; // false

    protected static SimpleGMLWriter Simple_GMLWriter = new SimpleGMLWriter();
    
    // protected IStoreConnection mConnection2; 
    
    public SemSearchLDM() {
	super();

	//mStrategy = "A";
    }
    
    public SemSearchLDM(String query, String from, String mots, int timeLimit, 
	    String startTime, String userId,  Map<String, String> optionsMap) {
	this();
	
	mQuery = query; mPoints = from; mMots = mots; mTimeLimit = timeLimit; 
	mStartTime = startTime; mUserId = userId; mOptionsMap = optionsMap;

	if(mTimeLimit==0) mTimeLimit= DefaultTimeLimit;
	
    }

    
    protected void connectDB2() {
	// Connect to DB
	//mConnection2 = new StoreConnectionPostgreSQL( mHost, mDB, mUser, mPw);
	//mConnection2.connect(java.sql.Connection.TRANSACTION_READ_COMMITTED);
	
    }
    
    @Override
    public String evalQuery(boolean keywords) throws Exception  {

	int actualGeoID = 4326; // For now default WGS84 // mRewritter.GetInfoItemA().geometryID;

	SimpleRuleParser queryParser  = new SimpleRuleParser();	
	
	// Create Result and general id, which will be used to get for the route and iso details
	Random ranGen = new Random();
	String sessionID = Integer.toString(ranGen.nextInt(Integer.MAX_VALUE) );
		
	printTimeTotal("Eval LDM Started: " +  sessionID); 

	// cleanAllViews();

	// Connect and set up DB
	connectDB(false);
	connectDB2();
	
	
	// Read and parse main query	
	String queryText =  Util.getContentEnBloc(mQuery);
	Rule queryMain = queryParser.parse(queryText, 0, 0);
	
	ILiteral headAtom = queryMain.getHeadLiteral();
	ILiteral[] bodyAtoms = queryMain.getBodyLiterals();
	
	// Read and parse join tree
	ArrayList<Rule> joinTreeRules = new ArrayList<Rule>();	
	
	String[] joinTreeText =  Util.getContent(mQuery.replace(".txt", ".jtx"));
	
	for (int j = 0; j < joinTreeText.length; j++) {
	    // System.out.println("Parse: " + joinTreeText[j]);
	    if(joinTreeText[j].length()==0)
		continue;
	    joinTreeRules.add(queryParser.parse(joinTreeText[j], 0, 0));	
	}
	
	// Keep last subquery for output
	String lastQueryName = null;
	int aggCount = 1;
	Hashtable<String, QueryResultLDM> resultMemoryStore = new Hashtable<String, QueryResultLDM>();
	ArrayList<String> aggViewsDrop = new ArrayList<String>();
	
	// Two cases sub rule or main rule
	for( Rule joinRule : joinTreeRules) {
	
	    
	    ILiteral joinRuleHead = joinRule.getHeadLiteral();
	    String ruleHeadName = joinRuleHead.getPredicateName();
	    QueryResultLDM queryResult = null;
	    
	    printTimeTotal(ruleHeadName + " Subquery started"); 
    
	    // Normal Atoms
	    if(ruleHeadName.contains("_N")) {
		
		queryResult = evalNormalAtom(joinRule, sessionID, resultMemoryStore);
	        
	    }
	    // Stream Atoms
	    else if(ruleHeadName.contains("_F")) {
		
		if(joinRule.hasAggregate()) {			   
		    queryResult = evalAggregateAtom(joinRule, sessionID, aggCount, resultMemoryStore, aggViewsDrop); // bodyAtoms		    
		    aggCount++;
		}

	    }
	    // Spatial Atoms
	    else if(ruleHeadName.contains("_S")) {

		if(joinRule.getBodyLength()!=3)
		    throw new Exception("Spatial joins possible with tree atoms");

		queryResult = evalSpatialJoin(joinRule, resultMemoryStore, actualGeoID); 
		
	    }
	    // Normal join
	    else if(ruleHeadName.contains("_J")) {
		
		if(joinRule.getBodyLength()>3 || joinRule.getBodyLength()<2)
		    throw new Exception("Normal joins possible with two/tree atoms");
		
		String nameAtom1 = joinRule.getBodyLiteral(0).getPredicateName();
		QueryResultLDM resultLeft = resultMemoryStore.get(nameAtom1);
		
		if(joinRule.getBodyLength()==2) {
		    String nameAtom2 = joinRule.getBodyLiteral(1).getPredicateName();
		    QueryResultLDM resultRight = resultMemoryStore.get(nameAtom2);

		    queryResult = evalNormalJoinTwoAtoms(joinRule, resultLeft.getResults(), resultRight.getResults(), false);
		}
		else {
		    String nameAtom2 = joinRule.getBodyLiteral(1).getPredicateName();
		    QueryResultLDM resultMiddle = resultMemoryStore.get(nameAtom2);
		    String nameAtom3 = joinRule.getBodyLiteral(2).getPredicateName();
		    QueryResultLDM resultRight = resultMemoryStore.get(nameAtom3);
		    
		    queryResult = evalNormalJoinThreeAtoms(joinRule, resultLeft.getResults(), resultMiddle.getResults(), resultRight.getResults());
		    
		}
		
	    } 
	    
	    // Store to cache	    
	    resultMemoryStore.put(ruleHeadName, queryResult); 
	    lastQueryName = ruleHeadName;
	    
	    printTimeTotal(ruleHeadName + " Subquery ended. Rows: " + queryResult.getSize() ); 


	}
	
	// Get last evaluation, since this is the result
	QueryResultLDM finalResult = resultMemoryStore.get(lastQueryName);

	printTimeTotal("Query Result Rows: " + finalResult.mResults.size()); 
	printTimeTotalExt("Eval LDM Ended in: ");
	

	// Clean aggregate views up and close DB	
	for(String aggViewName : aggViewsDrop) {
	    
	    try {
		  mConnection.executeStatement("DROP CONTINUOUS VIEW " + aggViewName);		
	    }
	    catch (RuntimeException e) {  
		System.out.println("Drop View Exception: " + e.toString());
	    }	
	    
	}
	
	
	
	//closeDB(sessionID);
        mConnection.close();
        //mConnection2.close();
        
	// Print results
	if (mOuputResults && mLog)   {	    
	    // TODO Print details    
	}
	
	return "";

    }

    private QueryResultLDM evalAggregateAtom(Rule aggregateRule, String sessionID, int aggCount, 
	    Hashtable<String, QueryResultLDM> interResults, ArrayList<String> aggViewsDrop)  {
	
	QueryResultLDM  queryResult = null;
	ILiteral headAtom  = null;
	
	ILiteral[] bodyAtoms = aggregateRule.getBodyLiterals();
	headAtom = aggregateRule.getHeadLiteral();
	
	ILiteral aggregateAtom = null; // This algorithm works only with a single aggregate, multiple would be a good optimization
	ArrayList<ILiteral> normalAtoms = new ArrayList<ILiteral>();
	ArrayList<ILiteral> buildInLiterals = new ArrayList<ILiteral>(); 	
	ILiteral cachedAtom = null; // TODO: Extend to array
	
	String aggViewName = "";
	String normalViewName = "";
	
	
	// Split into aggregate, internal and normal Part, so we can evaluate a pure Onto CQ and a single aggregate
	for( ILiteral lit : bodyAtoms) {
	    
	    if(lit.IsAggregate()) {
		aggregateAtom = lit;
	    }
	    else {
		
		// some of the normal atoms are alrady saved, get them from cache
		    if (lit instanceof BuiltInLiteral) {
			buildInLiterals.add(lit);
		    }
		    else {
			if(interResults.containsKey(lit.getPredicateName())) {
			    cachedAtom = lit;
			}
			else {
			    normalAtoms.add(lit);
			}
		    }
	    }	

	}

	// try {
	
	// Create view name
	String aggName = aggregateAtom.getPredicate().getName() + aggCount;
	
	// Step 1
	//  Create and eval aggregate query
	try {
	    aggViewName = createAggregateAtomView(aggregateAtom, sessionID, aggCount);
	} catch (Exception e) {
	    e.printStackTrace();
	}
	
	// Step 2	
	// Create normal query
        normalViewName = evalNormalAtoms(normalAtoms.toArray(new ILiteral[0]), sessionID, aggName + "_N");

	if(aggCount > 1) { // Even
		
	    /// Delay needed, so PipelineDB can catch up..	    
	    try {
		printTimeTotal("190ms delay after cont. view creation");
		TimeUnit.MILLISECONDS.sleep(190);
            
	    } catch (InterruptedException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	    }
	}
        
        printTimeTotal("Stream Query Rewritten");
        
	// Step 3
	ITerm term1 = aggregateAtom.getTerm(0); // First variable
	ITerm term2 = aggregateAtom.getTerm(1); // Second variable
	
	String createTableName = "T_" + sessionID + "_" + aggName;
	String createTableVariables = "iid_" + term1.toString() +" integer, " + "att_" + term1.toString() + " text, " 
		+ "att_" + term2.toString() +  " text"; // " text" integer
		
	// Create temporary table for inserts
	//String createTable = "CREATE TABLE "  + createTableName + "(" + createTableVariables +  ")";
	//mConnection.executeStatement(createTable);
	
	//String mergeSQL =  Util.getContentEnBloc("templates/query_ldm_merge_tmpl.rq");
	//String intserSQL =  "INSERT INTO %1$s (SELECT DISTINCT %2$s FROM %3$s, %4$s WHERE %5$s)";
	String mergeSQL =  "SELECT DISTINCT %2$s FROM %3$s, %4$s WHERE %5$s";

	// term1 are the indiviudals, term2 is the aggregate
	String selectVariables =   "ro.iid_" + term1.toString() + ", ro.att_" + term1.toString() + ", rf.att_" + term2.toString();
	
	// Add missing head variables to select variables
	for(ITerm termHead : headAtom.getTerms()) {
	    if(!termHead.toString().contentEquals(term1.toString()) && !termHead.toString().contentEquals(term2.toString())) {
		// Assume that head variables are ontolog atoms (rf.idd) not stream atoms (rf.add)
		selectVariables = selectVariables + ", ro.iid_" + termHead.toString() + ", ro.att_" + termHead.toString();
	    }
	}
	
	// Merge SQL parameter
	Object[] args1 = new Object[5];
	args1[0] = createTableName;
	args1[1] = selectVariables;
	args1[2] = aggViewName + " AS rf";
	args1[3] = normalViewName + " AS ro";
	args1[4] =  "rf.iid_" + term1.toString() + " = " + "ro.iid_" + term1.toString();  // "rf.iid_x = ro.iid_x"; 
	
	// Add build-in literals
	if(buildInLiterals.size() > 0) {
	    String buildIn = "";
	    
	    for( ILiteral binl : buildInLiterals) {
		buildIn = buildIn + " AND ";
		String temp_binl = "rf.att_" + binl.getTerm(0).toString() + binl.getTerm(1).toString() + binl.getTerm(2).toString();
		buildIn = buildIn + temp_binl;
	    }
	    args1[4] = args1[4] + buildIn;
	}
	
	String mergeQuery = String.format(mergeSQL, args1);
	
	// Get Results
	//if((aggCount & 1) == 0)
	//    queryResult = getResultsDB(mergeQuery,0, true); // Even
	//else
	    queryResult = getResultsDB(mergeQuery,0); // Odd
	
	// Join with internal results (Not needed anymore, since decomposition creates separate joins)
	//if(cachedAtom != null) {
	//    QueryResultLDM results2 = interResults.get(cachedAtom.getPredicateName());
	//    queryResult = getResultsJoin(aggregateRule, queryResult.getResults(), results2.getResults(), true);
	//}
	
	// printTimeTotal("Stream Query Evaluated");	
	
	
	if(normalViewName.length() > 0)
	    mConnection.executeStatement("DROP VIEW " + normalViewName); 

	if(aggViewName.length() > 0) {
	    aggViewsDrop.add(aggViewName);
	    
	 // Drop later
	 // mConnection.executeStatement("DROP CONTINUOUS VIEW " + aggViewName); 
	}

	


	//printTime("Onto Query Created: " + unionSqls.getValue5());	
	
	return queryResult;
    
    }
    
    private QueryResultLDM evalNormalAtom(Rule normalRule, String sessionID, 
	    Hashtable<String, QueryResultLDM> interResults)  {
	
	QueryResultLDM queryResult = null;
	
	ILiteral ruleHead = normalRule.getHeadLiteral();
        String ruleHeadName = ruleHead.getPredicateName();
	ILiteral[] bodyAtoms = normalRule.getBodyLiterals();

	ArrayList<ILiteral> normalAtoms = new ArrayList<ILiteral>();
	ArrayList<ILiteral> buildInLiterals = new ArrayList<ILiteral>(); 	
	ILiteral cachedAtom = null; // TODO: Extend to array

	
	// Split into aggregate, internal and normal Part, so we can evaluate a pure Onto CQ and single aggregate
	for( ILiteral lit : bodyAtoms) {

	    
	  if (lit instanceof BuiltInLiteral) {
	      buildInLiterals.add(lit);
	  }
	  else {  
	    // some of the normal atoms are alrady saved, get them from cache
	    if(interResults.containsKey(lit.getPredicateName())) {
		cachedAtom = lit;
	    }
	    else {
		normalAtoms.add(lit);
	    }
	  }
	}
        
	// Create normal query
	ArrayList<ILiteral> tempL = new ArrayList<ILiteral>();
        String normalViewName = evalNormalAtoms(normalAtoms.toArray(new ILiteral[0]) , sessionID, ruleHeadName); 
        String subSQL = "SELECT * FROM " + normalViewName;
        
        queryResult = getResultsDB(subSQL, 0);
        
	// Join with internal results (Not needed anymore, since decomposition creates separate joins)
	//if(cachedAtom != null) {
	//    QueryResultLDM results2 = interResults.get(cachedAtom.getPredicateName());
	//    queryResult = getResultsJoin(normalRule, queryResult.getResults(), results2.getResults(), true);

	//}
        

        return queryResult;
	

    }
    
    protected QueryResultLDM evalSpatialJoin(Rule joinRule, Hashtable<String, QueryResultLDM> interResults, int geoID) // String relation, QueryResultLDM result1, QueryResultLDM result2, int geoID) 
    {

	ArrayList<Pair<ResultLDM, ResultLDM>> queryResultPair = null; // new QueryResultLDM(geoID);
	
	// Get results1 from cache
	QueryResultLDM queryResult1 = null;
	// Get results2 from cache
	QueryResultLDM queryResult2 = null;

	// Find spatial relation, and separate into two parts, left and right of spatial query
	//ArrayList<ILiteral> leftAtoms = new ArrayList<ILiteral>();
	//ArrayList<ILiteral> rightAtoms = new ArrayList<ILiteral>();
	ILiteral leftAtom = null;
	ILiteral rightAtom = null;
	ILiteral spatialAtom = null;

	ILiteral[] bodyLiterals = joinRule.getBodyLiterals();
	
	// This still need to be improved with more then 3 atoms
	leftAtom = bodyLiterals[0];
	queryResult1 = interResults.get(leftAtom.getPredicateName());
	
	if(bodyLiterals.length>1) {
	    spatialAtom = bodyLiterals[1];
	}
	if(bodyLiterals.length>2) {
	    rightAtom = bodyLiterals[2];
	    queryResult2 = interResults.get(rightAtom.getPredicateName());
	}
	
        // Filter with second spatial relation
	if (spatialAtom != null)
	{
	    	String relation = spatialAtom.getPredicate().getName();

        	if (relation.equals(SpatialPredicatesLDM.KeywordQueryNear) ) {
        	    queryResultPair = evalSpatialRelationNext2(queryResult1.getResults(), queryResult2.getResults());
        	    
        	} else if (relation.equals(SpatialPredicatesLDM.KeywordQueryWithin)) {     	    
        	    queryResultPair = evalSpatialRelationInside2(queryResult1.getResults(), queryResult2.getResults());

        	} else if (relation.equals(SpatialPredicatesLDM.KeywordQueryIntersects)) {
        	    
        	    queryResultPair = evalSpatialRelationIntersect(queryResult1.getResults(), queryResult2.getResults());
       	    
        	} else if (relation.equals(SpatialPredicatesLDM.KeywordQueryDisjoint)) {
        	    // Ignore
        	    
        	} else if (relation.equals(SpatialPredicatesLDM.KeywordQueryEquals)) {
        	    // Ignore
        	} 

	}	 
	
	// TODO: A first solution so we merge the variables which are in the head. Initial for only two variables but more are possible
	
	ILiteral headAtom = joinRule.getHeadLiteral();
	ITerm term1 = headAtom.getTerm(0); // First variable
	ITerm term2 = headAtom.getTerm(1); // Second variable

	ArrayList<ResultLDM> tempResults = mergeResultPairs(queryResultPair, term1.toString(), term2.toString());
	
	QueryResultLDM retRes = new QueryResultLDM(geoID);  // geoID
	retRes.mResults = tempResults;
	
	return retRes;
   }

    /*
    private ArrayList<ResultLDM>  mergeResultPairs(ArrayList<Pair<ResultLDM, ResultLDM>> resultPairs, ArrayList<String> variables) {

	ArrayList<ResultLDM> returnResults = new ArrayList<ResultLDM>();
	    
	for(Pair<ResultLDM, ResultLDM> resultPair : resultPairs) {

	    ResultLDM newResult = new ResultLDM();
	    returnResults.add(newResult);
	    
	    for(String varName : variables) {

		String val1 = resultPair.getValue0().getAttByVar(varName);
		newResult.setAttByVar(varName, val1);
	    }
	}
	
	return returnResults;
	
    }
    */
    
    private ArrayList<ResultLDM>  mergeResultPairs(ArrayList<Pair<ResultLDM, ResultLDM>> resultPairs, String var1, String var2) {
	
	ArrayList<ResultLDM> returnResults = new ArrayList<ResultLDM>();
	    
	for(Pair<ResultLDM, ResultLDM> resultPair : resultPairs) {
	    
	    String val1 = resultPair.getValue0().getAttByVar(var1);
	    if(val1 == null || val1.length()==0)
		val1 = resultPair.getValue1().getAttByVar(var1);
	    
	    String val2 = resultPair.getValue0().getAttByVar(var2); // getValue1
	    if(val2 == null ||val2.length()==0)
		val2 = resultPair.getValue1().getAttByVar(var2);
	    
	    ResultLDM newResult = new ResultLDM();
	    returnResults.add(newResult);

	    newResult.setAttByVar(var1, val1);
	    newResult.setAttByVar(var2, val2);
	    
	}
	
	return returnResults;
	
    }
   
    
    protected QueryResultLDM getResultsDB(String mainSQL, int geoID) { // , boolean useDB2

	 QueryResultLDM  queryResult = new QueryResultLDM(geoID); // = null;

	if (mLog) 
	    System.out.println("SQL: " + mainSQL);
	
	try {
	    
	    Statement stmt = null;
	    
	    //if(!useDB2)
	    stmt =mConnection.createStatement();
	   // else
	   //	stmt =mConnection2.createStatement();
	    
	    ResultSet rset = stmt.executeQuery(mainSQL);	    
	    queryResult.fill(rset); 
	    
	} catch (RuntimeException e) {
	    throw e;
	} catch (SQLException e) {
	    throw new RuntimeException(e.toString());
	}
	
	return queryResult;
	
   }
    
    protected QueryResultLDM evalNormalJoinThreeAtoms(Rule joinRule,  ArrayList<ResultLDM> resultsLeft,
	    ArrayList<ResultLDM> resultsMiddle, ArrayList<ResultLDM> resultsRight) {

	 ArrayList<Pair<ResultLDM, ResultLDM>> resultPairs = new ArrayList<Pair<ResultLDM, ResultLDM>>();

	 
	 // Projection variables taken from head atom (initially consider two variables)
	 ILiteral headAtom = joinRule.getHeadLiteral();
	 ITerm termProj1 = headAtom.getTerm(0); // First projection variable TODO: extend to more proj. variables
	 ITerm termProj2 = headAtom.getTerm(1); // Second proj. variable

	 ITerm termJoin1 = null; // Join variable for left and middle atom
	 ITerm termJoin2 = null; // Join variable for right and middle atom
	 
	 ILiteral[] bodyLiterals = joinRule.getBodyLiterals();
	 ILiteral leftAtom = bodyLiterals[0];
	 ILiteral middleAtom = bodyLiterals[1];
	 ILiteral rightAtom = bodyLiterals[2];    
	     
	 outerloop1: // Label for outer loop
	 for(ITerm termLeft :  leftAtom.getTerms()) {
		 
	     for(ITerm termMiddle :  middleAtom.getTerms()) {
	   
		 if(termLeft.toString().equalsIgnoreCase(termMiddle.toString())) {
		     termJoin1 = termLeft;
		     break outerloop1;
		 }
		     
	     }
	  }

	 outerloop2: // Label for outer loop
	 for(ITerm termRight :  rightAtom.getTerms()) {
		 
	     for(ITerm termMiddle :  middleAtom.getTerms()) {
	   
		 if(termRight.toString().equalsIgnoreCase(termMiddle.toString())) {
		     termJoin2 = termRight;
		     break outerloop2;
		 }
		     
	     }
	  }	 

	 Hashtable<String, ResultLDM> middleAtomMatches = new Hashtable<String, ResultLDM> ();
	  
	 // Save results of middle term in Hashtable, so we can check if left and right tuples have a match
	  for(ResultLDM tuple1 : resultsMiddle) {
	      String tKey = tuple1.getAttByVar(termJoin1.toString()) + "|" + tuple1.getAttByVar(termJoin2.toString());

	      middleAtomMatches.put(tKey, tuple1);
	      
	  } 
	  
	 // Simplest algorithm is "Nested Loop Join" (good if one table is small), TODO: try "Merge Join" or "Hash Match Join"
	 
	 for(ResultLDM tupleLeft : resultsLeft) {

	     for(ResultLDM tupleRight : resultsRight) {

		// If no join variables do a cross join
		if(termJoin1!=null && termJoin1!=null) {
		     
		    String t1v1 = tupleLeft.getAttByVar(termJoin1.toString());
		    String t2v1 = tupleRight.getAttByVar(termJoin2.toString());		    
		    
		    String tKey = t1v1 +  "|" + t2v1;
		    
		    // A match in the middle predicate, take the tupels
		    if(middleAtomMatches.containsKey(tKey)) {
			
			// We assume that the middle predicate has only join variables, we ignore the rest TODO: also take other var. into result
			
			Pair<ResultLDM, ResultLDM> res = new Pair<ResultLDM, ResultLDM>(tupleLeft, tupleRight);
			resultPairs.add(res);
			
			
		    }
		    
        	} 
		 
		 
	     }

	 }

	 // Create merge variables
	ArrayList<ResultLDM> tempResults = mergeResultPairs(resultPairs, termProj1.toString(), termProj2.toString());
		
	QueryResultLDM retRes = new QueryResultLDM(0); 
	retRes.mResults = tempResults;

	 return retRes;
   }   
   

    //protected QueryResultLDM getResultsJoin(Rule joinRule, Hashtable<String, QueryResultLDM> interResults) {
    protected QueryResultLDM evalNormalJoinTwoAtoms(Rule joinRule, ArrayList<ResultLDM> results1, ArrayList<ResultLDM> results2,
	    					boolean joinProjectionVariables) {

	 ArrayList<Pair<ResultLDM, ResultLDM>> resultPairs = new ArrayList<Pair<ResultLDM, ResultLDM>>();

	 // Projection variables taken from head atom (initially consider two variables)
	 ILiteral headAtom = joinRule.getHeadLiteral();
	 ITerm termProj1 = headAtom.getTerm(0); // First head variable
	 ITerm termProj2 = headAtom.getTerm(1); // Second variable

	 ITerm termJoin1 = null;
	 //ITerm termJoin2 = null; // For now only a single join variable...
	 
	 // Find the join variable for left and right atom
	 if (joinProjectionVariables) {
	     termJoin1 = termProj1;
	     //termJoin2 = termProj2;
	 }
	 else {
	     ILiteral[] bodyLiterals = joinRule.getBodyLiterals();
	     ILiteral leftAtom = bodyLiterals[0];
	     ILiteral rightAtom = bodyLiterals[1];    
	     
	     outerloop: // Label for outer loop
	     for(ITerm termLeft :  leftAtom.getTerms()) {
		 
		 for(ITerm termRight :  rightAtom.getTerms()) {
		   
		     if(termLeft.toString().equalsIgnoreCase(termRight.toString())) {
			 termJoin1 = termLeft;
			 break outerloop;
		     }
		     
		 }
	     }
	     
	 }
	 

	 // Simplest algorithm is "Nested Loop Join" (good if one table is small), TODO: try "Merge Join" or "Hash Match Join"
	 for(ResultLDM tuple1 : results1) {

	     for(ResultLDM tuple2 : results2) {

		// If no join variables do a cross join
		if(termJoin1!=null) {
		     
        		 //boolean match = true;
        		 String t1v1 = tuple1.getAttByVar(termJoin1.toString());
        		 // String t1v2 = tuple1.getAttByVar(termJoin2.toString());
        		 String t2v1 = tuple2.getAttByVar(termJoin1.toString());
        		 // String t2v2 = tuple2.getAttByVar(termJoin2.toString());
        
        		 // TODO: This is still a bug, and has to be extended to term2. For now they always merge		 
        		 if(t1v1!=null && t2v1!=null && ! t1v1.equals(t2v1))
        		     continue; //match = false;
        		 
        		 //if(t1v2!=null && t2v2!=null && ! t1v2.equals(t2v2))
        		 //    match = false;
        		 
        		 // If no values ignore this combination
        		 if(t1v1==null || t2v1==null )  // || t1v2==null || t2v2==null
        		     continue; //match = false;
        	} 
		 
		 Pair<ResultLDM, ResultLDM> res = new Pair<ResultLDM, ResultLDM>(tuple1, tuple2);
		 resultPairs.add(res);
		 
	     }

	 }

	 // Create merge variables

	ArrayList<ResultLDM> tempResults = mergeResultPairs(resultPairs, termProj1.toString(), termProj2.toString());
		
	QueryResultLDM retRes = new QueryResultLDM(0);  // ignore geoID
	retRes.mResults = tempResults;

	 
	 return retRes;
   }
    
    
   private String createAggregateAtomView(ILiteral aggregateAtom, String sessionID, int aggCount) throws Exception  {
	
	//  Create aggregate query
	String temp1 = aggregateAtom.getAddParameter();
	String[] aggPara = temp1.split(",");
	
	String aggFkt = aggPara[0];
	if(aggFkt.equalsIgnoreCase("count")) {
	    aggFkt = "COUNT(x)";
	}
	else if(aggFkt.equalsIgnoreCase("avg")) {
	    aggFkt = "AVG(x)";
	}
	else if(aggFkt.equalsIgnoreCase("sum")) {
	    aggFkt = "SUM(x)";
	}
	else if(aggFkt.equalsIgnoreCase("max")) {
	    aggFkt = "MAX(x)";
	}
	else if(aggFkt.equalsIgnoreCase("min")) {
	    aggFkt = "MIN(x)";
	}
	else if(aggFkt.equalsIgnoreCase("last")) {
	    aggFkt = "(array_agg(x))[1]"; // string_agg(x, ';')
	}
	else if(aggFkt.equalsIgnoreCase("first")) {
	    aggFkt = "(array_agg(x))[1]"; // array_agg(x)[1]
	}
	else if(aggFkt.equalsIgnoreCase("line")) {
	    aggFkt = "'LINESTRING (' || string_agg(x, ', ') || ')' "; // POINT
	}
	else if(aggFkt.equalsIgnoreCase("line_angle")) {
	    // For set the angle to 0 (heads straight). TODO: proper calculate the deg.
	    aggFkt = "MAX(trunc(random() * 10 + 2))";
	}
	else {
	    
	    // TODO: Geometrical aggregation
	    throw new Exception("Spatial aggregate not supported");
	}
	
	int window = Integer.parseInt(aggPara[1]);
	String aggregateView = "V_" + sessionID + "_" + aggregateAtom.getPredicate().getName() + aggCount + "_f";
	
	Object[] args0= new Object[6];
	args0[0] = aggregateView; // View name
	args0[1] = window + " seconds"; // Window
	ITerm term1 = aggregateAtom.getTerm(0); // Take the first variable
	args0[2] = "iid_" + term1.toString();
	args0[3] = aggFkt ; // + "(x)" // val1  // Aggregation 
	ITerm term2= aggregateAtom.getTerm(1); // Take the second variable
	args0[4] = "att_" + term2.toString();
	args0[5] = "stream_" + aggregateAtom.getPredicateName(); // "S" + aggregateAtom.getPredicate().toString(); // Stream name
	
	String queryAggTemp =  Util.getContentEnBloc("templates/query_ldm_agg_tmpl.rq");
	String aggQuery = String.format(queryAggTemp, args0);	
	
	mConnection.executeStatement(aggQuery); 
	
	return aggregateView;

	
    }    
    
    private String evalNormalAtoms(ILiteral[] normalAtoms, String sessionID, String namePostix)  { // ArrayList<ILiteral>

    
	String queryVar = "";
	String queryWhere = "";

        StringWriter whereWriter = new StringWriter();
       
	int viewColumns = 1; // headAtom.getArity(); 
		
	HashSet<String> headVarHash = new HashSet<String>();  
	ArrayList<String> headVar = new ArrayList<String>();  
	ArrayList<ILiteral> buildInLiterals = new ArrayList<ILiteral>(); 
	
	// Query where
	for( ILiteral lit : normalAtoms) {
	    
	    // Builtin like X < 20
	    if (lit instanceof BuiltInLiteral) {
		
		buildInLiterals.add(lit);
		
		/* TODO: Probem Owlgres ignores them, so remember them and add them to the final query
		
		whereWriter.write("FILTER(");
		
		for(ITerm buildin: lit.getTerms()) {
		    if(buildin instanceof Variable) 
			whereWriter.write("?" + buildin.toString());		    
		    else 
			whereWriter.write(buildin.toString());		 	
		}
		whereWriter.write(")  \n");
		
		*/
		
	    }
	    // Concepts
	    else if(lit.getArity()==1) {
		String varName = lit.getTerm(0).toString();
		
		whereWriter.write("?" + varName);
		whereWriter.write(" rdf:type ");
		whereWriter.write(" tuwt:" + lit.getPredicate().getName());
		whereWriter.write(". \n");
		if(!headVarHash.contains(varName))
		    headVar.add("?" + varName);
		headVarHash.add(varName);
	    }
	    // Roles
	    else if(lit.getArity()==2) {
		String varName1 = lit.getTerm(0).toString();
		String varName2 = lit.getTerm(1).toString();
		
		whereWriter.write(" ?" + varName1);
		whereWriter.write(" tuwt:" + lit.getPredicate().getName());
		whereWriter.write(" ?" + varName2);
		whereWriter.write(". \n");

		if(!headVarHash.contains(varName1))
		    headVar.add("?" + varName1);
		headVarHash.add(varName1);
		if(!headVarHash.contains(varName2))
		    headVar.add("?" + varName2);
		headVarHash.add(varName2);

		//viewColumns = 1;
		
	    }
	    else if(lit.getArity()>2) {
		throw new UnsupportedOperationException();
	    }
	}
	
        // Query variables
	for(String var: headVar) {
	    if(queryVar.length()>0)
		queryVar = queryVar + " ";
	    queryVar = queryVar + var;
	   
	}
	
	// Get Template
	String querySparql =  Util.getContentEnBloc("templates/query_ldm_nor_tmpl.rq");
	
	Object[] args1 = new Object[2];
	args1[0] = queryVar;
	args1[1] = whereWriter.toString();
	
	String owlgresQuery = String.format(querySparql, args1);

	
	//int index = 1;
	String viewName =  "V_" + sessionID + "_" + namePostix   ; // "_M"

	// Call Owlgres
	if (owlgresQuery != null) {
	    callOwlgres(owlgresQuery, viewName, viewColumns);
	    return viewName;
	}

	return "";
	
    }
	

    protected ArrayList<Pair<ResultLDM, ResultLDM>> evalSpatialRelationIntersect(ArrayList<ResultLDM> results1, ArrayList<ResultLDM> results2) {
        
	ArrayList<Pair<ResultLDM, ResultLDM>> results = new ArrayList<Pair<ResultLDM, ResultLDM>>();
  		    		
	HashSet<String> alreadyAdded = new HashSet<String>();  
	
  	for (int j = 0; j < results2.size(); j++) {
  	    ResultLDM feat2 = results2.get(j);
 
  	    Geometry geo2 = feat2.getGeometry() ;
 	    if(geo2 == null) continue;
 	    
  	    for (int i = 0; i < results1.size(); i++) {		
  		ResultLDM feat1 = results1.get(i);
  	 
  		Geometry geo1 = feat1.getGeometry() ;
  		if(geo1 == null) continue;
  		
 	        if(geo2.intersects(geo1)  ) {

	           String tKey = feat1.getID() + "_"  + feat2.getID();

 	           if(! alreadyAdded.contains(tKey)) {
 	            
 	               Pair<ResultLDM, ResultLDM> res = new Pair<ResultLDM, ResultLDM>(feat1, feat2);
 	               results.add(res);
 	               alreadyAdded.add(tKey);
 	               
 	           }
  		}
  	    }
  		
  	}
  		   	
       	return results;
    
    }
    
    protected ArrayList<Pair<ResultLDM, ResultLDM>> evalSpatialRelationInside2(ArrayList<ResultLDM> results1, ArrayList<ResultLDM> results2) {
        
	ArrayList<Pair<ResultLDM, ResultLDM>> results = new ArrayList<Pair<ResultLDM, ResultLDM>>();
  		    		
  	for (int j = 0; j < results2.size(); j++) {
  	    ResultLDM feat2 = results2.get(j);
 
  	    Geometry geo2 = feat2.getGeometry() ;
 	    if(geo2 == null) continue;
 	    
  	    for (int i = 0; i < results1.size(); i++) {		
  		ResultLDM feat1 = results1.get(i);
  	 
 	        if(geo2.contains( feat1.getGeometry())) {
 	           Pair<ResultLDM, ResultLDM> res = new Pair<ResultLDM, ResultLDM>(feat1, feat2);
 	           results.add(res);
  		}
  	    }
  		
  	}
  		   	
       	return results;
    
    }
    
    protected ArrayList<Pair<ResultLDM, ResultLDM>> evalSpatialRelationNext2(ArrayList<ResultLDM> pois1, ArrayList<ResultLDM> pois2) { 
	      
	ArrayList<Pair<ResultLDM, ResultLDM>> results = new ArrayList<Pair<ResultLDM, ResultLDM>>(); // PointOfInterest
	
 	    for (int i = 0; i < pois1.size(); i++) {		
 		ResultLDM poi1 = pois1.get(i);
		
 		poi1.createPointFromAtts();
 		
		for (int j = 0; j < pois2.size(); j++) {
		    ResultLDM poi2 =  pois2.get(j);
		    
		    poi2.createPointFromAtts();
		    
	     	    double dist = DistanceUtils.haversine(poi1.getPoint().getX(), poi1.getPoint().getY(), poi2.getPoint().getX(), poi2.getPoint().getY(), GeoUtil.EARTH_RADIUS);
	     	    if(dist <= mDistBetween) {
	     		Pair<ResultLDM, ResultLDM> res = new Pair<ResultLDM, ResultLDM>(poi1, poi2);
			results.add(res);
			
	     	    }
		}
	    }
 	    
 	    return results;
   }

    
    private void callOwlgres(String sparqlQuery, String viewName, int columns)
    {

	try {
	    queryRewrite.main(mHost,mDB, mUser, mPw,  true, true, viewName, "public", columns, sparqlQuery);

	} catch (RuntimeException e) {
	    throw e;
	} 


    }
   
   /* 
    
   public void cleanAllViews() {
	
	String sql = "select 'DROP VIEW IF exists ' || viewname || ' cascade;' as a from pg_views where viewname like 'v_%';";
	
	connectDB();
	
	try {
	    Statement stmt1 = mConnection.createStatement();
	    ResultSet rset = stmt1.executeQuery(sql); // .toString()

	    while (rset.next()) {
	     String sqlDel = rset.getString("a");
	     
	     mConnection.executeStatement(sqlDel);
	     	     
	    }
	    
	    
	} catch (RuntimeException e) {
	    throw e;
	} catch (SQLException e) {
	    throw new RuntimeException(e.toString());
	}	
	
	//mConnection.
	
	closeDB(null);
	
    }
    
    */
    
    
}
