package edu.tuwien.kbs.ldm.parser;

/**
 * Represents a constant in a literal.
 */
public class Constant implements ITerm {
	
    
    public static final String PRED_OPEN = "(";
    public static final String PRED_CLOSE = ")";
    public static final String PRED_EXT_OPEN = "[";
    public static final String PRED_EXT_CLOSE = "]";
    public static final String NAF = "not";
    public static final String CLAS_NEGATION = "-";

    
    /** The NULL constant. */
    public static final Constant NULL=new Constant(null);
    
    /** The "_" unbound constant.  */
    public static final String UNBOUND = "_";

    /** The value of the constant. */
    protected Object m_value;
    protected boolean m_isUnbound = false;

    /**
     * Creates a null constant.
     */
    public Constant() {
        this(null);
    }
    
    /**
     * Creates a constant.
     */
    public Constant(Object value) {
        m_value=value;
        if(m_value instanceof String && UNBOUND.compareTo(m_value.toString()) == 0) m_isUnbound = true;
    }
    
    /**
     * Creates a constant as boolean.
     */
    public Constant(boolean value) {
        m_value=value ? Boolean.TRUE : Boolean.FALSE;
    }
    
    /**
     * Creates a constant as integer.
     */
    public Constant(int value) {
        m_value=new Integer(value);
    }
    
    /**
     * Returns the value of the constant.
     */
    public Object getValue() {
        return m_value;
    }
    
    /**
     * Returns the depth of this term.
     */
    public int getDepth() {
        return 0;
    }
    
    /**
     * Returns if this term represents the unbound constant.
     */
    public boolean isUnbound() {
        return m_isUnbound;
    }
    
    /**
     * Converts this constant to a string.
     *
     * @return                                  the string representation of the constant
     */
    public String toString() {
        if (m_value==null)
            return "<null>";
        else if (m_value instanceof String)
            return m_value.toString(); // "'"+m_value.toString()+"'"
        else if (m_value instanceof Boolean)
            return '['+m_value.toString()+']';
        else
            return m_value.toString();
    }
    
    /**
     * Checks if this constant is equal to some other constant.
     *
     * @param other                             the other constant
     * @return                                  <code>true</code> if this constant is equal to some other constant
     */
    public boolean equals(Object other) {
        if (this==other)
            return true;
        if (!(other instanceof Constant))
            return false;
        Object otherValue=((Constant)other).m_value;
        return m_value==otherValue || (m_value!=null && m_value.equals(otherValue));
    }
    
    /**
     * Returns the hash-code of this variable.
     */
    public int hashCode() {
        return m_value==null ? 0 : m_value.hashCode();
    }
}
