package edu.tuwien.kbs.ldm.parser;


public interface ILiteral {


    public abstract boolean IsAggregate();

    public abstract String getAddParameter();
    
    /**
     * Returns the attributes (att1, att1, ...) for the view or table to be used.
     */
    public abstract String getEntityAttribueNames();

    /**
     * Returns the predicate of this literal.
     */
    public abstract Predicate getPredicate();

    public abstract String getPredicateName();
    
    /**
     * Returns the embedded predicate of this literal. (e.g. used by OntoLiteralUpdate)
     */
    public abstract Predicate[] getEmbeddedPredicates();

    /**
     * Returns the SQL view name, in case a view for this predicate is constructed.
     */
    public abstract String getPredicatNameAsView();

    /**
     * Returns if this Literal is part of a recursion.
     */   
    public abstract boolean IsRecursive();

    /**
     * Sets that this Literal is part of a recursion.
     */ 
    public abstract void setIsRecursive(boolean rec);

    /**
     * Returns whether this literal is positive.
     */
    public abstract boolean isPositive();

    /**
     * Returns the arity of the literal.
     */
    public abstract int getArity();

    /**
     * Returns <code>true</code> if given argument is bound to a variable.
     */
    public abstract boolean isArgumentBoundToVariable(int argumentIndex);

    /**
     * Returns <code>true</code> if given argument is bound to a constant.
     */
    public abstract boolean isArgumentBoundToConstant(int argumentIndex);

    /**
     * Returns the term at given position.
     */
    public abstract ITerm getTerm(int argumentIndex);

    /**
     * Returns the variable at given position.
     */
    public abstract Variable getArgumentVariable(int argumentIndex);

    /**
     * Returns the constant of the given argument.
     */
    public abstract Constant getArgumentConstant(int argumentIndex);

    /**
     * Returns the array of terms of this literal.

     */
    public abstract ITerm[] getTerms();

    /**
     * Returns <code>true</code> if this literal contains supplied variable.
     */
    public abstract boolean containsVariable(Variable variable);

    /**
     * Returns the position of the supplied variable. In case it is not found -1 will be returned
     */
    public abstract int positionOfVariable(Variable variable);

}