package edu.tuwien.kbs.routing;

import java.io.IOException;
import java.util.ArrayList;

import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Elements;
import nu.xom.Node;
import nu.xom.ParsingException;

import com.vividsolutions.jts.geom.*;
import com.vividsolutions.jts.io.ParseException;
import com.vividsolutions.jts.io.WKTReader;

import edu.tuwien.kbs.geo.GeoUtil;

public class Isochrone {

    public String Mode;
    public String StartTime;
    public int TimeLimit;
    protected ArrayList<Geometry> mElements;
    
    protected static String AIT_Namespace = "http://www.ait.ac.at/ariadne/gml/1.0";
    protected static String GML_LineString = "LineString"; // gml:
    protected static String GML_CurveProp = "curveProperty"; // gml:
    protected static String AIT_Edge = "Edge"; // ait:
    protected static String GML_Feature = "featureMember"; // gml:
    protected static String GML_Coordinates = "coordinates"; //gml:

    protected static String WKT_LineString = "LINESTRING"; 
    
    public Isochrone() {
	mElements = new ArrayList<Geometry>();
	
    }
    
    public Isochrone(String mode, String startTime, int timeLimit) {
	this();
	
	Mode = mode;
	StartTime = startTime;
	TimeLimit = timeLimit;
	
    }
    
    public ArrayList<Geometry>  getElements() {
	return mElements;
    }
    
    public Geometry[] getElementsAsArray() {
	return (Geometry[]) mElements.toArray(new Geometry[mElements.size()]);
    }

    
    public void fillDetailsFromWKT(String wktText, GeometryFactory geometryFactory) {

	WKTReader reader = new WKTReader(geometryFactory);
	String[] splitS = wktText.split(WKT_LineString);
	for(int i=1; i < splitS.length; i++) { // First is empty
	    Geometry line = null;
	     try {
		line = reader.read(WKT_LineString + splitS[i]);
	    } catch (ParseException e) {
		// Ignore
	    }
	    if(line!=null) mElements.add(line);
	}
	
    }

    public void fillDetailsFromGML(String xml) {

	Element root = null;
	
	try {
	    Builder parser = new Builder();
	    Document doc = parser.build(xml, AIT_Namespace);
	    root = doc.getRootElement();

	} catch (ParsingException ex) {
	    ex.printStackTrace();
	} catch (IOException ex) {
	    ex.printStackTrace();
	}
	
	Elements features = root.getChildElements(); // root.getChildElements(GML_Feature , GML_Namespace) ;
	// System.out.println("Iso:" + xml);
	
	for (int i = 0; i < features.size(); i++) {
	    	    
	    Element feature = features.get(i);
	    if(! feature.getLocalName().equals(GML_Feature)) continue;	    
	    Element edge = feature.getFirstChildElement(AIT_Edge, AIT_Namespace );
	    if(edge==null) continue;	
	    
	    Element curveProp = edge.getFirstChildElement(GML_CurveProp, GeoUtil.GML_Namespace);
	    if(curveProp==null) continue;
	    Element lineString = curveProp.getFirstChildElement(GML_LineString, GeoUtil.GML_Namespace);
	    if(lineString==null) continue;	

	    
	    Element coordinate = lineString.getFirstChildElement(GML_Coordinates, GeoUtil.GML_Namespace);	    
	    if (coordinate != null && coordinate.getValue() != null && coordinate.getValue().length() > 0) {
		// System.out.println("Ls: " + coordinate.getValue());
		LineString line = GeoUtil.coordinatesToLineString(coordinate.getValue());
		if(line!=null) mElements.add(line);
	    }
	}	  	
	
	
    
    }
    
 /*   public void fillDetailsOld(String xml) {
	
	// ArrayList<Geometry> geos = new ArrayList<Geometry>();
	ArrayList<Element> nodes = new ArrayList<Element>();
	
	try {
	    Builder parser = new Builder();
	    Document doc = parser.build(xml, AIT_Namespace);
	    Element root = doc.getRootElement();
	    getChildrenOld(root, nodes, GML_LineString);

	} catch (ParsingException ex) {
	    ex.printStackTrace();
	} catch (IOException ex) {
	    ex.printStackTrace();
	}
	  
	for (int i = 0; i < nodes.size(); i++) {
	    Element ls = nodes.get(i);
	    Element coordinate = ls.getFirstChildElement(GML_Coordinates,
		    GeoUtil.GML_Namespace);
	    if (coordinate != null) {
		LineString lineStr = GeoUtil.coordinatesToLineString(coordinate.getValue());
		mElements.add(lineStr);

	    }
	}	  

    }

    
    private void getChildrenOld(Node current, ArrayList<Element> nodes,  String selNsodeName) {

	if (current instanceof Element) {
	    Element temp = (Element) current;
	    if (temp.getQualifiedName().equals(selNsodeName)) {
		nodes.add(temp);
	    }
	}
	// Attributes are never returned by getChild()
	for (int i = 0; i < current.getChildCount(); i++) {
	    getChildrenOld(current.getChild(i), nodes, selNsodeName);
	}

    }
    */
    
    
}
