package edu.tuwien.kbs.routing;

import com.vividsolutions.jts.geom.Point;

import edu.tuwien.kbs.geo.GeoUtil;
import nu.xom.Element;

public class PointOfInterest extends FeatureOfInterest {

    public static String DefaultCity= "Wien";
    public static String SourceOSM  = "OSM";
    public static String SourceOGD  = "OGD Wien";
    public static String SourceFalter  = "Falter";
    
    
    public String Street;
    public String HouseNr;
    public String City;
    public String PostCode;
    public String Phone;
    public String Email;
    public String Url;
    public String Source; // int
    public String Description;
    public String OpeningTimes;
    public String Categories;

       
    public PointOfInterest()
    {
	super();
    }
    
    public PointOfInterest(String poiId, String name, String pointX, String pointY, int geoID)
    {
	super();
	
	Name = name;
	Id = poiId;	
	mLocPoint =  GeoUtil.createPoint(pointX, pointY, geoID, false);	
    }
    
    public PointOfInterest(String poiId, String name, Point loc) // double pointX, double pointY)
    {
	super();
	
	Name = name;
	Id = poiId;
	mLocPoint = loc;	
    }
    
    public String toCoord(boolean switchXY) {	
	return GeoUtil.pointToCoord(mLocPoint, switchXY, ',');
    }
    
    @Override
    public Element serializeToElement(String alternateName) {
    
	
	Element poi = super.serializeToElement(alternateName);
	
	if(Street != null) {
	    Element street = new Element("tuw:street", RouteResults.TU_Namespace);
	    String text = Street;
		if(HouseNr != null) {
		    text = text + " "  + HouseNr;
		}
	    street.appendChild(text);
	    poi.appendChild(street);
	    if(City==null) City = DefaultCity; // Simple rule, because first test are in Vienna, future should be spatial containment ...	    
	}	

 	if(City != null) {
 	   Element city = new Element("tuw:place", RouteResults.TU_Namespace);
 	   if(PostCode != null) 
 	       city.appendChild(PostCode + " " + City); // TODO: Implement a more international solution, just works for austria, germany, switzerland, not UK or US
 	   else
 	   	city.appendChild(City);
 	   poi.appendChild(city);
 	}

	
	if(Phone != null) {
	    Element phone = new Element("tuw:phone", RouteResults.TU_Namespace);
	    phone.appendChild(Phone);
	    poi.appendChild(phone);
	}	
	if(Email != null) {
	    Element email = new Element("tuw:email", RouteResults.TU_Namespace);
	    email.appendChild(Email);
	    poi.appendChild(email);
	}	

	if(Url != null) {
	    Element url = new Element("tuw:url", RouteResults.TU_Namespace);
	    url.appendChild(Url);
	    poi.appendChild(url);
	}
	
	if(Source != null) {
	    Element src = new Element("tuw:source", RouteResults.TU_Namespace);
	    int srcInt = Integer.parseInt(Source);
	    String srcText = "";
	    
	    if(srcInt >= 100 && srcInt < 200) srcText = SourceOSM;
	    if(srcInt >= 200 && srcInt < 300) srcText = SourceOGD;
	    if(srcInt >= 300 && srcInt < 400) srcText = SourceFalter;
	    
	    src.appendChild(srcText);
	    poi.appendChild(src);
	}
	
	if(Description != null) {
	    Element desc = new Element("tuw:description", RouteResults.TU_Namespace);
	    desc.appendChild(Description);
	    poi.appendChild(desc);
	}
	
	if(OpeningTimes != null) {
	    Element ot = new Element("tuw:opening", RouteResults.TU_Namespace);
	    ot.appendChild(OpeningTimes);
	    poi.appendChild(ot);
	}
	
	if(Categories != null) {
	    Element cat = new Element("tuw:categories", RouteResults.TU_Namespace);
	    cat.appendChild(Categories);
	    poi.appendChild(cat);
	}
	
//	if(PostCode != null) {
//	    Element pc = new Element("tuw:postcode", RouteResults.TU_Namespace);
//	    pc.appendChild(PostCode);
//	    poi.appendChild(pc);
//	}
//	if(City != null) {
//	    Element city = new Element("tuw:city", RouteResults.TU_Namespace);
//	    city.appendChild(City);
//	    poi.appendChild(city);
//	}
		
	return poi;
    }
    
    
    
}
