package edu.tuwien.kbs.routing;

import java.io.IOException;
import java.util.ArrayList;

import nu.xom.*;

import com.vividsolutions.jts.geom.*;

import edu.tuwien.kbs.util.Util;


public abstract class Route {

    public static String AIT_Namespace = "http://www.ait.ac.at/ariadne/gml/1.0";
    public static String GML_Namespace = "http://www.opengis.net/gml/3.2";

    
    public String Id;
    public String Mode = "";
    public String ModeSequence = "";

    public double Time;
    public String StartTime = "";
    public double Distance;
    public int Changes;
    public boolean Recommended;
    protected ArrayList<Geometry> mElements;
    protected PointOfInterest  mAssignedPOI; // PointOfInterest
    
    public int FromId; // String
    public String ToId;
    
    public String rawXML;
  
    public Route() {
	mElements = new ArrayList<Geometry>();
	
    }
   
    public void assignPOI( ArrayList<PointOfInterest> pois) {
	// TODO: Implement
    }
    
    public ArrayList<Geometry>  getElements() {
	return mElements;
    }
    
    public PointOfInterest getPOI() {
	return mAssignedPOI;
    }
    
    public String getModeSequenceShort() {
	return ModeSequence.replace(",", "").replace("FOOT", "F").replace("PUBLIC", "P").replace("CAR", "C").replace("BIKE", "B").replace("TO", "").replace("VIA", "V");
    }
    
    public void setPOI(PointOfInterest val) {
	mAssignedPOI = val;
	val.RouteAssigned = true;
    }
       
    public Element serializeToElement(String alternateName) {
    
	Element route = new Element("tuw:route", RouteResults.TU_Namespace);
	
	Attribute id = new Attribute("id", Id); // "tuw:id"
	route.addAttribute(id);
	
	Element mots = new Element("tuw:mode", RouteResults.TU_Namespace);
	mots.appendChild(Mode);
	route.appendChild(mots);

	Element motsSeq = new Element("tuw:sequence", RouteResults.TU_Namespace);
	motsSeq.appendChild(ModeSequence);
	route.appendChild(motsSeq);	
	
	Element start = new Element("tuw:startTime", RouteResults.TU_Namespace);
	start.appendChild(StartTime);
	route.appendChild(start);	
	
	
	Element time = new Element("tuw:time", RouteResults.TU_Namespace);
	time.appendChild(Double.toString(Time));
	route.appendChild(time);
	
	Element dist = new Element("tuw:distance", RouteResults.TU_Namespace);
	dist.appendChild(Double.toString(Distance));
	route.appendChild(dist);

	if(Recommended) {
	    Element rec = new Element("tuw:recommendedRoute", RouteResults.TU_Namespace);
	    route.appendChild(rec);
	}

	// Element change = new Element("tuw:changes", RouteResults.TU_Namespace);
	// change.appendChild(Integer.toString(Changes));
	// route.appendChild(change);
	
	if(mAssignedPOI!=null) {
	    Element poi = mAssignedPOI.serializeToElement(alternateName);
	    route.appendChild(poi);
	}
	 
	
	return route;
    }
    
    public boolean fillDetails(String xml, String modFunction, String poiName) {
	
	Element root = null;
	
	rawXML = xml;
	
	try {
	    Builder parser = new Builder();
	    Document doc = parser.build(xml, AIT_Namespace);

	    root = doc.getRootElement();
	    // getChildren(root, nodes, GML_LineString);

	} catch (ParsingException ex) {
	    //ex.printStackTrace();
	    System.out.println("Parsing Error of: " + poiName + ": " + ex.toString());
	    return false;
	} catch (IOException ex) {
	    ex.printStackTrace();
	    return false;
	}	

	// Meta Data
	Element metaData = root.getFirstChildElement("routeMetaData", AIT_Namespace);
		
	// Id
	Element el1 = root.getFirstChildElement("RouteName", AIT_Namespace);
	if(el1 != null) Id = el1.getValue();

	// Recommended Route
	Element el6 = root.getFirstChildElement("recommendedRoute", AIT_Namespace);
	if(el6 != null)  Recommended = true;

	// From / To
	Element el4 = metaData.getFirstChildElement("from",
		AIT_Namespace);
	if(el4 != null && Util.isNumeric(el4.getValue())) FromId = Integer.valueOf(el4.getValue());

	Element el5 = metaData.getFirstChildElement("to",
		AIT_Namespace);
	if(el5 != null && Util.isNumeric(el5.getValue())) ToId = el5.getValue();

	
	// Mode
	Element modesOfTrans= metaData.getFirstChildElement("modesOfTransport", AIT_Namespace);
	if(modesOfTrans != null) {
	    Elements mots = modesOfTrans.getChildElements("modeOfTransport", AIT_Namespace);
	    for (int i = 0; i < mots.size(); i++) {
		if(i > 0) Mode = Mode + ",";
		nu.xom.Attribute atmot = mots.get(i).getAttribute("mode");
		if(atmot != null) 
		    Mode = Mode + atmot.getValue();		
	    }
	}
	
	// Time
	Element el2 = metaData.getFirstChildElement("routingConfiguration", AIT_Namespace);
	if(el2 != null) {
	    nu.xom.Attribute at2 = el2.getAttribute("configurationID");
	    Element el2b = el2.getFirstChildElement("routeWeight", AIT_Namespace);
	    if(at2!=null && at2.getValue().equals("TravelTime1") && el2b != null) 
		Time = Double.valueOf(el2b.getValue());	 
	    Element el2c = el2.getFirstChildElement("startTime", AIT_Namespace);
	    if(at2!=null && at2.getValue().equals("TravelTime1") && el2c != null) 
		StartTime =  el2c.getValue();	    

	}
	    
	
	Element el3 = metaData.getFirstChildElement("scoringConfigurations", AIT_Namespace);	
	if(el3 != null) {
	    Element el3a = el3.getFirstChildElement("scoringConfiguration", AIT_Namespace);
	    if(el3a != null) {
		nu.xom.Attribute at3 = el3a.getAttribute("configurationID");
		Element el3b = el3a.getFirstChildElement("routeWeight", AIT_Namespace);
		if(at3!=null && at3.getValue().equals("Distance2") && el3b != null)
		    Distance = Double.valueOf( el3b.getValue());
	    }
	}

	// Get sequence of modes, but just use them, if changed
	String modSave = null;
	Elements edges = root.getChildElements("featureMember", GML_Namespace);
	for (int i = 0; i < edges.size(); i++) {
	    Element edgeEl = edges.get(i).getFirstChildElement("Edge", AIT_Namespace);
	    if (edgeEl != null) {
		nu.xom.Attribute atmot = edgeEl.getAttribute("modeOfTransportRef");
		if (atmot != null) {
		    String modTemp = atmot.getValue();
		    if (!modTemp.equals(modSave) && modTemp.length() > 3) { // Strange behavior, because sometimes mod is just "MOD" without more spec.
			if (ModeSequence.length() > 0)  ModeSequence = ModeSequence + ",";
			ModeSequence = ModeSequence + modTemp.substring(3);
		    }
		    modSave = modTemp;
		}
		Element edgeFkt = edgeEl.getFirstChildElement("function", AIT_Namespace);
		if (edgeFkt != null) {
		    String modTemp = edgeFkt.getValue();
		    if (modTemp.equalsIgnoreCase(modFunction)) {
			if (ModeSequence.length() > 0)  ModeSequence = ModeSequence + ",";
			ModeSequence = ModeSequence + modFunction;
			modSave = null;
		    }
		}
	    }

	}

	return true;

   }
    
    public void fillExtraDetails(Element metaData) {
    }
	
    
}



