package edu.tuwien.kbs.routing;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;

import nu.xom.Document;
import nu.xom.Element;

import com.vividsolutions.jts.geom.Point;

import edu.tuwien.kbs.geo.GeoUtil;


public class RouteResults {

    public static String TU_Namespace = "http://www.kr.tuwien.ac.at/myits/search/1.0";
   
    protected  String mId; 
    protected ArrayList<Route> mRoutes; 
    protected  Point mFrom; // Point
    protected  Point mTo; 
    protected  String mMots; 
    protected  int mTimeLimit;    
    protected  String mStartTime;    
    //String alternateName
    protected  String mKeywords; 
    
    public RouteResults() {
	mRoutes = new ArrayList<Route>();
    }
    
    public RouteResults(Point from, String mots, int timeLimit, String startTime, String id, String kwds) {
	this();
	
	mFrom = from; mMots = mots; mTimeLimit = timeLimit; mStartTime = startTime;mId=id;mKeywords=kwds;
    } 
    
    public RouteResults(Point from, Point to, String mots, int timeLimit, String startTime, String id, String kwds) {
	this(from, mots, timeLimit, startTime, id, kwds);
	
	mTo = to; 
    } 
    
    public ArrayList<Route>  getElements() {
	return mRoutes;
    }

    public  Route  get(int i) {
	return mRoutes.get(i);
    }
    
    public  String  getId() {
	return mId;
    }
    
    public int size() {
	return mRoutes.size();
    }
    
    public void  addRoute(Route r) {
	mRoutes.add(r);
    }
   
   public String getStats() {
       	
       double sumD = 0; double sumT = 0;
       double maxD = Double.MIN_VALUE; double maxT = Double.MIN_VALUE;
       double minD = Double.MAX_VALUE; double minT = Double.MAX_VALUE;
       double avgD = 0; double avgT = 0;
       double medianD = 0; double medianT = 0;
       double[] valuesSortedD = new double[mRoutes.size()];
       double[] valuesSortedT = new double[mRoutes.size()];
       StringBuffer modals = new StringBuffer();
       
	for (int i = 0; i < mRoutes.size(); i++) {
	    Route rt = mRoutes.get(i);
	    
	    valuesSortedD[i] = rt.Distance;
	    valuesSortedT[i] = rt.Time;
	    
	    sumD+=rt.Distance;
	    sumT+=rt.Time;
	    
	    if(maxD < rt.Distance) maxD = rt.Distance;
	    if(maxT < rt.Time) maxT = rt.Time;
	    if(minD > rt.Distance) minD = rt.Distance;
	    if(minT > rt.Time) minT = rt.Time;
	    
	    if(modals.length()>0)
		modals.append(",");
	    modals.append(rt.getModeSequenceShort());
	    
	    
	}
	
	if(mRoutes.size() > 0) {
	    // Calc Mean
	    avgD = sumD / mRoutes.size();
	    avgT = sumT / mRoutes.size();
	    if(mRoutes.size() == 1) {
		    medianD = sumD;
		    medianT = sumT;		
	    }
	    else if(mRoutes.size() == 2) {
		    medianD = avgD;
		    medianT = avgT;
	    }
	    else if(mRoutes.size() > 2) {
		// Calc Median
		Arrays.sort(valuesSortedD);
		Arrays.sort(valuesSortedT);
		int middle = ((valuesSortedD.length) / 2);
		if ((valuesSortedD.length % 2) == 0) {
		    medianD = ((double)valuesSortedD[middle] + (double)valuesSortedD[middle+1])/2;
		    medianT = ((double)valuesSortedT[middle] + (double)valuesSortedT[middle+1])/2;
		}
		else {
		    medianD = (double) valuesSortedD[middle/2];
		    medianT = (double) valuesSortedT[middle/2];
		}
	    }
	    // Output
	    DecimalFormat df = new DecimalFormat("#.###");
	    StringBuffer out = new StringBuffer();
	    
	    out.append("COUNT: " + mRoutes.size() + "; ");
	    
	    out.append("SUM_D: " + df.format(sumD) + "; ");
	    out.append("AVG_D: " + df.format(avgD)+ "; ");
	    out.append("MED_D: " + df.format(medianD)+ "; ");
	    out.append("MIN_D: " + df.format(minD)+ "; ");
	    out.append("MAX_D: " + df.format(maxD)+ "; ");
	    
	    out.append("SUM_T: " + df.format(sumT) + "; ");
	    out.append("AVG_T: " + df.format(avgT) + "; ");
	    out.append("MED_T: " + df.format(medianT) + "; ");
	    out.append("MIN_T: " + df.format(minT) + "; ");
	    out.append("MAX_T: " + df.format(maxT) + "; ");

	    out.append("MODAL: " + modals.toString() );

	    return out.toString();
	    
	}
	else {
	    return "";
	}
       
       
   }
   
   public String serializeToXml(boolean sort) {

	Element root = new Element("tuw:routes",TU_Namespace);
	
	// Put general information 
	Element id = new Element("tuw:id", TU_Namespace);
	id.appendChild(mId);
	root.appendChild(id);

	if(mFrom != null) {
	    Element from = new Element("tuw:from", TU_Namespace);
	    String coord =  GeoUtil.pointToCoord(mFrom, false, ' ');
	    from.appendChild(coord);
	    root.appendChild(from);
	}
	
	if(mTo != null) {
	    Element to = new Element("tuw:to", TU_Namespace);
	    String coordT =  GeoUtil.pointToCoord(mTo, false, ' ');
	    to.appendChild(coordT);
	    root.appendChild(to);	    
	}
	
	
	Element mots = new Element("tuw:modesOfTransport", TU_Namespace);
	mots.appendChild(mMots);
	root.appendChild(mots);
	
	Element timeLim = new Element("tuw:timeLimit", TU_Namespace);
	timeLim.appendChild(Integer.toString(mTimeLimit));
	root.appendChild(timeLim);

	Element start = new Element("tuw:startTime", TU_Namespace);
	start.appendChild(mStartTime);
	root.appendChild(start);

	Element kwds = new Element("tuw:keywords", TU_Namespace);
	kwds.appendChild(mKeywords);
	root.appendChild(kwds);
	
	
	if(sort) {
	    Collections.sort(mRoutes, new RouteDistanceComparator());
	}
	
	for (int i = 0; i < mRoutes.size(); i++) {
	    Element el = mRoutes.get(i).serializeToElement(null);
	    root.appendChild(el);
	}

	Document doc = new Document(root);
	
	return doc.toXML();
    }   
    
    
}
