package edu.tuwien.kbs.routing;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
// import java.util.UUID;

import org.apache.http.NameValuePair;
import org.apache.http.message.BasicNameValuePair;
import org.opensphere.geometry.algorithm.ConcaveHull;

import redis.clients.jedis.Jedis;

import com.vividsolutions.jts.algorithm.ConvexHull;
import com.vividsolutions.jts.geom.*;
import com.vividsolutions.jts.io.ParseException;

import edu.tuwien.kbs.geo.*;
import edu.tuwien.kbs.util.Util;

public class RouterFacade {
    
     // MOTS for Sequence of MOTS
    protected static String Edge_VIA = "VIA";
    protected static String Edge_TO = "TO";
    
    // AIT Server
    //protected static String mAIT_Server = "atlas.arsenal.ac.at";
    protected static int AIT_Port = 8080;
    protected static String AIT_Server_Path_GetRoutes = "myits-ws/rest/getRoutes";
    protected static String AIT_Server_Path_GetIso = "myits-ws/rest/getIsochrone";
    protected static String AIT_Server_Path_GetIsoWkt = "myits-ws/rest/getIsochroneWkt";
    protected static String AIT_Server_Path_GetViaRoutes = "myits-ws/rest/getViaRoutes";
    
    protected static boolean IsoAsWkt = true; // Turn on off for testing
    
    // Parameter
    protected static String AIT_Param_User = "userId";
    protected static String AIT_Param_From = "from";
    protected static String AIT_Param_To = "to";
    protected static String AIT_Param_Via = "via";
    protected static String AIT_Param_MOT = "mot";
    protected static String AIT_Param_StartTime = "startTime";
    protected static String AIT_Param_TimeLimit = "timeLimitSeconds";
    public static String AIT_RouteSeparator = "######";
    
    
    protected static String JedisRoutesNames = "Rts_";
    protected static String JedisIsoNames = "Iso_";

    protected Jedis mJedis;
    protected Random mRanGen;
    protected String mAIT_Server = "62.218.164.237"; // "atlas.arsenal.ac.at";
    
    protected static PrecisionModel mPrecisionModel = new PrecisionModel(
	    PrecisionModel.FLOATING);
    protected static GeometryFactory mGeometryFactory = new GeometryFactory(
	    mPrecisionModel);

    
    public RouterFacade() {
	
	mJedis = new Jedis("localhost");
	mRanGen = new Random();

    }
    
    public Geometry[] getIsochrone(String id, Point start, String mot, String startTime, int timeLimit, boolean logData, String userId, Map<String, String> optionsMap) // Geometry[] 
    {

	if(startTime==null || startTime.length() == 0) 
	    startTime = getActualTime();


	
	    // Call AIT Web-Service
	    ArrayList<NameValuePair> params = new ArrayList<NameValuePair>();
	    params.add(new BasicNameValuePair(AIT_Param_User, userId)); 
	    params.add(new BasicNameValuePair(AIT_Param_From, GeoUtil.pointToCoord(start, true, ',')));
	    params.add(new BasicNameValuePair(AIT_Param_MOT, mot));
	    params.add(new BasicNameValuePair(AIT_Param_StartTime, startTime)); 
	    params.add(new BasicNameValuePair(AIT_Param_TimeLimit, Integer.toString(timeLimit))); 
	    for (Map.Entry<String, String> entry : optionsMap.entrySet()) {
		    params.add(new BasicNameValuePair(entry.getKey(), entry.getValue())); 		
	    }
	    
	    Isochrone iso = new Isochrone();
	    
	    // Call AIT Web-Service.
	    if(IsoAsWkt) {
		// AIT_Server_Path_GetIso: Important change from GML to WKT (Iso about 50% smaller)
		String isoWKT = Util.httpGetRequest(mAIT_Server, AIT_Port, AIT_Server_Path_GetIsoWkt, params); 
		    iso.fillDetailsFromWKT(isoWKT, mGeometryFactory);
	    }
	    else {
		String isoXML = Util.httpGetRequest(mAIT_Server, AIT_Port, AIT_Server_Path_GetIso, params); 
		iso.fillDetailsFromGML(isoXML);
	    }
		
	    return iso.getElementsAsArray();
	    
    }
    

    public void storeIsochroneDetails(String sessionId, Geometry[] geo, GeometryFactory geoFactory, int complexIsochrone, boolean log) {

	ArrayList<Polygon> polys = new ArrayList<Polygon>();
	for (int i = 0; i < geo.length; i++) {
	    if(geo[i] instanceof Polygon) polys.add( (Polygon) geo[i]);
	}

	SimpleGMLWriter w = new SimpleGMLWriter();
	String gmlText = null;
	
	// Two sorts of isochrone can be returned, either a simple (convex hull) or the real one
	if(complexIsochrone==0) {
		GeometryCollection multi = new GeometryCollection(polys.toArray(new Polygon[polys.size()]), geoFactory);
		ConvexHull cv = new ConvexHull(multi);
		Geometry g2 = cv.getConvexHull();
		gmlText = w.writeIsochrone(g2);
	}
	else if(complexIsochrone==1) {
		GeometryCollection multi = new GeometryCollection(polys.toArray(new Polygon[polys.size()]), geoFactory);
		ConcaveHull cv = new ConcaveHull(multi, 0.002); // 0.05 0.19
		Geometry g2 = cv.getConcaveHull();
		gmlText = w.writeIsochrone(g2);
		
		// Write Log
		if (log) {
		    Geometry[] g2s = new Geometry[1];
		    g2s[0] = g2;
		    SimpleGMLWriter gmlw = new SimpleGMLWriter();
		    gmlw.writeGeometry(g2s, "log/SimpleIso_" + sessionId + ".gml");
		}
	}

	else if(complexIsochrone==2) {
		MultiPolygon g2 = new MultiPolygon(polys.toArray(new Polygon[polys.size()]), geoFactory);
		gmlText = w.writeIsochrone(g2);
	}
	

	
	mJedis.set(JedisIsoNames + sessionId, gmlText);
    }
    
    public String getIsochroneDetails(String sessionID) {	

	if(! mJedis.exists(JedisIsoNames + sessionID))  return null;
	
	String isos = mJedis.get(JedisIsoNames + sessionID);
	return isos;
	
    }
   
 
    private String getActualTime() {
	Calendar cal1 = Calendar.getInstance();
	// return String.format("%1$tY-%1$tm-%1$teT%1$tH:%1$tM:%1$tS", new Object[] {cal1.getTime()}); // %n
	SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd");
	SimpleDateFormat timeFormat = new SimpleDateFormat("HH:mm:ss");

	String date = dateFormat.format(cal1.getTime());
	String time = timeFormat.format(cal1.getTime());
	return date + "T" + time;
    }
    
    public void  clearRouteCache() {	
	Set<String> keys = mJedis.keys(JedisRoutesNames + "*");		
	if(keys.size() != 0) 
	    mJedis.del(keys.toArray(new String[keys.size()]));	

	Set<String> keys2 = mJedis.keys(JedisIsoNames + "*");	
	if(keys2.size() != 0) 
	    mJedis.del(keys2.toArray(new String[keys2.size()]));	
 
    
    }
    
    public String getRouteDetails(String sessionID, String routeID) {	

	if(! mJedis.hexists(JedisRoutesNames + sessionID, routeID))  return null;
	
	List<String> uris = mJedis.hmget(JedisRoutesNames + sessionID, routeID);
	return uris.get(0);
	
    }
 
    public RouteResults  getEmptyRoutes(String id, Point start,  String mots, String startTime, int timeLimit, String kwds)
    {
	if(startTime==null || startTime.length() == 0)  startTime = getActualTime();

	return new RouteResults(start, mots, timeLimit, startTime, id, kwds);

    }
	
    public RouteResults  getRoutes(String id, Point start, ArrayList<PointOfInterest> pois, int resultLimit, 
	    String mots, String startTime, int timeLimit, boolean testData, boolean logData, String kwds, String userId, Map<String, String> optionsMap)
    {
	if(startTime==null || startTime.length() == 0)  startTime = getActualTime();

	RouteResults routes = new RouteResults(start, mots, timeLimit, startTime, id, kwds);

	
	if(testData) {
	    
	    // Temporary work around, after direct call to webservice
	    SimpleGMLReader reader = new SimpleGMLReader();
	    ArrayList<Route> rs = reader.loadAITRoutes(); 
	    
	    // Assign POIs to routes
	    for (int s = 0; s < rs.size(); s++) {
		Route r = rs.get(s);
		r.assignPOI(pois);		
		routes.addRoute(r);
	    }
 
	}
	else {
	    StringBuffer pointsText = new StringBuffer();
	    int count = 0;
	    //Call of all POIs together 
	    for (int i = 0; i < pois.size(); i++) {		
		if(resultLimit <= i) break;		
		FeatureOfInterest poi = pois.get(i);
		
		if(i>0) pointsText.append(";");
		pointsText.append(poi.toCoord(true));
		count = i;
	    }
	    
	    ArrayList<NameValuePair> params = createAITServiceParameter(userId, GeoUtil.pointToCoord(start, true, ','), pointsText.toString(), mots, startTime, optionsMap);
	    // Generic User.  Use actual user from UI and Acutal time: Use time from UI
		    
	    // Call AIT Web-Service. 
	    String tempXML = Util.httpGetRequest(mAIT_Server, AIT_Port, AIT_Server_Path_GetRoutes, params);
	    
	    String routeXMLs[] = tempXML.split("(" + AIT_RouteSeparator +  ")");
	    
	    
	    // Important: The results from the AIT Server come in the same order as the they are in the input param. 
	    // Therefore we just match them simple by index (and do not search for id or cooridnates).

	    if(routeXMLs.length < count) { // Temporary fix, cause of AIT Server problems
		System.out.println("Fallback: Call each POI individual");
		routeXMLs = new String[pois.size()];
		for (int i = 0; i < pois.size(); i++) {
		    PointOfInterest poi = pois.get(i);
		    ArrayList<NameValuePair> params2 = createAITServiceParameter(userId, GeoUtil.pointToCoord(start, true, ','), poi.toCoord(true), mots, startTime, optionsMap);
		    routeXMLs[i] = Util.httpGetRequest(mAIT_Server, AIT_Port, AIT_Server_Path_GetRoutes, params2);
		}
	    }
	    
	    
	    for (int i = 0; i < pois.size(); i++) {		
		if(resultLimit <= i) break;		
		PointOfInterest poi = pois.get(i);
		String routeXML = routeXMLs[i];
		
		// Create and complete route instance
		RouteNR route = new RouteNR();
		boolean routeOK = route.fillDetails(routeXML, Edge_TO, poi.Name);
		if(!routeOK) 
		    continue; // If serialization problems ignore POI and route
		
		if(route.Id == null) {		    
		    route.Id = Integer.toString(mRanGen.nextInt(Integer.MAX_VALUE)); // route.Id = UUID.randomUUID().toString();
		}
		route.setPOI(poi);		
		routes.addRoute(route);
		
		// Store for details
		mJedis.hset(JedisRoutesNames + routes.getId(), route.Id, routeXML);
		
		if(logData) {
		    try {
			Util.setContents("log/Route_" + route.Id  + ".xml", routeXML);
		    } catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		    } 
		}

	    }
	    
	}
	
	return routes;
	
    }
    
    private ArrayList<NameValuePair> createAITServiceParameter(String userID, String from, String to, String mots, String time, Map<String, String> optionsMap) {
	
	ArrayList<NameValuePair> params = new ArrayList<NameValuePair>();
	params.add(new BasicNameValuePair(AIT_Param_User, userID)); 
	params.add(new BasicNameValuePair(AIT_Param_From, from)); 
	params.add(new BasicNameValuePair(AIT_Param_To, to)); 
	params.add(new BasicNameValuePair(AIT_Param_MOT, mots)); 
	params.add(new BasicNameValuePair(AIT_Param_StartTime, time));  
	    for (Map.Entry<String, String> entry : optionsMap.entrySet()) {
		    params.add(new BasicNameValuePair(entry.getKey(), entry.getValue())); 		
	    }

	return params;
   }
    
    public RouteResults  getViaRoutes(String id, Point from, Point to, ArrayList<PointOfInterest> pois, int resultLimit, String mots, 
	    String startTime, int timeLimit, boolean testData, boolean logData, String kwds, String userId, Map<String, String> optionsMap)
    {
	if(startTime==null || startTime.length() == 0)  startTime = getActualTime();
	
	RouteResults routes = new RouteResults(from, to, mots, timeLimit, startTime, id, kwds);


//	if(testData) {
//	    // Temporary work around, after direct call to webservice
//	    SimpleGMLReader reader = new SimpleGMLReader();
//	    ArrayList<Route> rs = reader.loadViaAITRoutes(); 
//	    
//	    // Assign POIs to routes
//	    for (int s = 0; s < rs.size(); s++) {
//		Route r = rs.get(s);
//		r.assignPOI(pois);		
//		routes.addRoute(r);
//	    }
//
//	
//	}
//	else {
//	}

	    // Call AIT Web-Service
	    StringBuffer pointsText = new StringBuffer();

	    //First implementation call for each POI
	    for (int i = 0; i < pois.size(); i++) {		
		if(resultLimit <= i) continue;
		PointOfInterest poi = pois.get(i);
				
		if(i>0) pointsText.append(";");
		pointsText.append(poi.toCoord(true));
	    }
	    
	    ArrayList<NameValuePair> params = createAITServiceParameter(userId, 
			GeoUtil.pointToCoord(from, true, ','), GeoUtil.pointToCoord(to, true, ','), mots, startTime, optionsMap);
		params.add(new BasicNameValuePair(AIT_Param_Via, pointsText.toString())); // poi.toCoord(true)
		    
	    // Call AIT Web-Service. 
	    String tempXML = Util.httpGetRequest(mAIT_Server, AIT_Port, AIT_Server_Path_GetViaRoutes, params);

	    String routeXMLs[] = tempXML.split("(" + AIT_RouteSeparator +  ")");
	
	    for (int i = 0; i < pois.size(); i++) {		
		if(resultLimit <= i) break;		
		PointOfInterest poi = pois.get(i);
		String routeXML = routeXMLs[i];
		
		RouteVia route = new RouteVia();

		// Create and complete route instance
		route.fillDetails(routeXML, Edge_VIA, poi.Name);
		if(route.Id == null) {		    
		    route.Id = Integer.toString(mRanGen.nextInt(Integer.MAX_VALUE)); 
		}
		route.setPOI(poi);		
		routes.addRoute(route);

		if(logData) {
		    try {
			Util.setContents("log/ViaRoute_" + route.Id  + ".xml", routeXML);
		    } catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		    } 
		}


		
		// Store for details
		mJedis.hset(JedisRoutesNames + routes.getId(), route.Id, routeXML);
	    }

	    
	return routes;
	
    }
    
    public RouteResults  getDirectRoutes(String id, Point from, Point to, String mots, 
	    String startTime, int timeLimit, boolean testData, boolean logData, String userId, Map<String, String> optionsMap)
    {
	if(startTime==null || startTime.length() == 0)  startTime = getActualTime();
	
	RouteResults routes = new RouteResults(from, to, mots, timeLimit,
		startTime, id, "");

	ArrayList<NameValuePair> params = createAITServiceParameter(userId,
		GeoUtil.pointToCoord(from, true, ','),
		GeoUtil.pointToCoord(to, true, ','), mots, startTime, optionsMap);
		    
	// Call AIT Web-Service. 
	String routeXML = Util.httpGetRequest(mAIT_Server, AIT_Port, AIT_Server_Path_GetRoutes, params);
		
	RouteVia route = new RouteVia();

	if (logData) {
	    try {
		Util.setContents("log/ViaRoute_" + route.Id + ".xml", routeXML);
	    } catch (Exception e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	    }
	}

	// Create and complete route instance
	route.fillDetails(routeXML, Edge_VIA, null);
	if (route.Id == null)
	    route.Id = Integer.toString(mRanGen.nextInt(Integer.MAX_VALUE));

	// route.setPOI(new PointOfInterest());
	routes.addRoute(route);

	// Store for details
	mJedis.hset(JedisRoutesNames + routes.getId(), route.Id, routeXML);

	return routes;
	
    }
    
    
    
}
