package edu.tuwien.kbs.search;



import java.util.*;
import org.javatuples.*;
import com.clarkparsia.owlgres.cli.queryRewrite;
import com.vividsolutions.jts.geom.*;

import edu.tuwien.kbs.keywords.*;
import edu.tuwien.kbs.util.*;


public class QueryRewritter {

    // Interesting fact, that Rewriting and eval changes drastically with 
    // chosen BaseAttribute and BaseFeature. From 790ms to 561ms (Rew.) and 2100ms to 1350ms (Eval.)
    public static String BaseConceptAttributes = "geot:QualitativeValue"; // "geot:Attributes";
    public static String BaseObjRoleAttributes = "geot:hasQualitativeValue";
    public static String BaseConceptBuilding = "geot:BuildingFeature"; 
    public static String BaseConceptFeature = "geot:_Feature"; // "geot:_Feature";
    public static String BaseConceptGeometries = "geot:_Geometry";
    public static String BaseConceptEvent = "geot:Event";
    
    public static String OwlgresViewNameA = "temp_query_a_"; // "temp_og_query_a"
    public static String OwlgresViewNameB = "temp_query_b_";
    public static String OwlgresViewNameC = "temp_query_c_";
    public static String OwlgresUnionsNameA = "temp_queries_a_"; // temp_og_queries_a
    public static String OwlgresUnionsNameB = "temp_queries_b_";
    public static String OwlgresUnionsNameC = "temp_queries_c_";

    public static String OwlgresViewField1 = "att1";
    public static String OwlgresViewField2 = "att2";
    
    private static boolean OptimizeBaseConcept = true; // Turn on off for testing

    protected String m_OutputGeometry;
    protected QueryTemplates m_QueryTemplates;
    protected SchemaInfo.SchemaInfoItem mSchemaA;
    protected SchemaInfo.SchemaInfoItem mSchemaB;
    protected SchemaInfo.SchemaInfoItem mSchemaC;
    protected SchemaInfo.SchemaInfoItem mSchemaEventA;
    //protected SchemaInfo.SchemaInfoItem mSchemaEventB;
    
    // From properties
    protected String mHost;
    protected String mDB;
    protected String mUser;
    protected String mPw;

    protected KeywordGenerator kwGen;

    public QueryRewritter() {
	
    }
    
    public QueryRewritter(String outputGeometry, String host, String db, String user, String pw) {
	this();
	
	mHost = host; mDB = db; mUser = user; mPw = pw;
	m_OutputGeometry =  outputGeometry;
	
	try {
	    m_QueryTemplates = new QueryTemplates();
	    m_QueryTemplates.load();
	    kwGen = new KeywordGenerator();
	} catch (Exception e) {
	    // TODO Auto-generated catch block
	    e.printStackTrace();
	}    

    }
    
    public SchemaInfo.SchemaInfoItem GetInfoItemA() {
	return mSchemaA;
    } 
    public SchemaInfo.SchemaInfoItem GetInfoItemB() {
	return mSchemaB;
    }
    public SchemaInfo.SchemaInfoItem GetInfoItemEventA() {
	return mSchemaEventA;
    }


    public Septet<String, String, String, String, String, String, String> createKBQueryFromCQ(String inputQuery, SchemaInfo schema, 
		ArrayList<String> temporaryViews, String sessionID) throws Exception {
    
	Septet<String, String, String, String, String, String, String>  returns = null; // String[] 
 	StringBuffer unionViewNamesA = new StringBuffer(); 
	StringBuffer unionViewNamesB = new StringBuffer();
	boolean multipleViews = false;
	String fixedKeyword1 = null;

	// Create Query from set of PREDICATES,
	    
	String[] queries = inputQuery.split(";");

	for (int i = 0; i < queries.length; i++) {
	    // For prototype it is just possible to create 2 Points Within
	    // Queries
	    String[] s = queries[i].split("(;" + SpatialPredicatesMyITS.KeywordQueryNext + ";)");
	    String query1 = s[0].trim();
	    String uris1[] = query1.split("[ ]");
	    mSchemaA = schema.getInfoItem(uris1,kwGen); // uris1[0]

	    if (unionViewNamesA.length() > 0)  unionViewNamesA.append(" UNION ");
	    if (unionViewNamesB.length() > 0) unionViewNamesB.append(" UNION ");

	    String retView1 = createDLViews(uris1, mSchemaA, i, OwlgresViewNameA + sessionID);
	    temporaryViews.add(retView1);

	    if (retView1 != null)
		unionViewNamesA.append("(SELECT * FROM " + retView1 + ")");

	    //	Let Owlgres rewrite the second query ... TODO: Any Nrs of queries
	    if (s.length > 1) {
		String query2 = s[1].trim();
		String uris2[] = query2.split("[ ]");
		mSchemaB = schema.getInfoItem(uris2, kwGen); // uris2[0]

		String retView2 = createDLViews(uris2, mSchemaB, i, OwlgresViewNameB + sessionID);
		temporaryViews.add(retView2);
		multipleViews = true;

		if (retView2 != null) unionViewNamesB.append("(SELECT * FROM " + retView2 + ")");
	    }
	}

	String sqlA = "CREATE OR REPLACE VIEW " + "owl." + OwlgresUnionsNameA + sessionID + " AS " + unionViewNamesA.toString();

	// Create return Strings
	if(multipleViews) {
	    String sqlB = "CREATE OR REPLACE VIEW " + "owl." + OwlgresUnionsNameB + sessionID + " AS " + unionViewNamesB.toString();
	    returns = new Septet<String,String,String,String,String,String,String>(sqlA, sqlB, null, fixedKeyword1, null, null,null); 
	    
	}
	else {
	    returns = new Septet<String,String,String,String,String, String,String>(sqlA, null, null, null, null, null,null); 	    
	}

	return returns; 
	    
    
    }
    
    // sqlA, sqlB, (sqlC), fixedKeywordA, (fixedKeywordB), keywordText,uriEvents
    public Septet<String, String, String, String, String, String, String> createKBQueryFromKeywords(String inputQuery, SchemaInfo schema, 
	    							ArrayList<String> temporaryViews, String sessionID) throws Exception {
    	
	Septet<String, String, String, String, String, String, String>  returns = null; // String[] 
    	//String[] queries = null;
	StringBuffer unionViewNamesA = new StringBuffer();  StringBuffer unionViewNamesB = new StringBuffer(); StringBuffer unionViewNamesC = new StringBuffer();
	boolean multipleViews = false;
	String fixedKeyword1 = null; String fixedKeyword2 = null;
	String keywordText = null;
	String uriEvents = null;
	
	// Create Query from set of KEYWORDS

	    	    
	    String[] inSplit = inputQuery.split(";"); 	
	    KeywordResults kwds = kwGen.getKeywords(inSplit, 20, false); // Old:1, choose 20 to find the best match regarding the next word 
	    String[] uris = kwGen.getUrisFromSearchStrings(kwds);
	    ArrayList<String> uriA = new ArrayList<String>();
	    ArrayList<String> uriB = new ArrayList<String>();
	    ArrayList<String> uriC = new ArrayList<String>();
	    
	    
	    keywordText = kwds.toString();
	    
	    // Top Level of parse tree are spatial predicates Next, Near, In. TODO: Any Nrs of spatial predicates 
	    for (int i = 0; i < uris.length; i++) {
		    if(SpatialPredicatesMyITS.hasSpatialPredicate(uris[i])) {
			if(fixedKeyword1 == null)
			    fixedKeyword1 = uris[i];
			else
			    fixedKeyword2 = uris[i];
			continue;
		    }
		    if(fixedKeyword1==null) {
			if (kwGen.checkConceptInclusion(uris[i], BaseConceptEvent)) 
			    uriEvents = uris[i]; // Special case events, also need a location related to event (default BaseConceptFeature)
			uriA.add(uris[i]); // LHS
		    }
		    else  
			if(fixedKeyword2==null)
			    uriB.add(uris[i]); // RHS
			else 
			    uriC.add(uris[i]); // RHS

	    }

	    // Handle for now just to size 4 (8 if spatial predicate)
	    if(uriA.size() > 5 || uriB.size() > 5) {
		Exception en = new Exception("Up to 5 keywords are allowed.");
		throw en;		
	    }
	    	    
	    //  LHS of the spatial pred. (For now just consider events on the LHS
	    if (uriA.size() > 0) {
		String[] tempA = uriA.toArray(new String[uriA.size()]);
		if(uriEvents!=null) {
		    mSchemaEventA = schema.getInfoItem(new String[]{uriEvents}, kwGen);
		    mSchemaA = schema.getInfoItem(new String[]{""}, kwGen); // Means defaut schema: TODO: Dynamic
		}
		else {
		    mSchemaA = schema.getInfoItem(tempA, kwGen); // uriA.get(0)
		}
		
		// Let Owlgres rewrite the first query ...
		String retView1 = createDLViews(tempA, mSchemaA, 0, OwlgresViewNameA + sessionID); 
		temporaryViews.add(retView1);
		if (retView1 != null)  unionViewNamesA.append("(SELECT * FROM " + retView1 + ")");
	    }
	    else {
		return new Septet<String,String,String,String,String, String,String>(null,null,null,null,null,keywordText,null);
	    }
	    
	    // RHS of the spatial pred. A
	    if (fixedKeyword1 != null && uriB.size() > 0) {
		multipleViews = true;
		// Level 2b
		String[] tempB = uriB.toArray(new String[uriB.size()]);
		mSchemaB= schema.getInfoItem(tempB, kwGen); // uriB.get(0)
		String retView2 = createDLViews(tempB, mSchemaB, 0, OwlgresViewNameB + sessionID); // Let Owlgres rewrite the second query ...
		temporaryViews.add(retView2);
		if (retView2 != null)  unionViewNamesB.append("(SELECT * FROM " + retView2 + ")");
		
		// RHS of the spatial pred. B
		if (fixedKeyword2 != null && uriC.size() > 0) {
			// Level 3b
			String[] tempC = uriC.toArray(new String[uriC.size()]);
			mSchemaC= schema.getInfoItem(tempC, kwGen); // uriB.get(0)
			String retView3 = createDLViews(tempC, mSchemaC, 0, OwlgresViewNameC + sessionID); // Let Owlgres rewrite the second query ...
			temporaryViews.add(retView3);
			if (retView3 != null)  unionViewNamesC.append("(SELECT * FROM " + retView3 + ")");
		 }
	    }
	    

	//}
	
	String sqlA = "CREATE OR REPLACE VIEW " + "owl." + OwlgresUnionsNameA + sessionID + " AS " + unionViewNamesA.toString();
  	
    	// Create return Strings
    	if(multipleViews) {
	    String sqlB = "CREATE OR REPLACE VIEW " + "owl." + OwlgresUnionsNameB + sessionID + " AS " + unionViewNamesB.toString();
	    String sqlC = null;
	    if(fixedKeyword2 != null)
	    	sqlC = "CREATE OR REPLACE VIEW " + "owl." + OwlgresUnionsNameC + sessionID + " AS " + unionViewNamesC.toString();

	    returns = new Septet<String,String,String,String,String,String,String>(sqlA, sqlB, sqlC, fixedKeyword1, fixedKeyword2, keywordText,uriEvents); 
    	}
    	else {
    	    returns = new Septet<String,String,String,String,String, String,String>(sqlA, null, null, null, null, keywordText,uriEvents); 	    
    	}
    	    
    	return returns; 
    }
    
    private String createDLViews(String[] keywords, SchemaInfo.SchemaInfoItem infoItem, int index,  String owlgresViewName) { 

	String owlgresQuery = null;
	String viewName = null;
	int columns = 1;
	
	// Check if all concepts are subsumed by BaseClassAttributes
	// boolean hasSpatialFeature = false;
	int hasSpatialFeature = -1;

	// There is always one keywords, check if the keywords is a concept or individual (and get the related concept for indv.).
	String[] assignedClass =  new String[keywords.length];
	boolean[] objIsInd =  new boolean[keywords.length];
	
	assignedClass[0] = keywords[0];
	objIsInd[0] = (keywords[0].contains(SemSearch.DefaultNamespacePrefixABox));
	if(objIsInd[0]) assignedClass[0] = kwGen.getAssertedConept(keywords[0]);
	if (kwGen.checkConceptInclusion(assignedClass[0], BaseConceptFeature)) hasSpatialFeature = 0; //hasSpatialFeature = true;

	// Set default to main concepts and roles
	String subjConcept = BaseConceptBuilding;
	String predRole1 = BaseObjRoleAttributes; String predRole2 = BaseObjRoleAttributes; 
	String predRole3 = BaseObjRoleAttributes; String predRole4 = BaseObjRoleAttributes;
	
	if (keywords.length >= 2) {	    
	    objIsInd[1] = (keywords[1].contains(SemSearch.DefaultNamespacePrefixABox));
	    assignedClass[1] = keywords[1];
	    if(objIsInd[1]) assignedClass[1] = kwGen.getAssertedConept(keywords[1]);
	    if (kwGen.checkConceptInclusion(assignedClass[1], BaseConceptFeature)) hasSpatialFeature = 1; // hasSpatialFeature = true;
	}	
	if (keywords.length >= 3) {
	    objIsInd[2] = (keywords[2].contains(SemSearch.DefaultNamespacePrefixABox));
	    assignedClass[2] = keywords[2];
	    if(objIsInd[2]) assignedClass[2] = kwGen.getAssertedConept(keywords[2]);
	    if (kwGen.checkConceptInclusion(assignedClass[2], BaseConceptFeature)) hasSpatialFeature = 2; //hasSpatialFeature = true;
	}	
	if (keywords.length >= 4) { 
	    objIsInd[3] = (keywords[3].contains(SemSearch.DefaultNamespacePrefixABox));
	    assignedClass[3] = keywords[3];
	    if(objIsInd[3]) assignedClass[3] = kwGen.getAssertedConept(keywords[3]);
	    if (kwGen.checkConceptInclusion(assignedClass[3], BaseConceptFeature)) hasSpatialFeature = 3; //hasSpatialFeature = true;
	}
	if (keywords.length >= 5) { 
	    objIsInd[4] = (keywords[4].contains(SemSearch.DefaultNamespacePrefixABox));
	    assignedClass[4] = keywords[4];
	    if(objIsInd[4]) assignedClass[4] = kwGen.getAssertedConept(keywords[3]);
	    if (kwGen.checkConceptInclusion(assignedClass[4], BaseConceptFeature)) hasSpatialFeature = 4; //hasSpatialFeature = true;
	}
	
	// Move SpatialFeature to front
	if (hasSpatialFeature > 0) {
	    String saveC = assignedClass[0];
	    String saveK = keywords[0];
	    boolean saveH = objIsInd[0];
	    assignedClass[0] = assignedClass[hasSpatialFeature];
	    keywords[0] = keywords[hasSpatialFeature];
	    objIsInd[0] = objIsInd[hasSpatialFeature];
	    assignedClass[hasSpatialFeature] = saveC;
	    keywords[hasSpatialFeature] = saveK;
	    objIsInd[hasSpatialFeature] = saveH;
	}
	
	// Choose templates according to Nr. of keywords. TODO: Make this more generic and create template on demand
	
	// SpatialFeature (SF) or QualitativeAttribute (QA)
	if (keywords.length == 1) {

	    // Case we have an QualitativeAttribute, then, we create a CQ of the SpatialFeature and the QualitativeAttribute connect both.
	    if (hasSpatialFeature == -1) {
		// Optimization step: Decrease the size of rewritten conj. queries with Owlgres
		if(OptimizeBaseConcept) {
		    SimpleConceptRoleLevel predRole_domain = optimizeSpatialFeatureForObject(assignedClass[0]); //  Pair<String, String>
		    if(predRole_domain != null ) {
			predRole1 = predRole_domain.RoleUri; // predRole_domain.getValue0();
			subjConcept = predRole_domain.ConceptUri; // predRole_domain.getValue1();			
		    }
		}
		owlgresQuery = m_QueryTemplates.fillTwoTerms(objIsInd[0], subjConcept, predRole1, keywords[0]);
		if(!objIsInd[0]) columns = 2;
	    }
	    // Case when we have a SpatialFeature, then build just the query for it (e.g. Cafe, Restaurant, Park, ...)
	    else {
		owlgresQuery = m_QueryTemplates.fillOneTerm(keywords[0]);
	    }	    
	    
	// (SF and QA) or (QA and QA)
	} else if (keywords.length == 2) {
	    // (QA and QA)
	    if(hasSpatialFeature == -1) {
		
		Pair<String, String> params = rearangeKeywords(objIsInd[0], objIsInd[1], keywords, 0);		
		    
		if (OptimizeBaseConcept) {
		    Triplet<String, String, String> subjPredA_B = optimizeSpatialFeatureForTwoRoles(assignedClass[0], assignedClass[1]);
		    if (subjPredA_B != null) {
			subjConcept = subjPredA_B.getValue0();
			predRole1 = subjPredA_B.getValue1();
			predRole2 = subjPredA_B.getValue2();
		    }
		}
		owlgresQuery = m_QueryTemplates.fillThreeTermsTwoPred( objIsInd[0], objIsInd[1], subjConcept, params.getValue0(),
			params.getValue1(), predRole1, predRole2);
		
	    }
	    else {
		// 1. Keyword is a SF and 2. Keyword is a QA, optimize find subject role
		if(OptimizeBaseConcept) {
		    String role = optimizeSpatialFeatureForSubjectObject(assignedClass[0], assignedClass[1]);
		    if(role != null ) predRole1 =role;
		}

		owlgresQuery = m_QueryTemplates.fillTwoTerms(objIsInd[1],  keywords[0], predRole1, keywords[1]);
		if(!objIsInd[1]) columns = 2;
	    }
	    
	// (SF, QA and QA) or (QA, QA, and QA)    
	} else if (keywords.length == 3) {
	    	    
	    if( hasSpatialFeature == -1) {

		// (QA, QA, and QA)   
		Triplet<String, String, String> params = rearangeKeywords(objIsInd[0], objIsInd[1],  objIsInd[2], keywords, 0);	   
		
		if(OptimizeBaseConcept) {
		    
		    Quartet<String, String, String, String> subjPredA_B_C = optimizeSpatialFeatureForThreeRoles(assignedClass[0], assignedClass[1], assignedClass[2]);
		    if (subjPredA_B_C != null) {
			subjConcept = subjPredA_B_C.getValue0();
			predRole1 = subjPredA_B_C.getValue1(); predRole2 = subjPredA_B_C.getValue2(); 
			predRole3 = subjPredA_B_C.getValue3();
		    }
		}
		
		owlgresQuery = m_QueryTemplates.fillFourTermsNoPred(objIsInd[0], objIsInd[1],  objIsInd[2], subjConcept, params.getValue0(), params.getValue1(), 
			 params.getValue2(), predRole1, predRole2, predRole3);
		
	    }
	    else {
		
		Pair<String, String> params = rearangeKeywords(objIsInd[1], objIsInd[2], keywords, 1);		
		if(OptimizeBaseConcept) {
		    String role1 = optimizeSpatialFeatureForSubjectObject(assignedClass[0], assignedClass[1]);
		    if(role1 != null ) predRole1 =role1;
		    String role2 = optimizeSpatialFeatureForSubjectObject(assignedClass[0], assignedClass[2]);
		    if(role2 != null ) predRole2 =role2;
		}
	    
		owlgresQuery = m_QueryTemplates.fillThreeTermsTwoPred(objIsInd[1], objIsInd[2], keywords[0], params.getValue0(), params.getValue1(), predRole1, predRole2);
	    }
	 // (SF, QA, QA, QA and QA) 
	} else if (keywords.length == 4) {

	    if(hasSpatialFeature == -1) {
	    
		// (QA, QA, QA and QA)   
		Triplet<String, String, String> paramsTemp = rearangeKeywords(objIsInd[0], objIsInd[1],  objIsInd[2], keywords, 0); 
		Quartet<String, String, String, String> params = new Quartet(paramsTemp.getValue0(), paramsTemp.getValue1(), paramsTemp.getValue2(), keywords[3]); 
		
		if(OptimizeBaseConcept) {
		    
		    Quintet<String, String, String, String, String> subjPredA_B_C = optimizeSpatialFeatureForFourRoles(assignedClass[0], assignedClass[1], assignedClass[2], assignedClass[3]);
		    if (subjPredA_B_C != null) {
			subjConcept = subjPredA_B_C.getValue0();
			predRole1 = subjPredA_B_C.getValue1(); predRole2 = subjPredA_B_C.getValue2(); 
			predRole3 = subjPredA_B_C.getValue3(); predRole4 = subjPredA_B_C.getValue4();
		    }
		}
		
		owlgresQuery = m_QueryTemplates.fillFiveTermsNoPred(objIsInd[0], objIsInd[1], objIsInd[2], objIsInd[3], subjConcept, params.getValue0(), params.getValue1(), 
			 params.getValue2(), params.getValue3(), predRole1, predRole2, predRole3, predRole4);

	    }
	    else {
		Triplet<String, String, String> params = rearangeKeywords(objIsInd[1], objIsInd[2],  objIsInd[3], keywords, 1);
		
		if(OptimizeBaseConcept) {
		    String role1 = optimizeSpatialFeatureForSubjectObject( assignedClass[0], assignedClass[1]);
		    if (role1 != null) predRole1 = role1;
		    String role2 = optimizeSpatialFeatureForSubjectObject( assignedClass[0], assignedClass[2]);
		    if (role2 != null) predRole2 = role2;
		    String role3 = optimizeSpatialFeatureForSubjectObject( assignedClass[0], assignedClass[3]);
		    if (role3 != null) predRole3 = role3;
		}

		owlgresQuery = m_QueryTemplates.fillFourTermsNoPred(objIsInd[1], objIsInd[2],  objIsInd[3], keywords[0], params.getValue0(), params.getValue1(), 
		    params.getValue2(), predRole1, predRole2, predRole3);
	    }
	    
	} else if (keywords.length == 5) {
	    
	    
	}
	
	// Call Owlgres
	if (owlgresQuery != null) {
	    viewName = owlgresViewName + index;
	    callOwlgres(owlgresQuery, viewName, columns);
	    return viewName;
	}

	return null;

    }

    private Pair<String, String> rearangeKeywords(boolean obj1IsInd, boolean obj2IsInd,  String[] keywords, int startIndex) {

	if(!obj1IsInd && obj2IsInd) {
	    return new Pair<String, String>(keywords[startIndex+1], keywords[startIndex]);
	}
	else {
	    return new Pair<String, String>(keywords[startIndex], keywords[startIndex+1]);
	}
	
    }
    
    private Triplet<String, String, String> rearangeKeywords(boolean obj1IsInd, boolean obj2IsInd, boolean obj3IsInd, String[] keywords , int startIndex) {
	    
	Triplet<String, String, String> ret = new Triplet<String, String, String>(keywords[startIndex], keywords[startIndex+1], keywords[startIndex+2]);
	    
	    if(!obj1IsInd && obj2IsInd && obj3IsInd) { // CII
		ret = new Triplet<String, String, String>(keywords[startIndex+1], keywords[startIndex+2], keywords[startIndex]);
	    }
	    else if(obj1IsInd && !obj2IsInd && obj3IsInd) { // ICI
		ret = new Triplet<String, String, String>(keywords[startIndex+1], keywords[startIndex], keywords[startIndex+2]);
	    }
	    if(!obj1IsInd && !obj2IsInd && obj3IsInd) { // CCI
		ret = new Triplet<String, String, String>(keywords[startIndex+2], keywords[startIndex], keywords[startIndex+1]);
	    }
	    else if(!obj1IsInd && obj2IsInd && !obj3IsInd) { // CIC
		ret = new Triplet<String, String, String>(keywords[startIndex+1], keywords[startIndex], keywords[startIndex+2]);
	    }
	    return ret;
    }


    private Triplet<String, String, String> optimizeSpatialFeatureForTwoRoles(String concept1Uri, String concept2Uri)
    {
	return optimizeSpatialFeatureForTwoRoles(kwGen, concept1Uri, concept2Uri);
    }
    
    // Given two object concepts, find a subject concept and two predicate roles.
    public Triplet<String, String, String> optimizeSpatialFeatureForTwoRoles(KeywordGenerator keywordGen, String concept1Uri, String concept2Uri)
    {
	
	ArrayList<SimpleConceptRoleLevel> results1 =  keywordGen.getSortedDomainObjRolesForRange(concept1Uri);
	ArrayList<SimpleConceptRoleLevel> results2 =  keywordGen.getSortedDomainObjRolesForRange(concept2Uri);

	// Simple heuristics: Take the two objectRoles which have the lowest level and 
	// C_1 subsumes C_2 (or opposite). Take the more specific (C_1).
	if(results1.size() > 0 && results2.size() > 0) {
	    
	    for(SimpleConceptRoleLevel roleDomLevel1: results1) {		
		for(SimpleConceptRoleLevel roleDomLevel2: results2) {
		    String subjectConcept = null;
		    if(roleDomLevel1.ConceptUri.equals(roleDomLevel2.ConceptUri)) {
			subjectConcept = roleDomLevel1.ConceptUri;
		    } else if(keywordGen.checkConceptInclusion(roleDomLevel1.ConceptUri, roleDomLevel2.ConceptUri)) {
			subjectConcept = roleDomLevel1.ConceptUri;
		    } else if(keywordGen.checkConceptInclusion(roleDomLevel2.ConceptUri, roleDomLevel1.ConceptUri)) {
			subjectConcept = roleDomLevel2.ConceptUri;
		    }
		    
		    if(subjectConcept != null) {
			return new Triplet<String, String, String>(subjectConcept, roleDomLevel1.RoleUri, roleDomLevel2.RoleUri);
		    }		    
		}				
	    }	    
	}	
	
	return null;	
    }    
    
    // Given three concepts, find a subject, two concept and two roles.
    private Quartet<String, String, String, String> optimizeSpatialFeatureForThreeRoles(String object1Uri, String object2Uri, 
	    String object3Uri)
    {
	
	ArrayList<SimpleConceptRoleLevel> results1 =  kwGen.getSortedDomainObjRolesForRange(object1Uri);
	ArrayList<SimpleConceptRoleLevel> results2 =  kwGen.getSortedDomainObjRolesForRange(object2Uri);
	ArrayList<SimpleConceptRoleLevel> results3 =  kwGen.getSortedDomainObjRolesForRange(object3Uri);

	// Simple heuristics: Take the three roles which have the lowest level and 
	// C_1 subsumes C_2 (or opposite). Take the more specific (C_1).
	if(results1.size() > 0 && results2.size() > 0 && results3.size() > 0) {
	    
	    for (SimpleConceptRoleLevel roleDomLevel1 : results1) {
		for (SimpleConceptRoleLevel roleDomLevel2 : results2) {
		    for (SimpleConceptRoleLevel roleDomLevel3 : results3) {

			String subjectConcept1 = null;
			// get subjectConcept for role1 and role2
			SimpleConceptRoleLevel tempRole = null;
			if (roleDomLevel1.ConceptUri.equals(roleDomLevel2.ConceptUri)) {
			    subjectConcept1 = roleDomLevel1.ConceptUri;
			    tempRole = roleDomLevel1; // Does not matter which one
			} else if (kwGen.checkConceptInclusion(roleDomLevel1.ConceptUri,roleDomLevel2.ConceptUri)) {
			    subjectConcept1 = roleDomLevel1.ConceptUri;
			    tempRole = roleDomLevel1;
			} else if (kwGen.checkConceptInclusion(roleDomLevel2.ConceptUri,roleDomLevel1.ConceptUri)) {
			    subjectConcept1 = roleDomLevel2.ConceptUri;
			    tempRole = roleDomLevel2;
			}
			
			if (subjectConcept1 != null) {
			 // get subjectConcept for intermediate role from above and role3
			    String subjectConcept2 = null;
			    
			    if (roleDomLevel3.ConceptUri.equals(tempRole.ConceptUri)) {
				subjectConcept2 = roleDomLevel3.ConceptUri;
			    } else if (kwGen.checkConceptInclusion(roleDomLevel3.ConceptUri,tempRole.ConceptUri)) {
				subjectConcept2 = roleDomLevel3.ConceptUri;
			    } else if (kwGen.checkConceptInclusion(tempRole.ConceptUri,roleDomLevel3.ConceptUri)) {
				subjectConcept2 = tempRole.ConceptUri;
			    }			    
			    			    
			    if (subjectConcept2 != null) {
				    return new Quartet<String, String, String, String>(subjectConcept2, roleDomLevel1.RoleUri,
					    roleDomLevel2.RoleUri, roleDomLevel3.RoleUri);				
			    }

			}
		    }
		}
	    }    
	}	
	
	return null;
	
    }    

    // Given four concepts, find a subject concepts and three roles.
    private Quintet<String, String, String, String, String> optimizeSpatialFeatureForFourRoles(String object1Uri, String object2Uri, 
	    String object3Uri, String object4Uri)
    {
	
	ArrayList<SimpleConceptRoleLevel> results1 =  kwGen.getSortedDomainObjRolesForRange(object1Uri);
	ArrayList<SimpleConceptRoleLevel> results2 =  kwGen.getSortedDomainObjRolesForRange(object2Uri);
	ArrayList<SimpleConceptRoleLevel> results3 =  kwGen.getSortedDomainObjRolesForRange(object3Uri);
	ArrayList<SimpleConceptRoleLevel> results4 =  kwGen.getSortedDomainObjRolesForRange(object4Uri);

	// Simple heuristics: Take the four roles which have the lowest level and 
	// C_1 subsumes C_2 (or opposite). Take the more specific (C_1).
	if(results1.size() > 0 && results2.size() > 0 && results3.size() > 0 && results4.size() > 0) {
	    
	    for (SimpleConceptRoleLevel roleDomLevel1 : results1) {
		for (SimpleConceptRoleLevel roleDomLevel2 : results2) {
		    for (SimpleConceptRoleLevel roleDomLevel3 : results3) {
 		      for (SimpleConceptRoleLevel roleDomLevel4 : results3) {

			
			String subjectConcept1 = null;
			// get subjectConcept for role1 and role2 
			SimpleConceptRoleLevel tempRole = null;
			if (roleDomLevel1.ConceptUri.equals(roleDomLevel2.ConceptUri)) {
			    subjectConcept1 = roleDomLevel1.ConceptUri;
			    tempRole = roleDomLevel1; // Does not matter which one
			} else if (kwGen.checkConceptInclusion(roleDomLevel1.ConceptUri,roleDomLevel2.ConceptUri)) {
			    subjectConcept1 = roleDomLevel1.ConceptUri;
			    tempRole = roleDomLevel1;
			} else if (kwGen.checkConceptInclusion(roleDomLevel2.ConceptUri,roleDomLevel1.ConceptUri)) {
			    subjectConcept1 = roleDomLevel2.ConceptUri;
			    tempRole = roleDomLevel2;
			}

			String subjectConcept2 = null;
			if (subjectConcept1 != null) {
			 // get subjectConcept for intermidiate role from above and role3
			    if (roleDomLevel3.ConceptUri.equals(tempRole.ConceptUri)) {
				subjectConcept2 = roleDomLevel3.ConceptUri;
			    } else if (kwGen.checkConceptInclusion(roleDomLevel3.ConceptUri,tempRole.ConceptUri)) {
				subjectConcept2 = roleDomLevel3.ConceptUri;
			    } else if (kwGen.checkConceptInclusion(tempRole.ConceptUri,roleDomLevel3.ConceptUri)) {
				subjectConcept2 = tempRole.ConceptUri;
			    }			    
			}

			String subjectConcept3 = null;
			if (subjectConcept1 != null) {
			 // get subjectConcept for intermidiate role from above and role3
			    if (roleDomLevel4.ConceptUri.equals(tempRole.ConceptUri)) {
				subjectConcept3 = roleDomLevel4.ConceptUri;
			    } else if (kwGen.checkConceptInclusion(roleDomLevel4.ConceptUri,tempRole.ConceptUri)) {
				subjectConcept3 = roleDomLevel4.ConceptUri;
			    } else if (kwGen.checkConceptInclusion(tempRole.ConceptUri,roleDomLevel4.ConceptUri)) {
				subjectConcept3 = tempRole.ConceptUri;
			    }			    
			}
			
			if (subjectConcept2 != null) {
				    return new Quintet<String, String, String, String, String>(subjectConcept2, roleDomLevel1.RoleUri,
					    roleDomLevel2.RoleUri, roleDomLevel3.RoleUri, roleDomLevel4.RoleUri);				
			}

			
 		      }
		    }
		}
	    }    
	}	
	
	return null;
	
    }     
    
    
    
    // Given an the object concept, find subject concept and predicate role
    private SimpleConceptRoleLevel optimizeSpatialFeatureForObject(String objectUri) // Pair<String, String>
    {
	ArrayList<SimpleConceptRoleLevel> results =  kwGen.getSortedDomainObjRolesForRange(objectUri); // Triplet<String, String, Integer>
	
	// Heuristics: As closer the range is to our attributeConcept, as more specific (to the attribute) is the objRole
	if(results.size() > 0) {
	    return results.get(0);
	}
	
	return null;
		
    }
 
    // Given an the object concept, find subject concept and predicate role
    private String optimizeSpatialFeatureForSubjectObject(String subjectUri, String objectUri)
    {
	ArrayList<SimpleConceptRoleLevel> results =  kwGen.getSortedDomainObjRolesForRange(objectUri);

	// Return the first to find (has the lowest level)
	if(results.size() > 0) {
	    
	    // Check if subjectConcept is equal or included by the roleDom Concept 
	    for(SimpleConceptRoleLevel roleDomLevel1: results) {
		if(roleDomLevel1.ConceptUri.equals(subjectUri) || 
			kwGen.checkConceptInclusion(subjectUri, roleDomLevel1.ConceptUri) )
		    return roleDomLevel1.RoleUri;
	    }	    
	    
	}
	
	return null;
	
	
    }
    
    private void callOwlgres(String sparqlQuery, String viewName, int columns)
	    throws RuntimeException {

	try {
	    queryRewrite.main(mHost,mDB, mUser, mPw,  true, viewName, "owl", columns, sparqlQuery);
	    
	} catch (RuntimeException e) {
	    throw e;
	} 
	

    }
    
     public String createSpatialQuery(String viewName, String viewName2, 
	    String boundingBox, Point pointOrigin, double distanceBetween, String spatialRelation, 
	    boolean schemaBasA, boolean measureDistanceStartEnd, Point pointStart, Point pointEnd, 
	    boolean events, String startTime, String viewName3, String spatialRelation2){  
	    
	    StringBuffer mainSQL = new StringBuffer();
	    String distanceQuery = null;
	    SchemaInfo.SchemaInfoItem schemaA = mSchemaA; 
	    SchemaInfo.SchemaInfoItem schemaB = mSchemaB;
	    SchemaInfo.SchemaInfoItem schemaC = mSchemaC;
	    if(schemaBasA) {
		schemaA = mSchemaB;
		schemaB = null;
		schemaC = null;
	    }
	    	    
	    String valuesXCoor = String.format("ST_X(ST_Transform(ST_Transform(ST_Centroid(gpm.%1$s), 3395), %2$s))", 
		    Util.createObjArray(schemaA.fieldGeo, m_OutputGeometry));
	    String valuesYCoor = String.format("ST_Y(ST_Transform(ST_Transform(ST_Centroid(gpm.%1$s), 3395), %2$s))", 
		    Util.createObjArray(schemaA.fieldGeo, m_OutputGeometry));
            String view_attr_geometry = String.format("ST_Transform(gpm.%1$s,%2$s)",Util.createObjArray(schemaA.fieldGeo,schemaA.geometryID));
            
            String prefixA = "";
            if (schemaA.joinFieldPrefix != null && schemaA.joinFieldPrefix.length() > 0) 
        	prefixA = "'" + schemaA.joinFieldPrefix +  "_' || ";
            String owlgres_attr_id = "owl1." + OwlgresViewField1;
            
            // This order has to be represented by the positions in class QueryResult
            String view_attr_name = String.format("rtrim(gpm.%1$s, ' ')", schemaA.fieldName);;
            if(events) 
        	view_attr_name = String.format("rtrim(evn.%1$s, ' ')", mSchemaEventA.fieldName);
            if(spatialRelation != null)
        	view_attr_name = String.format("rtrim(gpm.%1$s, ' ') || ' / ' || rtrim(gp2.%2$s, ' ')", schemaA.fieldName, schemaB.fieldName);    
        	
            String view_attr_additional = "";
            if(schemaA.fieldSource!=null) 
            	view_attr_additional = String.format("%1$s, gpm.%2$s", view_attr_additional, schemaA.fieldSource); // "%1$s, gpm.%2$s"
            if(schemaA.fieldPostcode!=null) 
            	view_attr_additional = String.format("%1$s, gpm.%2$s, gpm.%3$s", view_attr_additional, schemaA.fieldPostcode, schemaA.fieldCity);
            if(schemaA.fieldStreet!=null) 
            	view_attr_additional = String.format("%1$s, gpm.%2$s, gpm.%3$s", view_attr_additional,  schemaA.fieldStreet, schemaA.fieldHouseNr);

            if(!events) 
        	view_attr_additional = createAdditionalFields(view_attr_additional, schemaA, "gpm.");
            else
               	view_attr_additional = createAdditionalFields(view_attr_additional, mSchemaEventA, "evn.");

         
            String measureDistance = null;
            if(!measureDistanceStartEnd) {
    	    	String pointString = String.format("ST_SetSRID(ST_Point(%1$s,%2$s),%3$s)", Util.createObjArray(pointOrigin.getX(), pointOrigin.getY(), schemaA.geometryID));
    	    	measureDistance =  String.format("ST_Distance_Sphere(%1$s, %2$s) as dist", view_attr_geometry, pointString);
            }
            else {
    	    	String pointStartString = String.format("ST_SetSRID(ST_Point(%1$s,%2$s),%3$s)", Util.createObjArray(pointStart.getX(), pointStart.getY(), schemaA.geometryID));
    	    	String pointEndString = String.format("ST_SetSRID(ST_Point(%1$s,%2$s),%3$s)", Util.createObjArray(pointEnd.getX(), pointEnd.getY(), schemaA.geometryID));
      		measureDistance =  String.format("LEAST((ST_Distance_Sphere(%1$s, %2$s) * 1.2) , (ST_Distance_Sphere(%1$s, %3$s) * 1)) as dist", view_attr_geometry, pointStartString, pointEndString);
      		//measureDistance =  String.format("((ST_Distance_Sphere(%1$s, %2$s) * 0.6) + (ST_Distance_Sphere(%1$s, %3$s) * 0.4)) as dist", view_attr_geometry, pointStartString, pointEndString);
            }
            
            // Fields and FROM SpatialTable in Postgis
            mainSQL.append(String.format("(SELECT %1$s AS id, %2$s AS name, %4$s, %5$s as posx, %6$s as posy, ST_AsText(%3$s) as geom %7$s FROM %8$s AS gpm \n", 
        	    Util.createObjArray(owlgres_attr_id, view_attr_name, view_attr_geometry, measureDistance, valuesXCoor, valuesYCoor, view_attr_additional, schemaA.mainTable))); // view_attr_name2,
            // Join SpatialTable and Owlgres View
            mainSQL.append(String.format(" INNER JOIN %1$s AS owl1 ON (%2$s CAST(gpm.%3$s AS varchar)) = %4$s \n", 
        	    Util.createObjArray(viewName, prefixA, schemaA.fieldId, owlgres_attr_id)));
 
            if(events) {
        	mainSQL.append(String.format(" INNER JOIN %1$s AS evn ON CAST(%2$s AS varchar) = %3$s \n", 
        	    Util.createObjArray(mSchemaEventA.mainTable, "evn." + mSchemaEventA.fieldId, "owl1." + OwlgresViewField2)));
            } 
            
            // Make spatial join A
            if(spatialRelation != null) { 
 
                String owlgres2_attr_id = null;
                String prefixB = "";
                // Multiple views means joining to queries through spatial relations as (Near, Inside, Along)
                if(spatialRelation != null) { 
                    owlgres2_attr_id = viewName2 + "." + OwlgresViewField1;
                    if (schemaB.joinFieldPrefix != null && schemaB.joinFieldPrefix.length() > 0)  
                        prefixB = "'" + schemaB.joinFieldPrefix +  "_' || ";
                }
                String owlgres3_attr_id = null;
                String prefixC = "";
                // Multiple views means joining to queries through spatial relations as (Near, Inside, Along)
                if(spatialRelation2 != null) { 
                    owlgres3_attr_id = viewName3 + "." + OwlgresViewField1;
                    if (schemaC.joinFieldPrefix != null && schemaC.joinFieldPrefix.length() > 0)  
                        prefixC = "'" + schemaC.joinFieldPrefix +  "_' || ";
                }
       	
        	
        	distanceQuery = SpatialPredicatesMyITS.createDistanceQuery(spatialRelation, distanceBetween, schemaA.fieldGeo, schemaB.fieldGeo, schemaA.geometryID);
        	
        	mainSQL.append(String.format("  INNER JOIN %1$s AS gp2 ON %2$s AND CAST(gpm.%3$s AS varchar) <> CAST(gp2.%4$s  AS varchar)  \n",
        	       Util.createObjArray(schemaB.mainTable, distanceQuery, schemaA.fieldId, schemaA.fieldId)));
        	mainSQL.append(String.format("   INNER JOIN %1$s AS owl2 ON (%4$s CAST(gp2.%3$s AS varchar)) = owl2.att1 \n", 
        	       Util.createObjArray(viewName2, owlgres2_attr_id, schemaB.fieldId, prefixB))); // %2$s
        	
                // Make spatial join B
                if(spatialRelation2 != null) { 
                    distanceQuery = SpatialPredicatesMyITS.createDistanceQuery(spatialRelation2, distanceBetween, schemaA.fieldGeo, schemaC.fieldGeo, schemaA.geometryID);
                    mainSQL.append(String.format("  INNER JOIN %1$s AS gp3 ON %2$s AND CAST(gpm.%3$s AS varchar) <> CAST(gp3.%4$s  AS varchar)  \n",
                 	       Util.createObjArray(schemaC.mainTable, distanceQuery, schemaA.fieldId, schemaA.fieldId)));
                    mainSQL.append(String.format("   INNER JOIN %1$s AS owl3 ON (%4$s CAST(gp3.%3$s AS varchar)) = owl3.att1 \n", 
                 	       Util.createObjArray(viewName3, owlgres3_attr_id, schemaC.fieldId, prefixC))); // %2$s
                             
                }      	
        	
            }
            
            // WHERE Clause, check containment (&&) in bounding box
            String whereValA = String.format("ST_Transform(gpm.%1$s,%2$s)",Util.createObjArray(schemaA.fieldGeo,schemaA.geometryID));
	    String box3dA = String.format("ST_GeomFromEWKT('SRID=%1$s; %2$s')", Util.createObjArray(schemaA.geometryID,boundingBox));
            mainSQL.append(String.format("WHERE %2$s && %1$s ", Util.createObjArray(box3dA, whereValA)));
            
            if(spatialRelation != null) { 
        	String whereValB = String.format("ST_Transform(gp2.%1$s,%2$s)",Util.createObjArray(schemaB.fieldGeo,schemaB.geometryID));
        	mainSQL.append(String.format("AND %2$s && %1$s ", Util.createObjArray(box3dA, whereValB)));
        	
        	if(spatialRelation2 != null) {
            	    String whereValC = String.format("ST_Transform(gp3.%1$s,%2$s)",Util.createObjArray(schemaC.fieldGeo,schemaC.geometryID));
            	    mainSQL.append(String.format("AND %2$s && %1$s ", Util.createObjArray(box3dA, whereValC)));
       	
        	}
            }
            
            // Special treatment for events, has to be generalized
            if(events) {
        	// AND %1$s <= date_trunc('day', now()) AND  %2$s >= date_trunc('day', now())
        	String range = "";
        	if(startTime!=null) {
        	     range = String.format(" AND %1$s <= date_trunc('day', '%3$s'::date) AND  %2$s >= date_trunc('day', '%3$s'::date) ", 
        		    Util.createObjArray("evn." + mSchemaEventA.fieldDateStart, "evn." + mSchemaEventA.fieldDateEnd, startTime.replace('T',' ')));
        	}
        	else {
               	     range = String.format(" AND %1$s <= date_trunc('day', now()) AND  %2$s >= date_trunc('day', now())", 
        		    Util.createObjArray("evn." + mSchemaEventA.fieldDateStart, "evn." + mSchemaEventA.fieldDateEnd));
        	}
        	mainSQL.append(range); 
        	String dayBitmaks = dayAsBitmask();
        	if(mSchemaEventA.fieldDateIn != null) 
        	    mainSQL.append(String.format(" AND (%1$s & B'%2$s') > B'0000000' ", Util.createObjArray("evn." + mSchemaEventA.fieldDateIn, dayBitmaks)));
            } 
            
            mainSQL.append("ORDER BY dist )");

            return mainSQL.toString();
	    
	}


  	
     	private String dayAsBitmask() {
     	    
     	    Calendar calendar = Calendar.getInstance();
     	    //  Sunday is day=1
     	    int day = calendar.get(Calendar.DAY_OF_WEEK);
     	    
     	    switch(day) {
     	    	case 1: return "0000001";
     	    	case 2: return "1000000";
     	    	case 3: return "0100000";
     	    	case 4: return "0010000";
     	    	case 5: return "0001000";
   	    	case 6: return "0000100";
   	    	case 7: return "0000010";
    	     }
     	    
     	    return "0000000";
     	    
     	}
     
     	private String createAdditionalFields(String view_attr_additional, SchemaInfo.SchemaInfoItem schemaA, String tablePrefix) {
		
	        if(schemaA.fieldUrl!=null) 
	        	view_attr_additional = String.format("%1$s, %2$s", view_attr_additional, tablePrefix + schemaA.fieldUrl);
	        if(schemaA.fieldEmail!=null) 
	        	view_attr_additional = String.format("%1$s, %2$s", view_attr_additional, tablePrefix + schemaA.fieldEmail);
	        if(schemaA.fieldPhone !=null) 
	        	view_attr_additional = String.format("%1$s, %2$s", view_attr_additional, tablePrefix + schemaA.fieldPhone);
	        if(schemaA.fieldDesc!=null) 
	        	view_attr_additional = String.format("%1$s, %2$s", view_attr_additional, tablePrefix + schemaA.fieldDesc);
	        if(schemaA.fieldOpen!=null) 
	        	view_attr_additional = String.format("%1$s, %2$s", view_attr_additional, tablePrefix + schemaA.fieldOpen);
	        
	        return view_attr_additional;
		
	 }
	    
   	
    
}
