package edu.tuwien.kbs.search;

import java.io.FileInputStream;
import java.io.IOException;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.apache.lucene.spatial.DistanceUtils;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.PrecisionModel;
import com.vividsolutions.jts.operation.buffer.BufferOp;

import sun.reflect.generics.reflectiveObjects.NotImplementedException;
import edu.tuwien.kbs.*;
import edu.tuwien.kbs.db.*;
import edu.tuwien.kbs.geo.GeoUtil;
import edu.tuwien.kbs.keywords.SpatialPredicatesMyITS;
import edu.tuwien.kbs.ldm.ResultLDM;
import edu.tuwien.kbs.routing.*;
import edu.tuwien.kbs.util.*;

public class SemSearch implements ISemSearch {

    protected static PrecisionModel Precision_Model = new PrecisionModel(
	    PrecisionModel.FLOATING);
    protected static GeometryFactory Geometry_Factory = new GeometryFactory(
	    Precision_Model);

    protected static String QueryfileName = "input_con_query.rq";
    protected static String DefaultGeometry = "4326";
 
    public static String DefaultNamespacePrefixTBox= "geot";
    public static String DefaultNamespacePrefixABox= "geoa";

    public static String Router_Modality_Car = "CAR"; 
    public static String Router_Modality_Bike = "BIKE"; 
    public static String Router_Modality_Public = "PUBLIC"; 
    public static String Router_Modality_Foot = "FOOT"; 
    public static double Router_MeterPreSeconds_Foot = 1.4;  // 9 km/h
    public static double Router_MeterPreSeconds_Car = 12.78; // With average speed in vienna of 46 km/h
    public static double Router_MeterPreSeconds_Bike = 5.56; // 20 km/h
    public static double Router_MeterPreSeconds_Public = 8.33; // With average speed of 30 km/h
   
    public static int Buffer_Quadrant_Segments = 3;
    public static double Buffer_Distance_Iso = 0.001; // 0.00029;
    public static double Buffer_Distance_Iso_Small = 0.0001; 
    public static double Buffer_Distance_Box = 0.001;
   
    protected Map<String, Object> mCmd;
    protected ArrayList<String> mCmdList;

    protected ArrayList<String> mTemporaryViews;
    protected IStoreConnection mConnection; 
    protected long startTime;
    protected long lastTimePoint;
    protected boolean mMultipleViews = false;
    protected boolean mMultipleViews2 = false;
    protected boolean mEvents;
    
    // From properties
    protected String mHost;
    protected String mDB;
    protected String mUser;
    protected String mPw;
    protected String mPosgresGeoSchemaID = "gis"; // "public"
    protected String mPosgresOntoSchemaID = "owl"; // "owl"
    //protected String mAITServer;
    
     // From arguments
    protected boolean mLog = false;   // Turn on off for write files to log directory
    protected boolean mOuputResults = true; // Turn on/off for testing
    protected boolean mKeepViews = false ; // Turn on/off for testing
    protected static boolean OptimizeSpatialQuery = true; // Turn on off for testing

    protected String mStrategy;
    protected double mDist;
    protected double mDistBetween;
    protected int mResultSizePublic;
    protected int mComplexIsochrone = 0;
    
    // From call
    protected String mQuery;
    protected String mPoints;
    protected String mMots;
    protected int mTimeLimit;
    protected String mStartTime;
    protected String mUserId = null;
   
    protected QueryRewritter mRewritter;
    protected RouterFacade mRouterAccess;
    protected SchemaInfo mSchema;
    protected Map<String, String> mOptionsMap;
    
    
    public SemSearch() {

	mCmd = new HashMap<String, Object>();
	mCmdList = new ArrayList<String>();

	// Default settings
	readProperties();
	
	mSchema = new SchemaInfo();	
	mSchema.load(); // Create schema for Table A and B (todo: should have more queries)	
	mTemporaryViews = new ArrayList<String>();
	mMultipleViews = false;
	
	mRewritter = new QueryRewritter(DefaultGeometry, mHost, mDB, mUser, mPw);
	mRouterAccess = new RouterFacade();	

	lastTimePoint = 0;

    }

    public String evalQuery(boolean keywords) throws Exception {
	throw new NotImplementedException();
    }
    
 

    public void setStartCal(long s) {
	startTime = s;
	lastTimePoint = s;
    }
    
    public double calculateDistanceForMotsLimit() {
	if(mMots.equals(Router_Modality_Car)) {
	    return mTimeLimit * Router_MeterPreSeconds_Car;
	} else if(mMots.equals(Router_Modality_Bike)) {
	    return mTimeLimit * Router_MeterPreSeconds_Bike;
	} else if(mMots.equals(Router_Modality_Foot)) {
	    return mTimeLimit * Router_MeterPreSeconds_Foot;	
	} else if(mMots.equals(Router_Modality_Public)) {
	    return mTimeLimit * Router_MeterPreSeconds_Public;
	} 
	
	return 0;
    }
    
    public void printOut(RouteResults routes, boolean silent) {

	System.out.println("Result (Id, Name, Time (min), Distance (m)):");

	if (silent) {	   
	    System.out.println("Size: " + routes.size());  // Print count
	} else {
	    for (int i = 0; i < routes.size(); i++) {
		Route route = routes.get(i);
		StringBuffer s1 = new StringBuffer();
		s1.append(route.Id);
		s1.append("; ");
		s1.append(route.getPOI().Name);
		s1.append(" (" + route.getPOI().toCoord(true) +  ")");
		s1.append("; ");
		s1.append((route.Distance / 1));		
		s1.append("; ");
		s1.append((route.Time / 60000));

		System.out.println(s1.toString());
	    }
	}

    } 
    
    protected void readProperties() {
	
	Properties properties = new Properties();
	try {
	    
	  properties.load(new FileInputStream(SemSearchCli.PropertiesFileName));
	  mHost = properties.getProperty("dbhost");
	  mDB = properties.getProperty("database");
	  mUser = properties.getProperty("dbuser");
	  mPw = properties.getProperty("dbpassword");
	  mPosgresGeoSchemaID = properties.getProperty("dbschema_gis");
	  mPosgresOntoSchemaID = properties.getProperty("dbschema_ont");
	  mDistBetween = Integer.parseInt(properties.getProperty("geodistance_next"));
	  mResultSizePublic = Integer.parseInt(properties.getProperty("result_size_public")); 
	  mComplexIsochrone = Integer.parseInt(properties.getProperty("return_complex_isochrone")); 
	  //mAITServer = properties.getProperty("ait_server");
	  
	} catch (IOException e) {
	    e.printStackTrace();
	}
    }
    

    
    
    @Override
    public void parse(String[] args) {

	for (int i = 0; i < args.length; i++) {

	    mCmdList.add(args[i]);

	    if (args[i].startsWith("--")) {
		String key = args[i].substring(2);
		Object value = null;

		// Output
		if (key.equals("o")) { value = true; }
		// KeepViews
		else if (key.equals("k")) { value = true; }
		// d (Database), r1 (Radius 1), r2 (Radius 2),
		else {  value = args[i + 1]; }
		
		mCmd.put(key, value);
	    }
	}
	
	// mKeepViews = mCmd.containsKey("k");
	mDistBetween = Double.valueOf((String) mCmd.get("r2"));
	mStrategy = (String) mCmd.get("s");
	
    }

   public void cleanAllViews() {
	
	String sql = "select 'drop view if exists ' || viewname || ' cascade;' as a from pg_views where view_name LIKE 'V_';";
	
	connectDB();
	
	try {
	    Statement stmt1 = mConnection.createStatement();
	    ResultSet rset = stmt1.executeQuery(sql); // .toString()

	    while (rset.next()) {
	     String sqlDel = rset.getString("a");
	     
	     mConnection.executeStatement(sqlDel);
	     	     
	    }
	    
	    
	} catch (RuntimeException e) {
	    throw e;
	} catch (SQLException e) {
	    throw new RuntimeException(e.toString());
	}	
	
	
	closeDB(null);
	
    }
    
    
    protected String[] getInputPoints() {

	return mPoints.split(";");   
    }    
    
    protected void printTime(String text) {
	
	if (!mOuputResults ) return;	
	
	long actTime = System.nanoTime();
	long milliseconds = (actTime - lastTimePoint) / 1000000; // startTime
	lastTimePoint = actTime;
	System.out.println(milliseconds + " ms: " + text);
    }

    protected void printTimeTotal(String text) {
	
	if (!mOuputResults ) return;	
	
	long actTime = System.nanoTime();
	long milliseconds = (actTime - startTime) / 1000000; // startTime
	System.out.println(milliseconds + " ms: " + text);
    }
    
    protected void printTimeTotalExt(String text) {
	
	if (!mOuputResults ) return;	
	
	long actTime = System.nanoTime();
	long milliseconds = (actTime - startTime) / 1000000; // startTime
	System.out.println(text + "ms ;" + milliseconds);
    }

    
    protected void connectDB() {
	connectDB(true);
    }
    
    protected void connectDB(boolean externalSchemas) {
	// Connect to DB
	mConnection = new StoreConnectionPostgreSQL( mHost, mDB, mUser, mPw);
	mConnection.connect(java.sql.Connection.TRANSACTION_READ_COMMITTED);
	
	// Add schemas and optimizations
	// setUpDB(externalSchemas);

    }
    
    protected void closeDB(String sessionID) {

	// Clean up temporary views and close connection
	if(sessionID != null && sessionID.length() > 0) 
	    cleanViews(sessionID); 
	mConnection.close();


    }
    
    protected void setUpDB(boolean externalSchemas) {
	
	if(!externalSchemas) return;
	
	String schema = null;
	if(mPosgresGeoSchemaID != null && mPosgresGeoSchemaID.length() > 0) 
	    schema = "," + mPosgresGeoSchemaID;
	if(mPosgresOntoSchemaID != null && mPosgresOntoSchemaID.length() > 0) 
	    schema = schema + "," + mPosgresOntoSchemaID;

	mConnection.executeStatement("SET search_path TO \"$user\""  + schema + ";");
	// mConnection.executeStatement("SET enable_seqscan TO off;");
    }
    
 
    protected void cleanViews( String sessionId) { // boolean multipleViews,
	
	if (!mKeepViews) {
	    try {
		mConnection.executeStatement("DROP VIEW " + "owl." + QueryRewritter.OwlgresUnionsNameA + sessionId);
		if (mMultipleViews) {
		    mConnection.executeStatement("DROP VIEW " + "owl." + QueryRewritter.OwlgresUnionsNameB  + sessionId);
		    if (mMultipleViews2)
			mConnection.executeStatement("DROP VIEW " + "owl." + QueryRewritter.OwlgresUnionsNameC  + sessionId);
		}
		
		for (int i = 0; i < mTemporaryViews.size(); i++) {
		    String tempView = mTemporaryViews.get(i);
		    mConnection.executeStatement("DROP VIEW " + "owl."
			    + tempView);
		}
	    } catch (Exception e) {
		System.out.println("Error: " + e.toString());
	    }

	}
	
    }
    
    protected ArrayList<PointOfInterest> getResultsWithSpatialPredicats(String spatialRelation, String spatialRelation2,  
	    String  mainSQL, String mainSQL2,  int actualGeoID, ArrayList<PointOfInterest> initalPOIs, 
	    boolean doubleCheckContainment, Geometry[]  isochBufferFull) {

	ArrayList<PointOfInterest> filteredPOIs = null;

	// Get first filter results

        // Filter with first spatial relation
	if (spatialRelation.equals(SpatialPredicatesMyITS.KeywordQueryNext) || spatialRelation.equals(SpatialPredicatesMyITS.KeywordQueryAlong)) {
	    QueryResultGIS queryResult = getResults(mainSQL, actualGeoID, true);
	    // This has to be smarter, get Results2 according to
	    ArrayList<PointOfInterest> pois2 = queryResult.selectPOIs( doubleCheckContainment, isochBufferFull);
	    filteredPOIs = evalSpatialRelationNext(initalPOIs, pois2);
	} else if (spatialRelation.equals(SpatialPredicatesMyITS.KeywordQueryInside)) {
	    // Do not check containment, as for  larger objects (district), they  will be partly outside isochrone
            QueryResultGIS queryResult = getResults(mainSQL, actualGeoID, false);
	    ArrayList<AreaOfInterest> areas = queryResult.selectAreaOfInts(false, isochBufferFull); 
	    filteredPOIs = evalSpatialRelationInside(initalPOIs, areas);
	} else if (spatialRelation.equals(SpatialPredicatesMyITS.KeywordQueryDisjoint)) {
            QueryResultGIS queryResult = getResults(mainSQL, actualGeoID, false);
	    // Do not check containment, as for  larger objects (district), they  will be partly outside isochrone
	    ArrayList<AreaOfInterest> areas = queryResult.selectAreaOfInts(false, isochBufferFull); 
	    filteredPOIs = evalSpatialRelationDisjoint(initalPOIs, areas);
	} else if (spatialRelation.equals(SpatialPredicatesMyITS.KeywordQueryIntersect)) {
	} else if (spatialRelation.equals(SpatialPredicatesMyITS.KeywordQueryEquals)) {
	} else if (spatialRelation.equals(SpatialPredicatesMyITS.KeywordQueryNotEquals)) {
	}
	
        // Filter with second spatial relation
	if (spatialRelation2 != null)
	{
		// Get second filter results

        	if (spatialRelation2.equals(SpatialPredicatesMyITS.KeywordQueryNext) || spatialRelation2.equals(SpatialPredicatesMyITS.KeywordQueryAlong)) {
    	            QueryResultGIS queryResult2 = getResults(mainSQL2, actualGeoID, true);
        	    // This has to be smarter, get Results2 according to
        	    ArrayList<PointOfInterest> pois3 = queryResult2.selectPOIs( doubleCheckContainment, isochBufferFull);
        	    filteredPOIs = evalSpatialRelationNext(filteredPOIs, pois3);
        	} else if (spatialRelation2.equals(SpatialPredicatesMyITS.KeywordQueryInside)) {
        	    QueryResultGIS queryResult2 = getResults(mainSQL2, actualGeoID, false);
        	    // Do not check containment, as for  larger objects (district), they  will be partly outside isochrone
        	    ArrayList<AreaOfInterest> areas2 = queryResult2.selectAreaOfInts(false, isochBufferFull); 
        	    filteredPOIs = evalSpatialRelationInside(filteredPOIs, areas2);
        	} else if (spatialRelation2.equals(SpatialPredicatesMyITS.KeywordQueryDisjoint)) {
        	    QueryResultGIS queryResult2 = getResults(mainSQL2, actualGeoID, false);
        	    // Do not check containment, as for  larger objects (district), they  will be partly outside isochrone
        	    ArrayList<AreaOfInterest> areas2 = queryResult2.selectAreaOfInts(false, isochBufferFull); 
        	    filteredPOIs = evalSpatialRelationDisjoint(filteredPOIs, areas2);
        	} else if (spatialRelation2.equals(SpatialPredicatesMyITS.KeywordQueryIntersect)) {
        	} else if (spatialRelation2.equals(SpatialPredicatesMyITS.KeywordQueryEquals)) {
        	} else if (spatialRelation2.equals(SpatialPredicatesMyITS.KeywordQueryNotEquals)) {
        	}
	}

	return filteredPOIs;

    }

    
    protected ArrayList<PointOfInterest> evalSpatialRelationNext(ArrayList<PointOfInterest> pois1, ArrayList<PointOfInterest> pois2) { 
	      
	ArrayList<PointOfInterest> results = new ArrayList<PointOfInterest>(); // PointOfInterest
	
 	    for (int i = 0; i < pois1.size(); i++) {		
		PointOfInterest poi1 = (PointOfInterest) pois1.get(i);
		
		for (int j = 0; j < pois2.size(); j++) {
		    PointOfInterest poi2 = (PointOfInterest)  pois2.get(j);
		    
	     	    double dist = DistanceUtils.haversine(poi1.getPointX(), poi1.getPointY(), poi2.getPointX(), poi2.getPointY(), GeoUtil.EARTH_RADIUS);
	     	    if(dist <= mDistBetween) {
	     		poi1.Name = combineNames(poi1.Name, poi2.Name);
			results.add(poi1);
			break;	     		
	     	    }
		}
	    }
 	    
 	    return results;
   }
 
    protected ArrayList<PointOfInterest> evalSpatialRelationDisjoint(ArrayList<PointOfInterest> pois, ArrayList<AreaOfInterest> areas) {
        
  	ArrayList<PointOfInterest> results = new ArrayList<PointOfInterest>();
  
  	for (int i = 0; i < pois.size(); i++) {		
  	    PointOfInterest poi = pois.get(i);
            boolean isContained = false;
  	    
	     for (int j = 0; j < areas.size(); j++) {
  	  	AreaOfInterest aoi = areas.get(j);
  	  	if(aoi.Area == null) continue;
  	 
 	        if(aoi.Area.contains(poi.getPoint())) {
 	           poi.Name = combineNames(poi.Name, aoi.Name);
 	           isContained = true;
  		}
  	     }
	     if(! isContained)
		 results.add(poi);
  	}
  		   	
       	return results;
    
    }
    
    
    protected ArrayList<PointOfInterest> evalSpatialRelationInside(ArrayList<PointOfInterest> pois, ArrayList<AreaOfInterest> areas) {
        
  	ArrayList<PointOfInterest> results = new ArrayList<PointOfInterest>();
  		    		
  	for (int j = 0; j < areas.size(); j++) {
  	    AreaOfInterest aoi = areas.get(j);
  	    if(aoi.Area == null) continue;

  	    Geometry area2 = aoi.Area.buffer(Buffer_Distance_Iso_Small, Buffer_Quadrant_Segments, BufferOp.CAP_SQUARE);
		    		  
  	    for (int i = 0; i < pois.size(); i++) {		
  	  	PointOfInterest poi = pois.get(i);
  	 
 	        if(area2.contains(poi.getPoint())) {
 	           poi.Name = combineNames(poi.Name, aoi.Name);
 	           results.add(poi);
 	           //break;
  		}
  	    }
  		
  	}
  		   	
       	return results;
    
    }
      
    protected String combineNames(String name1, String name2) {
	
	if(name1 != null &&  name2 != null && !name1.equalsIgnoreCase(name2))
		return name1 + " / " + name2;
	else if(name1 != null &&  name2 == null)
	    return name1;
	else if(name2 != null &&  name1 == null)
	    return name2;
	else
	    return name1;
 
    }

    protected QueryResultGIS getResults(String mainSQL, int geoID, boolean isPOI) {

	 QueryResultGIS  queryResult = null;

	if (mLog) System.out.println("SQL: " + mainSQL);
	
	try {
	    Statement stmt = mConnection.createStatement();
	    ResultSet rset = stmt.executeQuery(mainSQL); // .toString()

	    // Fetch each row from the result set
	    ResultSetMetaData rsmd = rset.getMetaData();
	    queryResult = new QueryResultGIS("", geoID); // Postgres_Result_PosId, Postgres_Result_PosName, Postgres_Result_PosX, Postgres_Result_PosY);
	    
	    queryResult.fill(rset, rsmd.getColumnCount(), isPOI);
	    
	} catch (RuntimeException e) {
	    throw e;
	} catch (SQLException e) {
	    throw new RuntimeException(e.toString());
	}
	
	return queryResult;
	
   }
    
   



}


