package edu.tuwien.kbs.search;

import java.util.ArrayList;
import java.util.Map;
import java.util.Random;

import org.javatuples.*;

import com.vividsolutions.jts.geom.*;
import com.vividsolutions.jts.operation.buffer.BufferOp;
import com.vividsolutions.jts.operation.linemerge.LineMerger;

import edu.tuwien.kbs.db.*;
import edu.tuwien.kbs.geo.GeoUtil;
import edu.tuwien.kbs.geo.SimpleGMLWriter;
import edu.tuwien.kbs.routing.*;

public class SemSearchNR extends SemSearch {

    public static int Result_Limit = 10; // The 11 closest should be a good limit as it reduces the amount of routes calculated (before 20, 15)
    public static int DefaultTimeLimit = 900; // 15 minutes in seconds
    protected boolean doubleCheckContainment = true; // false

    protected static SimpleGMLWriter Simple_GMLWriter = new SimpleGMLWriter();
    
      
    
    public SemSearchNR() {
	super();

	mStrategy = "B1"; // Default strategy call AIT isochrone
	
    }
    
    public SemSearchNR(String query, String from, String mots, int timeLimit, 
	    String startTime, String userId,  Map<String, String> optionsMap) { // boolean log, boolean outputResults, boolean keepViews
	this();
	
	mQuery = query; mPoints = from; mMots = mots; mTimeLimit = timeLimit; 
	mStartTime = startTime; mUserId = userId; mOptionsMap = optionsMap;
	// mLog = log; mOuputResults = outputResults; mKeepViews = keepViews;

	if(mTimeLimit==0) mTimeLimit= DefaultTimeLimit;
	
    }
    
    @Override
    public String evalQuery(boolean keywords) throws Exception  {

	String spatialRelation = null;
	String spatialRelation2 = null;

	// Create Result and general id, which will be used to get for the route and iso details
	Random ranGen = new Random();
	String sessionID = Integer.toString(ranGen.nextInt(Integer.MAX_VALUE) ); // Long.toString(ranGen.nextLong() );)
		
	printTime("Eval NR Started: " +  sessionID); // + mQuery + " (" + mPoints + ")"
	
	// Connect and set up DB
	connectDB();
	
	// Read input points (can be several)
	String[] points = mPoints.split(";");

	// Create Onto Part
	Septet<String, String, String, String, String, String, String> unionSqls = null;
	if(keywords)
	    unionSqls = mRewritter.createKBQueryFromKeywords(mQuery, mSchema, mTemporaryViews, sessionID); 
	else
	    unionSqls = mRewritter.createKBQueryFromCQ(mQuery, mSchema, mTemporaryViews, sessionID); 
	    
	
	// Detect events
	if(unionSqls.getValue6() != null) 
	    mEvents = true;
	
	// Create Union of A views
	if (unionSqls.getValue0() != null) {
	    mConnection.executeStatement(unionSqls.getValue0()); 
	    // Create Union of B views
	    if (unionSqls.getValue1() != null) {
		mMultipleViews = true;
		mConnection.executeStatement(unionSqls.getValue1());  
		spatialRelation = unionSqls.getValue3();
		// Create Union of C views
		if (unionSqls.getValue2() != null) {
		    mMultipleViews2 = true;
		    mConnection.executeStatement(unionSqls.getValue2());  
		    spatialRelation2 = unionSqls.getValue4();
		}
	    }
	}
	else {
	    // If no results possible return empty results
	    RouteResults routes =  mRouterAccess.getEmptyRoutes(sessionID, null,  mMots, mStartTime, mTimeLimit, unionSqls.getValue5()); 	
	    return routes.serializeToXml(true);	    
	}
	printTime("Onto Query Created: " + unionSqls.getValue5());
	
	// No dealing with multiple points for now	
	// We do not handle multiple input points, future work...
	int actualGeoID = mRewritter.GetInfoItemA().geometryID;
	String[] ps = points[0].split("[,]"); 
	Point origin = GeoUtil.createPoint(ps[0],ps[1], actualGeoID, true); 


	// Create Corridor for query, considering each point
	String isochBufferWKB = null;
	Geometry[]  isochBufferFull = null;
	if (mStrategy.equals("A")) {
	    
	    printTime("Corridor Received");
	    // Simple Bounding box
	    isochBufferWKB = GeoUtil.createBox2DForIso(origin, actualGeoID, mDist, mLog);
	    doubleCheckContainment = false;

	} else if (mStrategy.contains("B")) {

	    // Call AIT Router and calculate Isochrone
	    Geometry[] gc = mRouterAccess.getIsochrone(sessionID, origin, mMots, mStartTime, mTimeLimit, mLog, mUserId, mOptionsMap);
	    printTime("Corridor Received");
	    
	    // Calc Corridor
	    isochBufferFull = createBufferForIso(gc, mLog, mStrategy, sessionID);	    
	    isochBufferWKB = GeoUtil.writeToWKB(isochBufferFull,true); // Simply Buffer for SQL
	    doubleCheckContainment = false; // true;
	    
	    // Store for details
	    mRouterAccess.storeIsochroneDetails(sessionID, gc,  Geometry_Factory, mComplexIsochrone, mLog);
	}
	printTime("Corridor Calculated");
	
	// Create SQL spatial query; Not a good solution now, because infoItemA and B could be different for the underlying relations
	String mainSQL1 = null; String mainSQL2 = null; String mainSQL3 = null;
	if(!OptimizeSpatialQuery || spatialRelation == null) {
	    mainSQL1 = mRewritter.createSpatialQuery(QueryRewritter.OwlgresUnionsNameA + sessionID, QueryRewritter.OwlgresUnionsNameB + sessionID, 
		    isochBufferWKB, origin, mDistBetween, spatialRelation, false, false, null, null, mEvents, mStartTime, 
		    QueryRewritter.OwlgresUnionsNameC + sessionID, spatialRelation2);  
	}
	else {
	    mainSQL1 = mRewritter.createSpatialQuery(QueryRewritter.OwlgresUnionsNameA + sessionID, null, isochBufferWKB, origin, 
		    mDistBetween, null, false, false, null, null, mEvents, mStartTime, null, null); 
	    mainSQL2 = mRewritter.createSpatialQuery(QueryRewritter.OwlgresUnionsNameB + sessionID, null, isochBufferWKB, origin, 
		    mDistBetween, null, true, false, null, null, false, mStartTime, null, null); 
	    if(spatialRelation2 != null) {
		    mainSQL3 = mRewritter.createSpatialQuery(QueryRewritter.OwlgresUnionsNameC + sessionID, null, isochBufferWKB, origin, 
			    mDistBetween, null, true, false, null, null, false, mStartTime, null, null); 
	    }
	}
	printTime("Spatial Query Created");
	
	// Execute main SQL
	QueryResultGIS queryResult = getResults(mainSQL1, actualGeoID, true);	
	// Checks the isochrone again (because some of Postgis results are not correct, and check for spatial relation if optimized)
	ArrayList<PointOfInterest> pois = queryResult.selectPOIs(doubleCheckContainment, isochBufferFull); 

	if(OptimizeSpatialQuery && spatialRelation != null) {
	    pois = getResultsWithSpatialPredicats(spatialRelation, spatialRelation2, mainSQL2, mainSQL3, actualGeoID, pois, doubleCheckContainment, isochBufferFull);
	}
	
	printTime("Query Run with Instances: " + pois.size());
	
	// Call router for each route. 
	int resLimit = Result_Limit;
	//if(mMots.contains(Router_Modality_Public) && mResultSizePublic > 0) 
	//    resLimit = mResultSizePublic; // Change result size for PUBLIC
	
	RouteResults routes =  mRouterAccess.getRoutes(sessionID, origin, pois, resLimit, mMots, 
		mStartTime, mTimeLimit, false, mLog, unionSqls.getValue5(), mUserId, mOptionsMap); 		
	printTime("Routes Calculated: (" + routes.getStats()  +  ")");

	// Close and clean up DB
	closeDB(sessionID);
	
	printTimeTotal("Eval NR Ended" );
	
	// Print results
	if (mOuputResults && mLog)  printOut(routes,false);	    
	
	return routes.serializeToXml(true);
	

    }

     	
  protected Geometry[]  createBufferForIso(Geometry[] geometries, boolean log, String strategy, String sessionID) { 

	if (strategy.equals("B1")) {

	    for (int i = 0; i < geometries.length; i++) { //
		Geometry g = geometries[i];
		if (g != null) {
		    Geometry gb = g.buffer(Buffer_Distance_Iso, Buffer_Quadrant_Segments, BufferOp.CAP_SQUARE); // BufferOp.CAP_ROUND
		    geometries[i] = gb;
		}
	    }

	} else if (strategy.contains("B2")) {
	    LineMerger lineMerger = new LineMerger();
	    GeometryCollection gc = new GeometryCollection(geometries,
		    Geometry_Factory);
	    lineMerger.add(gc);
	    ArrayList merged_geos = (ArrayList) lineMerger.getMergedLineStrings();
	    ArrayList<Geometry> merged_geos2 = new ArrayList<Geometry>();

	    // Buffer for each linestring
	    for (int i = 0; i < merged_geos.size(); i++) { // geometries.length
		Geometry g = (Geometry) merged_geos.get(i); // geometries[i];
		if (g != null) {
		    Geometry gb = g.buffer(Buffer_Distance_Iso, Buffer_Quadrant_Segments, BufferOp.CAP_SQUARE); 
		    merged_geos2.add(gb); 
		}
	    }
	    geometries = merged_geos2 .toArray(new Geometry[merged_geos2.size()]);

	}
	
	// Write Log
	if (log)  Simple_GMLWriter.writeGeometry(geometries, "log/BufferIso_" + sessionID + ".gml");
	

	return geometries; // retVal;

    }

    
  
   
}
