package edu.tuwien.kbs.search;

import java.util.ArrayList;
import java.util.Map;
import java.util.Random;

import org.javatuples.*;

import com.vividsolutions.jts.geom.*;

import edu.tuwien.kbs.db.QueryResultGIS;
import edu.tuwien.kbs.geo.GeoUtil;
import edu.tuwien.kbs.routing.PointOfInterest;
import edu.tuwien.kbs.routing.RouteResults;

public class SemSearchVIA extends SemSearch {

    public static int Eval_Poi_Threshold = Integer.MAX_VALUE; // 105;
    public static int Result_Limit = 10; // The 10 closest should be ok for now
    public static double BUFFER_Distance_Box1 = 0.002; // 0.0015
    public static double BUFFER_Distance_Box2 = 0.02;     
    protected static String PointfileName = "input_points_VIA.rq";
    
    
    public SemSearchVIA() {
	super();
	
   }

    public SemSearchVIA(String query, String from_to, String mots, int timeLimit, String startTime, 
	    boolean log, boolean outputResults, String userId,  Map<String, String> optionsMap) {
	this();
	
	mQuery = query; mPoints = from_to; mMots = mots; mTimeLimit = timeLimit; 
	mStartTime = startTime; mLog = log; mOuputResults = outputResults; mUserId = userId; mOptionsMap = optionsMap;

    
    }
	
    @Override
    public String evalQuery(boolean keywords) throws Exception {
	
	String spatialRelation = null;
	//boolean doubleCheckContainment = true;
	
	// Create Result and general id, which will be used to get for the route and iso details
	Random ranGen = new Random();
	String sessionID = Integer.toString(ranGen.nextInt(Integer.MAX_VALUE));
	
	printTime("Eval Via Started for: " + sessionID); // mQuery + " (" + mPoints + ")"
	
	// Connect and set up DB
	connectDB();
	
	// Read input points (can be several)
	String[] points = mPoints.split(";");
//	if(points.length < 2 || points[0].length() < 2 || points[1].length() < 2) {
//	    throw new Exception("Invalid input parameter: Point A and B have to be defined as --p \"p1x,p1y;p2x,p2y\" ");
//	}

	// Create Onto Part
	Septet<String, String, String, String, String, String, String> unionSqls = null;
	if(keywords)
	    mRewritter.createKBQueryFromKeywords(mQuery,  mSchema, mTemporaryViews, sessionID);
	else
	    mRewritter.createKBQueryFromCQ(mQuery,  mSchema, mTemporaryViews, sessionID);
	    
	
	// Detect events
	if( unionSqls.getValue6() != null) mEvents = true;
	// Create Union of A views
	if (unionSqls.getValue0() != null) {
        	mConnection.executeStatement(unionSqls.getValue0()); // sqlA);
        	// Create Union of B views
        	if (unionSqls.getValue1() != null) {
        	    mConnection.executeStatement(unionSqls.getValue1()); // sqlB);
    		    mMultipleViews = true;
    		    spatialRelation = unionSqls.getValue3();
        	}
	}
	else {
	    // If no results possible return empty results
	    RouteResults routes =  mRouterAccess.getEmptyRoutes(sessionID, null,  mMots, mStartTime, mTimeLimit, unionSqls.getValue5()); 	
	    return routes.serializeToXml(true);	    
	}

	printTime("Onto Query Created: " + unionSqls.getValue5());
		
	// We only point A to point B
	int actualGeoID = mRewritter.GetInfoItemA().geometryID;
	String[] psA = points[0].split("[,]"); 
	Point pointA = GeoUtil.createPoint(psA[0],psA[1], actualGeoID, true); 
	String[] psB = points[1].split("[,]"); 
	Point pointB = GeoUtil.createPoint(psB[0],psB[1], actualGeoID, true); 

	
	// Get statistics and est. size for query
	int estimSize = estimateSize();
	RouteResults routes = null;
	
	// Few POIs, query of POIs and 2 routes from A to POI to B are calculated
	if(estimSize < Eval_Poi_Threshold) {

	    // Create a bounding box considering point A and B
	    Pair<Geometry, Point> retPair = GeoUtil.createBox2DForIso(pointA, pointB, BUFFER_Distance_Box1, actualGeoID, mLog);
	    String box2D = GeoUtil.writeToWKB(new Geometry[] {retPair.getValue0()}, false); //  retPair.getValue0();	    
	    printTime("Corridor Calculated");
	    
	    // Take Centroid of point for distance measure. This is an heuristic, but better then using start- and endpoint
	    // Create SQL spatial query;
	    String mainSQL1 = null; String mainSQL2 = null; 
	    boolean measureStartEnd = false;
	    
	    if(mMots.contains(Router_Modality_Public))
		    measureStartEnd = true; // If MOT public, weight points that are closer to start/end higher, otherwise weight for points in the middle are best
	    
	    if(!OptimizeSpatialQuery || spatialRelation == null) {
		mainSQL1 = mRewritter.createSpatialQuery(QueryRewritter.OwlgresUnionsNameA + sessionID, QueryRewritter.OwlgresUnionsNameB + sessionID, 
		    box2D, retPair.getValue1(), mDistBetween, spatialRelation, false, measureStartEnd, pointA, pointB, 
		    mEvents, mStartTime, null, null); 
	    }
	    else {
		    mainSQL1 = mRewritter.createSpatialQuery(QueryRewritter.OwlgresUnionsNameA + sessionID, null, box2D, retPair.getValue1(), 
			    mDistBetween, null, false, measureStartEnd, pointA, pointB, mEvents, mStartTime, null, null); 
		    mainSQL2 = mRewritter.createSpatialQuery(QueryRewritter.OwlgresUnionsNameB + sessionID, null, box2D, retPair.getValue1(), 
			    mDistBetween, null, true, measureStartEnd, pointA, pointB, false, mStartTime, null, null);	   
	    }
	    printTime("Spatial Query Created");
	    
	    // Execute main SQL
	    QueryResultGIS queryResult = getResults(mainSQL1, actualGeoID, true);
	    printTime("Query Run with Instances: " + queryResult.getSize());
	    
	    // If no results returned, extend search range of Box2DForIso
	    if(queryResult.getSize() == 0) {
		    retPair = GeoUtil.createBox2DForIso(pointA, pointB, BUFFER_Distance_Box2, actualGeoID, mLog);
		    box2D = GeoUtil.writeToWKB(new Geometry[] {retPair.getValue0()}, false); //  retPair.getValue0();	    
		    // printTime("Corridor Calculated 2. time (Var. A)");

		    mainSQL1 = mRewritter.createSpatialQuery(QueryRewritter.OwlgresUnionsNameA + sessionID, 
			   QueryRewritter.OwlgresUnionsNameB + sessionID,  box2D, retPair.getValue1(), mDistBetween, 
			   spatialRelation, false, measureStartEnd, pointA, pointB, mEvents,mStartTime, null, null); 

		    // printTime("Spatial Query Created 2. time");
		    queryResult = getResults(mainSQL1, actualGeoID, true);
		    printTime("Query Run 2. time with Instances: " + queryResult.getSize() );		
	    }

	    ArrayList<PointOfInterest> pois = queryResult.selectPOIs(false, null); // isochBuffer Result_Limit

	    if(OptimizeSpatialQuery && spatialRelation != null) {
		pois = getResultsWithSpatialPredicats(spatialRelation, null, mainSQL2, null, actualGeoID, pois, false, null);
	    }

	    // Call router for each route
	    int resLimit = Result_Limit;
	    //if(mMots.contains(Router_Modality_Public) && mResultSizePublic > 0) 
	    //	resLimit = mResultSizePublic; // Reduse result size for PUBLIC

	    routes = mRouterAccess.getViaRoutes(sessionID, pointA, pointB, pois, resLimit, mMots, 
		    mStartTime, mTimeLimit, false, mLog, unionSqls.getValue5(), mUserId, mOptionsMap);
	    printTime("Routes Calculated: (" + routes.getStats()  +  ")");
	    
	}
	// Many POIs, alternative routes and corridor are first calculated, then query of POIs
	else {	    
	    
	    printTime("Var. B, Corridor Calculated");

	    
	}
	
	// Close and clean up DB
	closeDB(sessionID);

	printTimeTotal("Eval Via Ended" );
	
	// Print results
	if (mOuputResults && mLog)  printOut(routes,false);	    

	return routes.serializeToXml(true);
	
    }
       
    protected int estimateSize() { // String[] queries
    
	// For now just return random nr
	Random randomGen = new Random();   
	return randomGen.nextInt(10);
	
	
    }
    

}
