package edu.tuwien.kbs.util;

import java.util.HashMap;
import java.util.Map;

import edu.tuwien.kbs.keywords.KeywordGenerator;


public class SchemaInfo {

    	protected static String SchemaFileName =  "schema_gis.conf";
	protected Map<String, SchemaInfoItem> schemaItems; 
	protected static String DefaultItem = "DEFAULT";
	
	public SchemaInfo()
	{
	    schemaItems = new HashMap<String, SchemaInfoItem>();
	    schemaItems.put(SchemaInfo.DefaultItem, new SchemaInfoItem());
	}
	    
	public void load() {
		    
	    String[] items =  Util.getContent(SchemaFileName);
	    
	    // Start inserting data
	    for (int i = 0; i < items.length; i++) {
		String[] vals = items[i].split("[;]");

		SchemaInfoItem info = new SchemaInfoItem(vals[1].trim(), vals[2].trim(), vals[3].trim(), 
			vals[4].trim(), vals[5].trim(), Integer.parseInt(vals[6].trim()));		    

		if(vals.length > 7)  info.fieldSource= vals[7].trim();
		if(vals.length > 8)  info.fieldPostcode = vals[8].trim();
		if(vals.length > 9)  info.fieldCity = vals[9].trim();
		if(vals.length > 10)  info.fieldStreet = vals[10].trim();
		if(vals.length > 11)  info.fieldHouseNr = vals[11].trim();
		if(vals.length > 12)  info.fieldUrl = vals[12].trim();
		if(vals.length > 13)  info.fieldEmail = vals[13].trim();
		if(vals.length > 14)  info.fieldPhone = vals[14].trim();
		if(vals.length > 15)  info.fieldDesc = vals[15].trim();
		if(vals.length > 16)  info.fieldOpen = vals[16].trim();
		if(vals.length > 17)  info.fieldDateStart = vals[17].trim();
		if(vals.length > 18)  info.fieldDateEnd = vals[18].trim();
		if(vals.length > 19)  info.fieldDateIn = vals[19].trim();

		schemaItems.put(vals[0], info);
	    }
	}

	public SchemaInfoItem getDefaultInfoItem() {
	    SchemaInfoItem retVal = null;
	    
	    if(schemaItems.containsKey(DefaultItem)) 
		retVal = schemaItems.get(DefaultItem);
	    
	    return retVal;
	    
	}
	
	public SchemaInfoItem getInfoItem(String[] keys, KeywordGenerator kGen) { // String key
	    
	    SchemaInfoItem retVal = null;
	    
	   for (int i = 0; i < keys.length; i++) {
	   
	       String key = keys[i];
	   
		    // First choice, direct entry
		    if(schemaItems.containsKey(key))  retVal = schemaItems.get(key);
		    
		    if(retVal!=null)  {
			break;
		    }
		    else{			
			// Second choice, entry whith a super-concept (can be improved to chooce not any but the closet s. c.)
			for (Map.Entry<String, SchemaInfoItem> entry : schemaItems.entrySet()) {
			    String schemaKey = entry.getKey();
			    if(kGen.checkConceptInclusion(key, schemaKey)) {
				retVal = entry.getValue();
				break;
			    }
			}
		    }	    
	   }
	    	    
	    // Last choice default 
	    if(retVal==null) retVal = schemaItems.get(DefaultItem);
	    
	    return retVal;
	
	}
	    
	
	public class SchemaInfoItem {
	    
	    public String mainTable = null; // "gis.planet_osm_point";
	    public int geometryID = 0;  // String  "4326"; // "900913"; 4326 is WGS84 and 900913 is Mercador
	    public String fieldGeo = null; //  "way";
	    public String fieldId = null; //  "osm_id" ;
	    public String fieldName = null; //  "name";
	    public String joinFieldPrefix = null; //  "";
	    public String fieldSource = null;

	    public String fieldPostcode = null;
	    public String fieldCity = null;
	    public String fieldStreet = null;
	    public String fieldHouseNr = null;
	    public String fieldUrl = null;
	    public String fieldEmail = null;
	    public String fieldPhone = null;
	    public String fieldDesc = null;
	    public String fieldOpen = null;
	    
	    public String fieldDateStart = null;
	    public String fieldDateEnd = null;
	    public String fieldDateIn = null;
	    
	    public SchemaInfoItem()
	    {}
	    
	    public SchemaInfoItem(String table,  String geoField, String idField, String nameField, String prefix, int geoID) {//, String pc, String cty, String str, String hnr) {
		mainTable = table;
		geometryID = geoID;
		fieldGeo = geoField;
		fieldId = idField;
		fieldName = nameField;
		joinFieldPrefix = prefix;
		//postgisFieldPostcode = pc;
		//postgisFieldCity = cty;
		//postgisFieldStreet = str;
		//postgisFieldHouseNr = hnr;
	    }
	    
	}

	
}
